/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core;

import com.kerio.im.common.error.Code;
import com.kerio.im.core.Shutdown;
import com.kerio.im.core.api.event.ShutdownEvent;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.remote.CommandEvent;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Handle im service shutdown. Try to close all internal threads. Try to close
 * ExtService and XmppService. Shutdown system with appropriate result code.
 *
 * @author tskoda
 */
public class Shutdown {
    
    private static enum STATE {
        DEFAULT,
        SHUTDOWN_CALLED;
    }

    private static final Logger LOG = Logger.getLogger(Shutdown.class.getName());
    private static final String MESSAGE = "control";
    private static final String SHUTDOWN = "service-im-shutdown";
    private static final String SHUTDOWNTIMEOUT = "service-im-shutdown-timeout";
    private static final int TIMEOUT = 10; // 10 seconds in miliseconds
    private final ScheduledExecutorService timeoutExecutor;
    private final ExecutorService shutdownExecutor;
    private final ThreadPool threadPool;
    private final ServiceBus bus;
    private STATE state;

    public Shutdown(final ServiceBus bus, final ThreadPool threadPool) {
        bus.subscribe(CommandEvent.class, new ShutdownListener());
        this.bus = bus;
        this.threadPool = threadPool;
        this.timeoutExecutor = Executors.newSingleThreadScheduledExecutor(new CustomThreadFactory(SHUTDOWNTIMEOUT, true));
        this.shutdownExecutor = Executors.newSingleThreadExecutor(new CustomThreadFactory(SHUTDOWN, false));
        this.state = STATE.DEFAULT;
    }

    public void shutdown(Code code) {
        synchronized(this) {
            if (this.state == STATE.DEFAULT) {
                this.state = STATE.SHUTDOWN_CALLED;
                this.shutdownExecutor.submit(new ShutdownTask(code));
                this.timeoutExecutor.schedule(new ShutdownTimeoutTask(code), TIMEOUT, TimeUnit.SECONDS);
            }
        }
    }

    private class ShutdownTimeoutTask implements Runnable {

        private final Code code;

        public ShutdownTimeoutTask(Code code) {
            this.code = code;
        }

        @Override
        public void run() {
            LOG.log(Level.WARNING, "IM service shutdown timeout, forcing runtime to close");
            Runtime.getRuntime().halt(this.code.ordinal()); //bypass shutdown hooks, finalization, force closing
        }
    }

    private class ShutdownTask implements Runnable {

        private final Code code;

        public ShutdownTask(Code code) {
            this.code = code;
        }

        @Override
        public void run() {
            //any exception must not be thrown
            try {
                LOG.log(code != Code.OK? Level.WARNING: Level.INFO, "Performing IM service shutdown");
                Shutdown.this.bus.publish(new ShutdownEvent(this.code), new ServiceBus.ExceptionHandler() {

                    @Override
                    public boolean pass(Exception e) {
                        LOG.log(Level.FINE, "Notifying shutdown listener failed, continue with next", e);
                        return true;
                    }
                });
                
                System.exit(this.code.ordinal()); //call all shutdown hooks
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Performing IM service shutdown failed", e);
            }
        }
    }

    private class ShutdownListener implements ServiceBus.EventListener<CommandEvent> {
        public static final String EXIT_MESSAGE = "exit";
        
        
        
        @Override
        public void event(final CommandEvent event) {
            if (!event.interested(MESSAGE)) {
                return;
            }

            if (event.getContent().containsKey(EXIT_MESSAGE) && EXIT_MESSAGE.equals(event.getContent().get(EXIT_MESSAGE))) {
                LOG.log(Level.FINE, "Received stop command message");
                Shutdown.this.shutdown(Code.OK);
            }
        }
    }
}
