/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.cert;

import com.kerio.im.bean.Certificate;
import com.kerio.im.common.error.Code;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.service.CertificateService;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.configuration.StoreConfiguration;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.external.ExtService;
import com.kerio.im.xmpp.api.XmppDomains;
import com.kerio.im.xmpp.api.XmppException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class Certificates implements CertificateService {

    private static final Logger LOG = Logger.getLogger(Certificate.class.getName());
    private final ExtService extService;
    private final CertificateStore serverCertificates;
    private final XmppDomains domains;

    public Certificates(final ExtService extService, final XmppDomains domains, final ServiceBus bus, final StoreConfiguration configuration) {
        this.extService = extService;
        this.domains = domains;
        this.serverCertificates = new CertificateStore(configuration.get(GlobalKey.SERVICE_REMOTE_DOMAIN));
        loadCertificates();
        
        domains.update(this.serverCertificates.map()); 
        bus.subscribe(CommandEvent.class, new CertificateEventSync(this));
    }

    private void loadCertificates() {
        LOG.log(Level.FINE, "Getting server certificate from external service");
        
        this.extService.getServerCertificates(new ExtService.CertificateHandler() {
            
            @Override
            public boolean process(String id) {
                return true; //we want all 
            }
            
            @Override
            public boolean handle(String id, Set<String> aliases, String certificate, String key, boolean isDefault, boolean isValid) {
                Certificate cert = new Certificate(id, isValid, certificate, key);
                serverCertificates.put(cert, aliases, isDefault);
                return true; //we want all 
            }
        });
        
        LOG.log(Level.INFO, "Getting server certificate from external service successful");
    }

    public void updateCertificate(String newDefaultCertificateId) {
        final Map<String, Certificate> result = this.serverCertificates.updateDefault(newDefaultCertificateId);
        
        if (! result.isEmpty()) {
            this.domains.update(result);
        } else {
            throw new XmppException("Certificate with {0} identifier doesn't exist", Code.RESOURCE_ERROR, newDefaultCertificateId);
        }
    }

    public void addCertificate(String newCertificateId) {
        final Map<String, Certificate> result = new HashMap<>();
        
        this.extService.getServerCertificates(new ExtService.CertificateHandler() {
            
            @Override
            public boolean process(String id) {
                return newCertificateId.equals(id);
            }

            @Override
            public boolean handle(String id, Set<String> aliases, String certificate, String key, boolean isDefault, boolean isValid) {
                Certificate cert = new Certificate(id, isValid, certificate, key);
                result.putAll(serverCertificates.add(cert, aliases));
                return false; //we found our certificate
            }
            
        });
        
        if (! result.isEmpty()) {
            this.domains.update(result);
        } else {
            throw new XmppException("Certificate with {0} identifier doesn't exist", Code.RESOURCE_ERROR, newCertificateId);
        }
        
    }

    public void removeCertificate(String id) {
        final Map<String, Certificate> result = this.serverCertificates.removeById(id);
        
        if (! result.isEmpty()) {
            this.domains.update(result);
        }
        
        LOG.log(Level.INFO, "Removing certificate for {0} aliases successful", result.keySet());
        
    }
    
}
