/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.derby.jdbc.EmbeddedDataSource40;

/**
 *
 * @author tskoda
 */
public class DerbyConnectionFactory {

    private static final Logger LOG = Logger.getLogger(DerbyConnectionFactory.class.getName());
    private static final String CREATE = "create";

    public static DerbyConnectionFactory create(DerbyDestination destination) {
        return new DerbyConnectionFactory(destination, new DerbyConnectionFactoryImpl());
    }

    public static void close(ResultSet resultSet) {
        if (resultSet != null) {
            try {
                resultSet.close();
            } catch (SQLException e) {
                LOG.log(Level.INFO, "Can not close result set", e);
            }
        }
    }

    public static void close(Statement statement) {
        if (statement != null) {
            try {
                statement.close();
            } catch (SQLException e) {
                LOG.log(Level.INFO, "Can not close statement", e);
            }
        }
    }

    public static void close(Connection connection) {
        if (connection != null) {
            try {
                connection.close();
            } catch (SQLException e) {
                LOG.log(Level.INFO, "Can not close connection", e);
            }
        }
    }

    public static void rollback(Connection connection) {
        if (connection != null) {
            try {
                connection.rollback();
            } catch (SQLException e) {
                LOG.log(Level.INFO, "Can not rollback connection", e);
            }
        }
    }

    public static class DerbyConnectionFactoryImpl {

        public Connection newConnection(String path) throws SQLException {
            EmbeddedDataSource40 ds = new EmbeddedDataSource40();
            ds.setDatabaseName(path);
            ds.setCreateDatabase(CREATE);
            Connection result = ds.getConnection();
            result.setAutoCommit(false);
            return result;
        }

        public Connection shutdownConnection(String path) throws SQLException {
            EmbeddedDataSource40 ds = new EmbeddedDataSource40();
            ds.setDatabaseName(path);
            ds.setShutdownDatabase("shutdown");

            return ds.getConnection();
        }

    }

    private final DerbyDestination destination;
    private final DerbyConnectionFactoryImpl derbyConnectioFactory;

    public DerbyConnectionFactory(DerbyDestination destination, DerbyConnectionFactoryImpl derbyConnectionFactory) {
        this.destination = destination;
        this.derbyConnectioFactory = derbyConnectionFactory;
    }

    /**
     * Open if exist or create and open if does not exist. If database seems to
     * be corrupted (folder exists some files inside but not openable) the
     * current folder is going to be renamed (if unsuccessful, IOException is
     * thrown) and new db created.
     *
     * You need to check schema and perform upgrade by yourself or apply default
     * schema when this is new db.
     *
     * @return true only if new db was created
     * @throws IOException when disk or permission error occurs
     * @throws SQLException
     */
    public boolean open() throws SQLException, IOException {
        boolean create = !this.destination.exists();
        Connection testConnection = null;

        try {
            LOG.log(Level.FINE, "Openning database {0}, {1}", new Object[]{this.destination.toPath(), create ? "creating new one" : "using existing"});
            testConnection = this.derbyConnectioFactory.newConnection(this.destination.toPath());

        } catch (SQLException e) {
            LOG.log(Level.FINE, "Openning database {0} failed, database is probably incomplete ({1}:{2})", new Object[]{this.destination.toPath(), e.getErrorCode(), e.getSQLState()});
            this.destination.repair();
            create = true;

            LOG.log(Level.FINE, "Openning database {0}, {1}", new Object[]{this.destination.toPath(), create ? "creating new one" : "using existing"});
            testConnection = this.derbyConnectioFactory.newConnection(this.destination.toPath());
        } finally {
            close(testConnection);
        }

        LOG.log(Level.FINE, "Openning database {0} successful", this.destination.toPath());
        return create;
    }

    public Connection newConnection() throws SQLException {
        return this.derbyConnectioFactory.newConnection(this.destination.toPath());
    }

    public void shutdown() {
        try {
            this.derbyConnectioFactory.shutdownConnection(this.destination.toPath());
            LOG.log(Level.FINE, "Shutdown database {0} failed", this.destination.toPath());
        } catch (SQLException ex) {
            String engineShutdownSucc = "XJ015";
            String databaseShutdownSucc = "08006";
            if (engineShutdownSucc.equals(ex.getSQLState()) || (databaseShutdownSucc.equals(ex.getSQLState()))) {
                //ok
            } else {
                LOG.log(Level.FINE, "Shutdown database {0} successful", this.destination.toPath());
            }
        } catch (Exception ex) {
            LOG.log(Level.FINE, "Shutdown database {0} failed", ex);
        }
    }

}
