/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class DerbyDestination {

    private static final Logger LOG = Logger.getLogger(DerbyDestination.class.getName());
    
    static final String RENAME_FORMAT = "%1$s.tmp.%2$d";
    static final int RENAME_SEQUENCE_START = 1;

    static DerbyDestination open(File parent, String name) throws IOException {
        DerbyDestination result = new DerbyDestination(parent, name);
        try {
            result.open();
        } catch (DestinationExistException e) {
            LOG.log(Level.INFO, "Failed to open destination {0}, folder exists. Going to move original files.", new File(parent, name).getAbsolutePath());
            result.repair();
        }
        return result;
    }
    
    static class DestinationExistException extends IOException {
    }

    private final File parent;
    private final File database;
    private Boolean exists;

    DerbyDestination(File parent, String name) {
        this.parent = parent;
        this.database = new File(this.parent, name);
    }

    public void repair() throws IOException {
        if (this.exists == null) {
            throw new IOException("Destination not openned, call open() first");
        } else if (this.exists) {
            if (emptyFolder()) {
                this.exists = !delete();
            } else { //move no matter if it looks like derby folder or not
                this.exists = !move();
            }
        }
    }

    public boolean exists() throws IOException {
        if (exists == null) {
            throw new IOException("Destination not openned, call open() first");
        }
        return this.exists.booleanValue();
    }

    public String toPath() {
        return database.getAbsolutePath();
    }

    void open() throws DestinationExistException {
        this.exists = exist();

        if ((this.exists) && (! looksLikeDerby())) {
            throw new DestinationExistException();
        }
    }
    
    private boolean emptyFolder() {
        String[] children = this.database.list();
        return this.database.isDirectory() && (children == null || children.length == 0);
    }
    
    private boolean delete() throws IOException {
        LOG.log(Level.FINE, "Deleting empty folder {0}", this.database.getAbsoluteFile());

        if (this.database.delete()) {
            return true;
        } else {
            throw new IOException("Folder " + this.database.getAbsolutePath() + " can't be deleted");
        }  
    }

    private boolean move() throws IOException {
        File uniqueName = uniqueFile();
        LOG.log(Level.FINE, "Renaming folder {0} to {1}", new Object[]{this.database.getAbsoluteFile(), uniqueName});

        if (this.database.renameTo(uniqueName)) {
            return true;
        } else {
            throw new IOException("Folder " + this.database.getAbsolutePath() + " can't be renamed");
        }
    }

    private File uniqueFile() {
        String name = this.database.getName();
        File result = null;
        int count = RENAME_SEQUENCE_START;
        while (result == null) {
            String newName = String.format(RENAME_FORMAT, name, count);
            File file = new File(this.parent, newName);
            if (!file.exists()) {
                result = file;
            }
            count++;
        }
        return result;
    }

    private boolean looksLikeDerby() {
        File seq0 = new File(this.database, "seg0");
        File log = new File(this.database, "log");
        File properties = new File(this.database, "service.properties");

        return seq0.isDirectory() && log.isDirectory() && properties.isFile();
    }

    private boolean exist() {
        return this.database.exists() && this.database.canRead() && this.database.canWrite();
    }

}
