/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import com.kerio.im.core.CustomThreadFactory;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.event.ShutdownEvent;
import com.kerio.im.core.api.service.ConfigurationService;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.conversation.ConversationServiceImpl;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.external.ExtService;
import com.kerio.im.external.ExtService.ArchiveSettings;
import java.io.File;
import java.io.IOException;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class Archive {

    private enum MESSAGE_OPERATION {

        archive("archive");

        public final String val;

        private MESSAGE_OPERATION(String name) {
            this.val = name;
        }
    }

    private enum MESSAGE_PARAM {

        event("event"),
        export_source("export.source"),
        export_destination("export.destination");

        public final String val;

        private MESSAGE_PARAM(String name) {
            this.val = name;
        }

    }

    private enum EVENT_VALUE {

        changed,
        export
    }

    public static final String CONVERSATION_LISTENER_NAME = "archive";
    private static final String IM_SUBFOLDER_DEFAULT = "xmpp";
    private static final Logger LOG = Logger.getLogger(Archive.class.getName());

    public static Archive open(ConversationServiceImpl conversationService, ExtService extService, ConfigurationService configuration, ServiceBus bus) {

        try {
            ArchiveSettings settings = extService.getArchiveSettings();
            File archiveParent = new File(settings.getPath());
            File archiveDestination = new File(archiveParent, IM_SUBFOLDER_DEFAULT);
            Long limit = configuration.get(GlobalKey.ARCHIVE_FILE_LIMIT) * 1000 * 1000; //MB -> B

            FileStore archiveStore = FileStore.open(archiveDestination, limit);
            try {
                Archive archive = new Archive(conversationService, archiveStore);
                bus.subscribe(CommandEvent.class, archive.new ArchiveEventListener(extService));
                bus.subscribe(ShutdownEvent.class, archive.new ArchiveShutdownListener());
                try {
                    archive.refresh(settings.isEnabled());
                    LOG.log(Level.INFO, "Openning message archive component successful");
                    return archive;
                } catch (IOException e) {
                    archive.close();
                    throw e;
                }
            } catch (IOException e) {
                archiveStore.close();
                throw e;
            }
        } catch (Exception e) {
            throw new ServiceException("Openning message archive component failed, messages aren't going to be archived", e, Code.INTERNAL_SERVER_ERROR);
        }
    }

    private final ConversationServiceImpl conversationRegister;
    private final FileStore archiveStore;
    private Boolean lastState;

    Archive(ConversationServiceImpl conversationRegister, FileStore archiveStore) {
        this.conversationRegister = conversationRegister;
        this.archiveStore = archiveStore;
        this.lastState = null; //UKNOWN STATE
    }

    public synchronized void refresh(boolean enabled) throws IOException {
        if ((this.lastState == null) || (!this.lastState.equals(enabled))) {

            //anyway close current archive if already initialized
            if (this.lastState != null) {
                close();
            }

            if (enabled) {
                this.archiveStore.open();
                this.conversationRegister.register(
                        CONVERSATION_LISTENER_NAME,
                        new AsyncFileStore(
                                Executors.newSingleThreadExecutor(new CustomThreadFactory("im-service-archive", false)),
                                this.archiveStore,
                                new ExceptionHandlerImpl()));
            }

            this.lastState = enabled;
        } else {
            LOG.log(Level.INFO, "{0} message archiving skipped, current state not changed", this.lastState == null || this.lastState == false ? "Stopping" : "Starting");
        }
    }

    synchronized void export(String source, String target) throws IOException {
        this.archiveStore.export(source, target);
    }

    /**
     * Close Archive. Unregister all registered conversation listeners. Close
     * archive store. Doesn't throw any exception.
     */
    synchronized void close() {

        ConversationServiceImpl.AsyncConversationService maybePreviouslyRegistered = this.conversationRegister.unregister(CONVERSATION_LISTENER_NAME);

        if ((maybePreviouslyRegistered != null) && (maybePreviouslyRegistered != this.archiveStore)) {
            maybePreviouslyRegistered.close();
            LOG.log(Level.INFO, "Closing stale asynchronous file storage successful");
        }

        if (this.archiveStore != null) {
            this.archiveStore.close();
            LOG.log(Level.INFO, "Closing archive store successful");
        }

    }

    class ExceptionHandlerImpl implements ServiceBus.ExceptionHandler {

        @Override
        public boolean pass(Exception e) {
            LOG.log(Level.INFO, "Failed to archive IM message", e);
            LOG.log(Level.SEVERE, "Failed to archive IM mesage, can't write to archive folder {0}", Archive.this.archiveStore.getDestination());
            return true;
        }

    }

    class ArchiveShutdownListener implements ServiceBus.EventListener<ShutdownEvent> {

        @Override
        public void event(ShutdownEvent event) {
            close();
        }

    }

    class ArchiveEventListener implements ServiceBus.EventListener<CommandEvent> {

        private final ExtService extService;

        public ArchiveEventListener(ExtService extService) {
            this.extService = extService;
        }

        @Override
        public void event(CommandEvent event) {
            if (!event.interested(MESSAGE_OPERATION.archive.name())) {
                return;
            }

            try {
                EVENT_VALUE eventType = EVENT_VALUE.valueOf(event.get(MESSAGE_PARAM.event.val));

                switch (eventType) {
                    case changed:
                        try {
                            boolean enabled = this.extService.getArchiveSettings().isEnabled();
                            try {
                                refresh(enabled);
                                LOG.log(Level.INFO, "{0} message archive component successful", enabled ? "Starting" : "Stopping");
                            } catch (IOException e) {
                                LOG.log(Level.INFO, (enabled ? "Starting" : "Stopping") + " message archive component failed", e);
                                //report as error when starting/stopping, administrator need to take an action
                                LOG.log(Level.SEVERE, "{0} message archive component failed, reason: {1}", new Object[]{enabled ? "Starting" : "Stopping", e.getMessage()});
                                event.fail(e.getMessage());
                            }
                        } catch (ServiceException e) {
                                LOG.log(Level.INFO, "Starting / Stopping message archive component failed", e);
                                //report as error when starting/stopping, administrator need to take an action
                                LOG.log(Level.SEVERE, "Starting / Stopping message archive component failed, reason: {0}", e.getMessage());
                                event.fail(e.getMessage());
                        }

                        break;

                    case export:
                        //only warnings when exporting, this is not critical
                        String source = event.get(MESSAGE_PARAM.export_source.val);
                        String destination = event.get(MESSAGE_PARAM.export_destination.val);
                        if ((source == null) || (source.isEmpty())) {
                            LOG.log(Level.WARNING, "Exporting archive failed, reason: no source provided");
                            event.fail("No source provided");
                        }

                        if ((destination == null) || (destination.isEmpty())) {
                            LOG.log(Level.WARNING, "Exporting archive failed, reason: no destination provided");
                            event.fail("No destination provided");
                        }

                        try {
                            export(source, destination);
                            LOG.log(Level.INFO, "Exporting archive successful");
                        } catch (IOException e) {
                            LOG.log(Level.INFO, "Exporting archive failed", e);
                            LOG.log(Level.WARNING, "Exporting archive failed, reason: {0}", e.getMessage());
                            event.fail(e.getMessage());
                        }

                        break;
                    default:
                        throw new Exception("Processing archive event failed, event " + eventType + " handler not implemented");
                }

                //ok response is added by default
            } catch (Exception e) {
                //unknown error is reported to error log, administrator may need to take an action
                LOG.log(Level.INFO, "Processing archive event failed", e);
                LOG.log(Level.SEVERE, "Processing archive event failed, reason: {0}", e.getMessage());
                event.fail(e.getMessage());
            }
        }
    }

}
