/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class FileStoreWriter {

    private static final Logger LOG = Logger.getLogger(FileStoreWriter.class.getName());
    private static String NEWLINE = "\n";
    protected static DateTimeFormatter DATE_TIME_FORMAT = DateTimeFormatter.ofPattern("dd/MM/yyyy HH:mm:ss", Locale.ENGLISH);

    static class SizeCheck {

        public static final int DEFAULT_CHECK_INTERVAL = 16;

        private final File file;
        private final long limit;
        private final int checkInterval;
        private int count = 0;
        
        SizeCheck(File file, long limit) {
            this(file, limit, DEFAULT_CHECK_INTERVAL);
        }

        SizeCheck(File file, long limit, int checkInterval) {
            this.file = file;
            this.limit = limit;
            this.checkInterval = checkInterval;
        }

        public boolean allow() {
            boolean result = true;
            if (++count % checkInterval == 0) {
                result = this.file.length() < limit;
            }

            return result;
        }
    }

    static class SizeCheckException extends IOException {
    }

    private final SizeCheck sizeCheck;
    private final OutputStreamWriter output;
    private final File file;

    FileStoreWriter(File file, SizeCheck sizeCheck) throws IOException {
        this(file, sizeCheck, new OutputStreamWriter(new FileOutputStream(file, true), "UTF-8"));
    }

    FileStoreWriter(File file, SizeCheck sizeCheck, OutputStreamWriter output) {
        this.sizeCheck = sizeCheck;
        this.output = output;
        this.file = file;
    }

    public void write(LocalDateTime time,  String s) throws IOException {
        String line = s.replace("\n", "");
        if (line.length() > 0) {
            if (!this.sizeCheck.allow()) {
                throw new SizeCheckException();
            }
            
            this.output.write("[" + time.format(DATE_TIME_FORMAT) + "] " + line + NEWLINE);
            this.output.flush(); //we want to archive line by line
        }
    }

    public void close() {
        try {
            this.output.flush();
            this.output.close();
        } catch (IOException e) {
            LOG.log(Level.INFO, "Failed to close file output", e);
        }
    }

    public File file() {
        return this.file;
    }
}
