/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.domain;

import com.kerio.im.bean.Domain;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.cert.Certificates;
import com.kerio.im.xmpp.api.XmppDomains;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class DomainControl {

    private static final Logger LOG = Logger.getLogger(DomainControl.class.getName());
    private final XmppDomains xmppDomains;
    private final Domain serviceDomain;
    private final ServiceBus bus;
    /**
     * Map where domain name is a key. We need such a mapping as we obtain only
     * domain name from external xmpp service.
     */
    private Map<String, Domain> current;

    public DomainControl(XmppDomains xmppDomains, Domain serviceDomain, ServiceBus bus) {
        this.xmppDomains = xmppDomains;
        this.serviceDomain = serviceDomain;
        this.bus = bus;
        this.current = new HashMap<String, Domain>();
    }

    public void remove(Domain domain) {
        remove(domain.getName());
    }

    public boolean remove(String removedDomainName) {
        boolean result = false;
        //never ever remove service domain
        if (!(this.serviceDomain.getName().equals(removedDomainName))) {
            LOG.log(Level.FINE, "Removing domain {0}", removedDomainName);
            Domain removed = this.current.remove(removedDomainName);
            try {
                if (removed != null) {
                    this.bus.publish(new DomainEvent(DomainEvent.Type.REMOVED, removed));
                }
            } finally {
                result = true;
                this.xmppDomains.remove(removedDomainName);
            }
        }
        
        return result;
    }

    public boolean add(Domain addedDomain) {
        boolean result = false;
        if (!this.current.containsKey(addedDomain.getName())) {
            LOG.log(Level.FINE, "Adding domain {0}", addedDomain);
            this.xmppDomains.activate(addedDomain.getName());
            this.current.put(addedDomain.getName(), addedDomain);
            this.bus.publish(new DomainEvent(DomainEvent.Type.ADDED, addedDomain));
            result = true;
        }
        return result;
    }

    int count() {
        return this.current.size();
    }

    Set<Domain> allExceptService() {
        Set<Domain> result = new HashSet<Domain>(this.current.values());
        result.remove(this.serviceDomain);
        return result;
    }

    public boolean check(Domain domain) {
        boolean result = false;
        if (!(this.serviceDomain.getName().equals(domain.getName()))) {
            if (!this.current.containsKey(domain.getName())) {
                this.add(domain);
            } else {
                LOG.log(Level.FINE, "Checking domain {0}", domain);
                this.bus.publish(new DomainEvent(DomainEvent.Type.FOUND, domain));
            }
            result = true;
        }
        return result;
    }

    public void repair(Domain domain) {
        if (this.current.containsKey(domain.getName())) {
            //no repair actions yet
            this.bus.publish(new DomainEvent(DomainEvent.Type.REPAIRED, domain));
        }
    }
    
    public boolean isLocalDomain(String domain) {
        return this.current.containsKey(domain);
    }
}
