/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.logging;

import com.kerio.im.common.Lookup;
import com.kerio.im.common.TypedProperties;
import com.kerio.im.core.api.CoreLogging;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;

/**
 *
 * @author development
 */
public class ConnectHandler extends Handler {

    private static final int MAX_ONE_LINE_LOG_PARAM_LENGTH = 400;
    private static final int MAX_ONE_LINE_LOG_LENGTH = 1000;
    private static final byte[] NEW_LINE = "\n".getBytes();
    /**
     * Stream to write to (Std.err or pipe)
     */
    private OutputStream out;
    /**
     * Used to format time
     */
    private SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss");
    /**
     * if the output should be shrinked to one line
     */
    private boolean simple;
    private ConnectFilter connectFilter;
    private StanzaAnonymizer stanzaAnonymizer;

    public ConnectHandler() {
        this(LoggingManager.getProperties());
    }

    protected ConnectHandler(TypedProperties properties) {
        this(properties, new ConnectFilter(properties), new StanzaAnonymizer());
    }

    protected ConnectHandler(TypedProperties properties, ConnectFilter filter, StanzaAnonymizer stanzaFilter) {
        this.connectFilter = filter;
        this.stanzaAnonymizer = stanzaFilter;
        this.setFilter(filter);
        this.simple = properties.get(LogKeys.ENGINE_SIMPLE);
        this.out = properties.get(LogKeys.ENGINE_OUTPUT);
    }

    @Override
    public void publish(LogRecord record) {
        try {
            if (isLoggable(record)) {
                publishImpl(record);
            }

        } catch (Exception e) {
            LoggingManager.STDERR_STREAM.println(e.getMessage());
        }
    }

    private void publishImpl(LogRecord record) throws IOException {
        StringBuilder builder = new StringBuilder(MAX_ONE_LINE_LOG_PARAM_LENGTH);

        addType(record, builder);
        addThreadId(record, builder);
        addTime(record, builder);

        boolean limitLength = simple && record.getLevel() != Level.SEVERE;

        addMessage(record, builder, limitLength);

        addSource(record, builder);
        addThrown(record, builder, this.simple);

        String message = builder.toString();
        compactMessage(message, limitLength);

        out.write(message.getBytes());
        out.write(NEW_LINE);
        out.flush();
    }

    private void addThreadId(LogRecord record, StringBuilder builder) {
        builder.append("[");
        builder.append(record.getThreadID());
        builder.append("] ");
    }

    private void addType(LogRecord record, StringBuilder builder) {
        if (this.connectFilter.isForeign(record.getLoggerName())) {
            builder.append("D");
        } else if (record.getLevel().intValue() >= Level.SEVERE.intValue()) {
            builder.append("E");
        } else if (record.getLevel().intValue() >= Level.WARNING.intValue()) {
            builder.append("W");
        } else {
            builder.append("D");
        }
    }

    private void addTime(LogRecord record, StringBuilder builder) {
        if (!simple) {
            builder.append("[");
            builder.append(sdf.format(record.getMillis()));
            builder.append("] ");
        }
    }

    private void addMessage(LogRecord record, StringBuilder builder, boolean limitLength) {
        if (record.getMessage() != null) {
            Object[] params = record.getParameters();
            String message = record.getMessage();
            if (message.contains("\n")) {
                message = message.replaceAll("\n", " ");
            }

            if (params != null) {
                for (int i = 0; i < params.length; i++) {
                    String str = params[i] != null ? params[i].toString() : "null";
                    if (simple) {
                        if (str.contains("\n")) {
                            str = str.replaceAll("\n", " ");
                        }
                        if (limitLength && str.length() > MAX_ONE_LINE_LOG_PARAM_LENGTH) {
                            str = str.substring(0, MAX_ONE_LINE_LOG_PARAM_LENGTH - 3) + "...";
                        }
                    }
                    message = message.replace("{" + i + "}", str);
                }
            }

            builder.append(this.stanzaAnonymizer.anonymize(record.getLoggerName(), message));
        }
    }

    private void addSource(LogRecord record, StringBuilder builder) {
        builder.append(" (");
        String className = record.getSourceClassName();
        if (className.contains(".")) {
            className = className.substring(className.lastIndexOf('.') + 1);
        }
        builder.append(className);
        if ((className != null) && (!className.isEmpty())) {
            builder.append(".");
        }
        builder.append(record.getSourceMethodName());
        builder.append(")");
    }

    private void addThrown(LogRecord record, StringBuilder builder, boolean flatten) {
        Throwable exception = record.getThrown();

        if ((exception == null) && (record.getParameters() != null) && (record.getParameters().length > 0)) {
            Object possiblyException = record.getParameters()[record.getParameters().length - 1];
            if (possiblyException instanceof Throwable) {
                exception = (Throwable) possiblyException;
            }
        }

        if (exception != null) {
            builder.append(": ");
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            PrintStream output = flatten ? new PrintStream(new TraceFilterStream(baos)) : new PrintStream(baos);
            exception.printStackTrace(output);
            builder.append(new String(baos.toByteArray()));
        }
    }

    private String compactMessage(String message, boolean limit) {
        String result;
        if (simple) {
            result = message.replaceAll("(\\.java)|(com\\.kerio\\.im\\.)|(java\\.lang\\.)|(java\\.)", "");
            result = result.replaceAll(" at ", "/");
            result = result.trim();
            if (limit && result.length() > MAX_ONE_LINE_LOG_LENGTH) {
                result = result.substring(0, MAX_ONE_LINE_LOG_LENGTH);
            }
        } else {
            result = message;
        }

        return result;
    }

    @Override
    public void flush() {
        try {
            this.out.flush();
        } catch (IOException e) {
            LoggingManager.STDERR_STREAM.println("Can not flush connect logger");
        }
    }

    @Override
    public void close() throws SecurityException {
    }
}
