/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.logging;

import com.kerio.im.common.CustomLevel;
import com.kerio.im.common.PrintWriterToLogger;
import com.kerio.im.common.TypedProperties;
import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import com.kerio.im.core.api.CoreLogging;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.remote.CommandEvent;
import java.io.ByteArrayInputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

/**
 * Manages logging configuration. Apply changes.
 *
 * @author tskoda
 */
public class LoggingManager implements CoreLogging {

    private static final Logger LOG = Logger.getLogger(LoggingManager.class.getName());
    /**
     * Redirected stdout LEVEL
     */
    private static final Level STDOUT_LEVEL = new CustomLevel("STDOUT", Level.INFO.intValue() + 1);
    /**
     * Redirected stderr LEVEL
     */
    private static final Level STDERR_LEVEL = new CustomLevel("STDERR", Level.SEVERE.intValue() + 1);
    /**
     * Logging properties.
     */
    private static final LoggingConfiguration CONFIGURATION;
    /**
     * Original error stream for output
     */
    static final PrintStream STDERR_STREAM;

    static {
        STDERR_STREAM = System.err;
        CONFIGURATION = new LoggingConfiguration(new TypedProperties());
    }

    static TypedProperties getProperties() {
        return CONFIGURATION.getProperties();
    }

    private static void install() {
        try {
            LogManager.getLogManager().readConfiguration(new ByteArrayInputStream(LoggingConfiguration.CONFIGURATION));

        } catch (Exception e) {
            throw new ServiceException("Parsing log configuration failed", e, Code.INTERNAL_SERVER_ERROR);
        }
    }

    private static void redirectStandardStreams() {
        if (!CONFIGURATION.isStdRedirected()) {
            Logger errLog = Logger.getLogger("std.err");
            PrintStream errStream = new PrintWriterToLogger(errLog, STDERR_LEVEL);
            System.setErr(errStream);

            Logger outLog = Logger.getLogger("std.out");
            PrintStream outStream = new PrintWriterToLogger(outLog, STDOUT_LEVEL);
            System.setOut(outStream);
            CONFIGURATION.stdRedirected();
        }
    }

    public static void install(Map<String, String> cmdProperties) {
        LoggingManager.CONFIGURATION.configure(cmdProperties);
        LoggingManager.install();
        LoggingManager.redirectStandardStreams();
    }

    public LoggingManager(ServiceBus bus) {
        bus.subscribe(CommandEvent.class, new CommandsListener());
    }

    protected class CommandsListener implements ServiceBus.EventListener<CommandEvent> {

        private static final String TYPE = "log";

        @Override
        public void event(CommandEvent event) {

            if (!event.interested(TYPE)) {
                return;
            }

            LOG.log(Level.FINE, "Configuring logging service");
            try {
                Map<String, String> properties = new HashMap<String, String>();
                properties.put(LogKeys.DEBUG.key, event.get(LogKeys.DEBUG.key));
                LoggingManager.CONFIGURATION.configure(properties);
                LoggingManager.install();
                LOG.log(Level.INFO, "Configuring logging service successful");
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Configuring logging service event failed", e);
                event.fail(e.getMessage());
            }
        }
    }
}
