/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import com.kerio.im.bean.Port;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 *
 * @author tskoda
 */
public class CombinedFilter implements Port.Filter {
    
    public static CombinedFilter without() {
        return new CombinedFilter();
    } 
    
    public static CombinedFilter with(Port.Filter... filters) {
        return new CombinedFilter(filters);
    }

    private final List<Port.Filter> filterChain;
    private boolean allEnabled;

    private CombinedFilter() {
        this.filterChain = new LinkedList<Port.Filter>();
        this.allEnabled = true;
    }

    private CombinedFilter(Port.Filter... filters) {
        this();
        if ((filters != null) && (filters.length > 0)) {
            this.filterChain.addAll(Arrays.asList(filters));
        }
    }

    @Override
    public boolean pass(InetAddress client) {

        boolean result = false;
        
        synchronized (this.filterChain) {
            
            if (!this.filterChain.isEmpty()) { //there is at least one enabled filter
                for (Port.Filter filter : this.filterChain) {
                    if (result = filter.pass(client)) {
                        break; //if any pass, then pass
                    }
                }
            } else { 
                //if empty filters and !allEnabled => all filters are disabled, disable all connection
                //if empty filters and allEnabled => no filter configured, pass
                result = this.allEnabled; 
            }
        }

        return result;
    }
    
    public void updateUnknown() {
        this.allEnabled = false;
    }

    public void update(Port.Filter filter, boolean enabled) throws UnknownHostException {
        synchronized (this.filterChain) {
            if (enabled) {
                this.filterChain.add(filter);
            } else {
                this.allEnabled = false;
            }
        }
    }

    public void update(CombinedFilter filter) {
        synchronized (this.filterChain) {
            this.filterChain.clear();
            this.filterChain.addAll(filter.filterChain);
            this.allEnabled = filter.allEnabled;
        }
    }

    @Override
    public String toString() {
        return "CombinedFilter{" + "filterChain=" + filterChain + ", allEnabled=" + allEnabled + '}';
    }
    
    
}
