/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import com.kerio.im.core.ThreadPool;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.domain.Domains;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.core.remote.RemoteControl;
import com.kerio.im.xmpp.api.event.PingEvent;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class EventPing implements ServiceBus.EventListener<CommandEvent> {

    private static final String MESSAGE = "get";
    private static final String TYPE = "type";
    private static final String TYPE_PING = "ping";
    private static final Logger LOG = Logger.getLogger(EventPing.class.getName());
    private static final long TIMEOUT = 20;
    private static final TimeUnit TIMEOUT_UNIT = TimeUnit.SECONDS;
    private static final AtomicLong ID = new AtomicLong();

    private enum Content {

        domain("domain"),
        server("server");
        final String id;

        private Content(String id) {
            this.id = id;
        }
    }
    private final ServiceBus bus;
    private final ThreadPool pool;
    private final RemoteControl remoteControl;
    private final Domains domains;
    private final Map<String, CommandEvent> pending;

    public EventPing(Domains domains, ServiceBus bus, ThreadPool pool, RemoteControl remoteControl) {
        this.bus = bus;
        this.bus.subscribe(PingEvent.class, new PingResponse());

        this.domains = domains;
        this.pool = pool;
        this.remoteControl = remoteControl;
        this.pending = Collections.synchronizedMap(new HashMap<String, CommandEvent>());
    }

    //"op":"get","param":{"type":"ping","domain":"tskoda-book.local","server":"kerio.com"
    @Override
    public void event(CommandEvent event) {
        if (!event.interested(MESSAGE)) {
            return;
        }

        try {
            String type = event.get(TYPE);

            if (TYPE_PING.equals(type)) {
                String domain = event.get(Content.domain.id);
                String server = event.get(Content.server.id);
                
                if (this.domains.isLocalDomain(server)) {
                    LOG.log(Level.INFO, "Ping request to locally managed domain {0} ignored", server);
                    event.fail("Locally managed domain");
                } else {
                    request(domain, server, event);
                    event.async(); //make handling of request async
                }
            }

        } catch (Exception e) {
            LOG.log(Level.INFO, "Processing ping request failed.", e);
            event.fail(e.getMessage());
        }
    }

    private void request(String domain, String server, CommandEvent request) {

        PingEvent pingEvent = new PingEvent(PingEvent.Type.REQUEST, PingEvent.getPingId(ID.getAndIncrement()), domain, server);

        LOG.log(Level.INFO, "Ping request from {0} to {1} with id {2} has been sent", new Object[]{pingEvent.getSender(), pingEvent.getRecipient(), pingEvent.getId()});

        this.pool.submit(new PingTimeout(pingEvent.getId()), TIMEOUT, TIMEOUT_UNIT);
        this.bus.publish(pingEvent);

        //only if everything went well
        this.pending.put(pingEvent.getId(), request);
    }

    //called from watchdog thread
    private void responseTimeout(String eventID) {
        CommandEvent removed = this.pending.remove(eventID);
        if (removed != null) {
            LOG.log(Level.INFO, "Ping request with id {0} failed, request timeout", eventID);
            removed.fail("Ping request timeout");
            this.remoteControl.answer(removed);
        }
    }

    //called when published
    private void response(PingEvent response) {
        CommandEvent removed = this.pending.remove(response.getId());
        if (removed != null) {
            if (response.getType() == PingEvent.Type.RESPONSE_FAILED) {
                LOG.log(Level.INFO, "Ping request from {0} to {1} failed", new Object[]{response.getRecipient(), response.getSender()});
                removed.fail("Ping request failed");
            } else {
                LOG.log(Level.INFO, "Ping request from {0} to {1} successful", new Object[]{response.getRecipient(), response.getSender()});
            }
            this.remoteControl.answer(removed);
        }
    }

    private class PingTimeout implements Runnable {

        private final String id;

        public PingTimeout(String id) {
            this.id = id;
        }

        @Override
        public void run() {
            EventPing.this.responseTimeout(this.id);
        }
    }

    private class PingResponse implements ServiceBus.EventListener<PingEvent> {

        @Override
        public void event(PingEvent event) {
            if (event.getType() != PingEvent.Type.REQUEST) {
                EventPing.this.response(event);
            }
        }
    }
}
