/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import com.kerio.im.common.InetAddressUtil;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.external.ExtService;
import com.kerio.im.external.ExtService.AccessHandler;
import com.kerio.im.xmpp.api.XmppPorts;
import java.net.UnknownHostException;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class EventService implements ServiceBus.EventListener<CommandEvent> {

    private static final Logger LOG = Logger.getLogger(EventService.class.getName());
    public static final String MESSAGE = "service";

    public static class SecurityAccessHandlerImpl extends ServiceAccessImpl implements ExtService.SecurityAccessHandler {

        private boolean exceptions = Boolean.FALSE;
        private boolean tlsRequired = Boolean.TRUE;

        @Override
        public void exceptions(boolean hasExceptions) {
            this.exceptions = hasExceptions;
        }

        @Override
        public void required(boolean tlsRequired) {
            this.tlsRequired = tlsRequired;
        }

        public boolean hasExceptions() {
            return exceptions;
        }

        public boolean isTlsRequired() {
            return tlsRequired;
        }

    }

    public static class ServiceAccessImpl implements ExtService.AccessHandler {

        private final CombinedFilter serviceFilter;
        private final ConnectionLimit connectionLimit;

        public ServiceAccessImpl() {
            this.serviceFilter = CombinedFilter.without();
            this.connectionLimit = ConnectionLimit.without();
        }

        @Override
        public void filterUnknown() {
            this.serviceFilter.updateUnknown();

        }

        @Override
        public void filterRange(String from, String to, boolean enabled) {
            try {
                this.serviceFilter.update(new FilterRange(from, to), enabled);
            } catch (UnknownHostException e) {
                LOG.log(Level.WARNING, "Invalid ip address in ip group", e);
            }
        }

        @Override
        public void filterMask(String net, String mask, boolean enabled) {
            try {
                this.serviceFilter.update(FilterMask.create(net, mask), enabled);
            } catch (UnknownHostException e) {
                LOG.log(Level.WARNING, "Invalid ip address in ip group", e);
            }
        }

        @Override
        public void filterLocalhost(boolean enabled) {
            try {
                this.serviceFilter.update(new FilterLocalhost(), enabled);
            } catch (UnknownHostException e) {
                LOG.log(Level.WARNING, "Invalid ip address in ip group", e);
            }
        }

        @Override
        public void filterHost(String ip, boolean enabled) {
            try {
                this.serviceFilter.update(new FilterHost(ip), enabled);
            } catch (UnknownHostException e) {
                LOG.log(Level.WARNING, "Invalid ip address in ip group", e);
            }
        }

        @Override
        public void limit(int count) {
            if (count > 0) {
                this.connectionLimit.update(count);
            }
        }

        public CombinedFilter getServiceFilter() {
            return serviceFilter;
        }

        public ConnectionLimit getConnectionLimit() {
            return connectionLimit;
        }

    }

    public static class ServiceHandlerImpl implements ExtService.ServiceHandler {

        private final Set<ServicePort> ports;
        private final boolean ssl;
        private final String hostname;
        private final ServiceAccessImpl access;
        private final String serviceId;
        private final Security security;
        private final ConnectionListener listener;

        public ServiceHandlerImpl(String serviceId, String hostname, boolean ssl, Security security, final ConnectionListener listener) {
            this.ports = new HashSet<ServicePort>();
            this.ssl = ssl;
            this.hostname = hostname;
            this.serviceId = serviceId;
            this.access = new ServiceAccessImpl();
            this.security = security;
            this.listener = listener;
        }

        public Set<ServicePort> getPorts() {
            return ports;
        }

        @Override
        public void port(int port, ExtService.ServiceHandler.TYPE type, String address, String serviceName) {
            String ifc = null;
            switch (type) {
                case localhost:
                    //it can be either 127.0.0.1 or ::1
                case address:
                    ifc = address;
                    break;
                case all:
                    ifc = "*";
                    break;
                default:
                    LOG.log(Level.WARNING, "Detected unknown address type: {0}", type);
            }
            if (ifc != null) {
                ServicePort servicePort = new ServicePort(this.serviceId, port, ifc, this.hostname, this.ssl, security, serviceName, this.access.serviceFilter, this.access.connectionLimit, this.listener);
                this.ports.add(servicePort);
            }
        }

        @Override
        public AccessHandler getAccess() {
            return this.access;
        }

    }

    public enum Operation {

        start,
        stop,
        changed,
        security
    }

    public enum Content {

        identifier("service.id"),
        action("service.action"),
        secure("service.secure"),
        tlsRequired("service.tls"),
        tlsExceptions("service.tls.exception");

        public final String id;

        private Content(String id) {
            this.id = id;
        }
    }
    private final ExtService extService;
    private final ServiceProcessor serviceProcessor;
    private final String hostname;

    public EventService(ExtService extService, ServiceProcessor serviceProcessor, String hostname) {
        this.extService = extService;
        this.hostname = hostname;
        this.serviceProcessor = serviceProcessor;
    }

    @Override
    public void event(CommandEvent event) {
        if (!event.interested(MESSAGE)) {
            return;
        }

        try {
            Operation operation = Operation.valueOf(event.get(Content.action.id));
            switch (operation) {
                case start:
                    String startServiceId = event.get(Content.identifier.id);
                    LOG.log(Level.FINE, "Starting service {0}", startServiceId);
                    Boolean ssl = Boolean.valueOf(event.get(Content.secure.id, Boolean.FALSE.toString()));
                    Set<ServicePort> servicePorts = this.allServicePorts(startServiceId, this.hostname, ssl, ConnectionListener.external(this.extService));
                    if (servicePorts.isEmpty()) {
                        LOG.log(Level.WARNING, "Service {0} hasn't any configured port", startServiceId);
                    }
                    this.serviceProcessor.start(startServiceId, servicePorts, XmppPorts.Type.CLIENT);
                    LOG.log(Level.INFO, "Starting {0} service {1} successful", new Object[]{ssl ? "secure" : "plain", startServiceId});
                    break;
                case stop:
                    String stopServiceId = event.get(Content.identifier.id);
                    LOG.log(Level.FINE, "Stopping service {0}", stopServiceId);
                    this.serviceProcessor.stop(stopServiceId, XmppPorts.Type.CLIENT);
                    LOG.log(Level.INFO, "Stopping service {0} successful", stopServiceId);
                    break;
                case changed:
                    String changedServiceId = event.get(Content.identifier.id);
                    LOG.log(Level.FINE, "Updating service access settings {0}", changedServiceId);

                    ServiceAccessImpl accessImpl = new ServiceAccessImpl();
                    this.extService.getPortAccess(changedServiceId, accessImpl);
                    for (ServicePort port : this.serviceProcessor.ports(changedServiceId)) {
                        port.update(accessImpl.serviceFilter, accessImpl.connectionLimit);
                    }

                    LOG.log(Level.INFO, "Updating service access settings successful {0}", changedServiceId);
                    break;
                case security:
                    LOG.log(Level.FINE, "Updating plain service security settings");
                    //boolean tlsRequired = optional(TLS_REQUIRED, event.get(Content.tlsRequired.id), this.serviceProcessor.getSecurity().isRequireTLS());
                    SecurityAccessHandlerImpl access = new SecurityAccessHandlerImpl();
                    this.extService.getSecurity(access);

                    if (access.isTlsRequired()) {
                        this.serviceProcessor.getSecurity().setRequired(access.hasExceptions(), access.getServiceFilter());
                    } else {
                        this.serviceProcessor.getSecurity().setOptional();
                    }

                    LOG.log(Level.INFO, "Updating plain service security settings successful");
                    break;
                default:
                    LOG.log(Level.WARNING, "Processing service event failed, no suitable operation found {0}", operation);
                    event.fail("Processing service event failed, no suitable operation found");
            }
        } catch (Exception e) {
            LOG.log(Level.WARNING, "Processing service event failed", e);
            event.fail(e.getMessage());
        }
    }

    Set<ServicePort> allServicePorts(final String serviceId, final String hostname, final boolean ssl, final ConnectionListener listener) {
        ServiceHandlerImpl serviceHandler = new ServiceHandlerImpl(serviceId, hostname, ssl, this.serviceProcessor.getSecurity(), listener);
        this.extService.getAllPorts(serviceId, serviceHandler);
        return serviceHandler.getPorts();
    }

}
