/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.core.service;

import com.google.common.collect.Sets;
import com.kerio.im.core.api.service.ConfigurationService;
import com.kerio.im.core.api.service.SecurityService;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class SecurityManager implements SecurityService {

    private static final Logger LOG = Logger.getLogger(SecurityManager.class.getName());
    private static final Set<String> UNSAFE_PROTOCOL_SET = Collections.unmodifiableSet(Sets.newHashSet(
            "SSLv3"));

    private static final Set<String> SAFE_CIPHERS_SET = Collections.unmodifiableSet(Sets.newHashSet(
            "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384",
            "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384",
            "TLS_RSA_WITH_AES_256_CBC_SHA256",
            "TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA384",
            "TLS_ECDH_RSA_WITH_AES_256_CBC_SHA384",
            "TLS_DHE_RSA_WITH_AES_256_CBC_SHA256",
            "TLS_DHE_DSS_WITH_AES_256_CBC_SHA256",
            "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA",
            "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA",
            "TLS_RSA_WITH_AES_256_CBC_SHA",
            "TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA",
            "TLS_ECDH_RSA_WITH_AES_256_CBC_SHA",
            "TLS_DHE_RSA_WITH_AES_256_CBC_SHA",
            "TLS_DHE_DSS_WITH_AES_256_CBC_SHA",
            "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256",
            "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256",
            "TLS_RSA_WITH_AES_128_CBC_SHA256",
            "TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256",
            "TLS_ECDH_RSA_WITH_AES_128_CBC_SHA256",
            "TLS_DHE_RSA_WITH_AES_128_CBC_SHA256",
            "TLS_DHE_DSS_WITH_AES_128_CBC_SHA256",
            "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA",
            "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA",
            "TLS_RSA_WITH_AES_128_CBC_SHA",
            "TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA",
            "TLS_ECDH_RSA_WITH_AES_128_CBC_SHA",
            "TLS_DHE_RSA_WITH_AES_128_CBC_SHA",
            "TLS_DHE_DSS_WITH_AES_128_CBC_SHA",
            "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384",
            "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256",
            "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384",
            "TLS_RSA_WITH_AES_256_GCM_SHA384",
            "TLS_ECDH_ECDSA_WITH_AES_256_GCM_SHA384",
            "TLS_ECDH_RSA_WITH_AES_256_GCM_SHA384",
            "TLS_DHE_RSA_WITH_AES_256_GCM_SHA384",
            "TLS_DHE_DSS_WITH_AES_256_GCM_SHA384",
            "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
            "TLS_RSA_WITH_AES_128_GCM_SHA256",
            "TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256",
            "TLS_ECDH_RSA_WITH_AES_128_GCM_SHA256",
            "TLS_DHE_RSA_WITH_AES_128_GCM_SHA256",
            "TLS_DHE_DSS_WITH_AES_128_GCM_SHA256"));

    private final Map<String, String[]> supportedCiphersPerProtocol;
    private String[] supportedProtocols;
    private final Object supportedProtocolsLock;

    public SecurityManager() {
        //configuration allows us to control what is in the enabled ciphers list
        this.supportedCiphersPerProtocol = new ConcurrentHashMap<String, String[]>(4); //possibly TLS, TLSv1, TLSv2, + SSL
        this.supportedProtocolsLock = new Object();
    }

    /**
     * Filter ciphers of ssl/tls engine/socket instance. It is going to be
     * called very often, so we do the filtering once per protocol. 
     *
     * @param protocol (ssl, tls, tlsv1, ... )
     * @param defaultEnabledCiphers ciphers enabled by protocol provider
     * @return array of safe ciphers (no RC4, ... )
     */
    @Override
    public String[] filterCiphers(String protocol, String[] defaultEnabledCiphers) {
        String[] enabled = this.supportedCiphersPerProtocol.get(protocol);
        //double-checked locking on cuncurrent hash map
        if (enabled == null) {
            synchronized (this.supportedCiphersPerProtocol) {
                enabled = this.supportedCiphersPerProtocol.get(protocol);
                if (enabled == null) {
                    enabled = safeCiphers(defaultEnabledCiphers);
                    this.supportedCiphersPerProtocol.put(protocol, enabled);
                    if (LOG.isLoggable(Level.FINE)) {
                        String[] disabled = unsafeCiphers(defaultEnabledCiphers);
                        LOG.log(Level.FINE, "Enabling only strong ciphers (>= 128-bit key, !RC4, !EXPORT, !IDEA, ...) for {0} protocol successful.", protocol);
                        LOG.log(Level.FINE, "{0} enabled ciphers({1}): {2}", new Object[]{protocol, enabled.length, Arrays.toString(enabled)});
                        LOG.log(Level.FINE, "{0} disabled ciphers({1}): {2}", new Object[]{protocol, disabled.length, Arrays.toString(disabled)});
                    }
                }
            }
        }

        return enabled;
    }

    private String[] safeCiphers(String[] defaultEnabledCiphers) {
        Set<String> safe = new HashSet<String>(Arrays.asList(defaultEnabledCiphers));
        safe.retainAll(SAFE_CIPHERS_SET);

        return safe.toArray(new String[safe.size()]);
    }

    private String[] unsafeCiphers(String[] defaultEnabledCiphers) {

        Set<String> unsafe = new HashSet<String>(Arrays.asList(defaultEnabledCiphers));
        unsafe.removeAll(SAFE_CIPHERS_SET);

        return unsafe.toArray(new String[unsafe.size()]);
    }

    @Override
    public String[] filterProtocols(String[] defaultEnabledProtocols) {
        //double-checked locking 
        if ((this.supportedProtocols == null) && (defaultEnabledProtocols != null)) {
            synchronized (this.supportedProtocolsLock) {
                if (this.supportedProtocols == null) {
                    ArrayList<String> safeProtocols = new ArrayList<String>();
                    ArrayList<String> unsafeProtocols = new ArrayList<String>();
                    for (String protocol: defaultEnabledProtocols) {
                        if (UNSAFE_PROTOCOL_SET.contains(protocol)) {
                            unsafeProtocols.add(protocol);
                        } else {
                            safeProtocols.add(protocol);
                        }
                    }
                    
                    this.supportedProtocols = safeProtocols.toArray(new String[safeProtocols.size()]);
                    
                    if (LOG.isLoggable(Level.FINE)) {
                        LOG.log(Level.FINE, "Enabling only safe SSL protocols successful.");
                        LOG.log(Level.FINE, "Enabled protocols({0}): {1}", new Object[]{safeProtocols.size(), safeProtocols});
                        LOG.log(Level.FINE, "Disabled protocols({0}): {1}", new Object[]{unsafeProtocols.size(), unsafeProtocols});
                    }
                }
            }
        }

        return this.supportedProtocols;
    }
    
    
}
