/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import com.kerio.im.bean.Port;
import java.net.InetAddress;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class ServicePort implements Port {
    
    private static final Logger LOG = Logger.getLogger(ServicePort.class.getName());

    private final String id;
    private final int port;
    private final String ifc;
    private final String host;
    private final boolean ssl;
    private final CombinedFilter filter;
    private final ConnectionLimit limit;
    private final String serviceName;
    private final Security security;
    private final Listener listener;

    public ServicePort(String id, int port, String ifc, String host, boolean ssl, Security security, String serviceName, CombinedFilter filter, ConnectionLimit limit, ConnectionListener listener) {
        this.id = id;
        this.port = port;
        this.ifc = ifc;
        this.host = host;
        this.ssl = ssl;
        this.security = security;
        this.filter = filter;
        this.limit = limit;
        this.serviceName = serviceName;
        this.listener = listener;
    }

    public void update(CombinedFilter filter, ConnectionLimit limit) {
        this.filter.update(filter);
        this.limit.update(limit);
    }

    @Override
    public Security getSecurity() {
        return this.security;
    }

    @Override
    public int getPort() {
        return this.port;
    }

    @Override
    public String getIfc() {
        return this.ifc;
    }

    @Override
    public String getHost() {
        return this.host;
    }

    @Override
    public boolean isSSL() {
        return this.ssl;
    }

    @Override
    public String getServiceName() {
        return this.serviceName;
    }

    public String getId() {
        return id;
    }

    @Override
    public boolean accept(InetAddress client) {
        boolean result;
        if (!(result = this.limit.pass())) {

            LOG.log(Level.WARNING, "Connection attempt to service {0} from IP address {1} rejected: too many connections. Connection limit is {2}.", new Object[]{
                        this.serviceName,
                        client.getHostAddress(),
                        this.limit.getLimit()});
        } else if (!(result = this.filter.pass(client))) {
            LOG.log(Level.WARNING, "Connection attempt to service {0} from IP address {1} rejected: access from this IP address is not allowed.", new Object[]{
                        this.serviceName,
                        client.getHostAddress()});
        }
        return result;
    }

    @Override
    public boolean connected(String remoteAddress, int port) {
        this.limit.connected();
        return this.listener.connected(this.id, this.ssl, remoteAddress, port);
    }

    @Override
    public void disconnected(String remoteAddress, int port) {
        this.limit.disconnected();
        this.listener.disconnected(this.id, remoteAddress, port);
    }

    @Override
    public String toString() {
        return "Port{" + "port=" + port + ", ifc=" + ifc + ", host=" + host + ", secure=" + ssl + '}';
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 97 * hash + this.port;
        hash = 97 * hash + (this.ifc != null ? this.ifc.hashCode() : 0);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Port other = (Port) obj;
        if (this.port != other.getPort()) {
            return false;
        }
        if ((this.ifc == null) ? (other.getIfc() != null) : !this.ifc.equals(other.getIfc())) {
            return false;
        }
        return true;
    }
}
