/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import com.kerio.im.common.error.Code;
import com.kerio.im.external.ExtService;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.XmppPorts;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class ServiceProcessor {

    private static final Logger LOG = Logger.getLogger(ServiceProcessor.class.getName());

    public static Set<ServicePort> asSet(ServicePort port) {
        Set<ServicePort> result = new HashSet<ServicePort>(1);
        result.add(port);
        return result;
    }

    public enum Status {

        Running,
        Stopped,
        Error,
        Unknown
    }
    private final XmppPorts xmppPorts;
    private final Map<String, Set<ServicePort>> serviceToPort;
    private final Map<String, Status> status;
    private final Security security;

    public ServiceProcessor(XmppPorts xmppPorts) {
        this.xmppPorts = xmppPorts;
        this.serviceToPort = new HashMap<String, Set<ServicePort>>(3);
        this.security = Security.required(Boolean.FALSE, CombinedFilter.without());
        this.status = Collections.synchronizedMap(new HashMap<String, Status>());
    }

    protected void setExtService(ExtService extService) {
        EventService.SecurityAccessHandlerImpl access = new EventService.SecurityAccessHandlerImpl();
        extService.getSecurity(access);

        if (access.isTlsRequired()) {
            getSecurity().setRequired(access.hasExceptions(), access.getServiceFilter());
        } else {
            getSecurity().setOptional();
        }
    }

    protected Set<ServicePort> ports(String serviceId) {
        synchronized (this.serviceToPort) {
            Set<ServicePort> result;
            if (this.serviceToPort.containsKey(serviceId)) {
                result = this.serviceToPort.get(serviceId);
            } else {
                result = new HashSet<ServicePort>(1);
                this.serviceToPort.put(serviceId, result);
            }
            return result;
        }
    }

    public Security getSecurity() {
        return this.security;
    }

    private void error(String serviceId) {
        if (serviceId != null) {
            this.status.put(serviceId, Status.Error);
        }
    }

    private void started(String serviceId) {
        this.status.put(serviceId, Status.Running);
    }

    private void stopped(String serviceId) {
        this.status.put(serviceId, Status.Stopped);
    }

    public Status status(String serviceId) {
        Status result = this.status.get(serviceId);
        if (result == null) {
            result = Status.Unknown;
        }
        return result;
    }

    public Set<String> runningServices() {
        Set<String> result = new HashSet();
        synchronized (this.serviceToPort) {
            for (Map.Entry<String, Status> entry : this.status.entrySet()) {
                if (entry.getValue() == Status.Running) {
                    result.add(entry.getKey());
                }
            }
        }
        return result;
    }

    /**
     *
     * @param serviceId service identifier
     * @param currentPorts service ports
     * @param hostname server hostname
     * @param isSecure if service is secure
     * @throws BindException
     * @throws IOException
     */
    public void start(String serviceId, Set<ServicePort> newPorts, XmppPorts.Type type) {
        try {
            synchronized (this.serviceToPort) {
                Set<ServicePort> currentPorts = ports(serviceId);
                Set<ServicePort> portsToStop = new HashSet(currentPorts);
                portsToStop.removeAll(newPorts);

                Set<ServicePort> portsToStart = new HashSet(newPorts);
                portsToStart.removeAll(currentPorts);

                if (!portsToStop.isEmpty()) {
                    stopAll(portsToStop, currentPorts, type);
                }

                if (!portsToStart.isEmpty()) {
                    startAll(portsToStart, currentPorts, type);
                    started(serviceId);
                } else {
                    throw new XmppException("Service {0} failed to start, no configured ports found", Code.RESOURCE_ERROR, serviceId);
                }
                
                started(serviceId);
            }
        } catch (XmppException e) {
            error(serviceId);
            throw e;
        } catch (Exception e) {
            error(serviceId);
            throw new XmppException("Service {0} failed to start", Code.RESOURCE_ERROR, e, serviceId);
        }
    }

    public void stop(String serviceId, XmppPorts.Type type) {
        try {
            synchronized (this.serviceToPort) {
                Set<ServicePort> currentPorts = ports(serviceId);
                Set<ServicePort> portsToStop = new HashSet<ServicePort>(currentPorts);
                //portsToStop.remove(servicePort);
                stopAll(portsToStop, currentPorts, type); //stopping every port for received serviceId
                stopped(serviceId);
            }
        } catch (XmppException e) {
            error(serviceId);
            throw e;
        } catch (Exception e) {
            error(serviceId);
            throw new XmppException("Failed to start service {0}", Code.RESOURCE_ERROR, e, serviceId);
        }
    }

    public String getServiceId(int port) {
        String result = null;
        synchronized (this.serviceToPort) {
            for (Map.Entry<String, Set<ServicePort>> entry : this.serviceToPort.entrySet()) {
                for (ServicePort entryPort : entry.getValue()) {
                    if (entryPort.getPort() == port) {
                        result = entry.getKey();
                        break;
                    }
                }
            }
        }
        return result;
    }

    private void startAll(Set<ServicePort> portsToStart, Set<ServicePort> currentPorts, XmppPorts.Type type) {
        LOG.log(Level.FINER, "Starting ports {0}", portsToStart);
        for (ServicePort port : portsToStart) {
            this.xmppPorts.start(type, port);
            currentPorts.add(port);
        }
    }

    private void stopAll(Set<ServicePort> portsToStop, Set<ServicePort> currentPorts, XmppPorts.Type type) {
        LOG.log(Level.FINER, "Stopping ports {0}", portsToStop);
        for (ServicePort port : portsToStop) {
            this.xmppPorts.stop(type, port);
            currentPorts.remove(port);
        }
    }
}
