/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.user;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.remote.CommandEvent;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class EventUserSync implements ServiceBus.EventListener<CommandEvent> {

    private static final Logger LOG = Logger.getLogger(EventUserSync.class.getName());
    private static final String COMMAND = "user";
    private static final String ENABLED_VALUE = "1";
    private static final String VISIBLE_VALUE = "1";
    private static final String IS_ALLOWED_SEND_OUTSIDE = "0";

    public enum Content {

        fullname("user.fullname"),
        event("user.event"),
        name("user.name"),
        domain("user.domain"),
        enabled("user.enabled"),
        gal("user.gal"),
        hasDomainRestriction("user.hasDomainRestriction");
        
        /**
         * Identifier of message content.
         */
        final String id;

        Content(String id) {
            this.id = id;
        }
    }

    public enum Operation {

        created,
        updated,
        removed,
        repair
        
    }
    private final UserControlList userControlList;

    public EventUserSync(UserControlList userControlList) {
        this.userControlList = userControlList;
    }

    @Override
    public void event(CommandEvent event) {
        if (!event.interested(COMMAND)) {
            return;
        }
        try {
            Operation operation = Operation.valueOf(event.get(Content.event.id));
            String userName = event.get(Content.name.id).toLowerCase(); //required
            String domain = event.get(Content.domain.id).toLowerCase(); //required
            UserControl userControl = this.userControlList.get(domain);
            if (userControl != null) {
                switch (operation) {
                    case created:
                        User createdUser = new User(
                                userName,
                                domain,
                                event.get(Content.fullname.id, userName), //optional
                                ENABLED_VALUE.equals(event.get(Content.enabled.id)), //required
                                VISIBLE_VALUE.equals(event.get(Content.gal.id)), //required
                                IS_ALLOWED_SEND_OUTSIDE.equals(event.get(Content.hasDomainRestriction.id))); //required
                        userControl.add(createdUser);
                        LOG.log(Level.INFO, "Processing user event successful, user {0} created", userName);
                        break;
                    case removed:
                        userControl.remove(userName, domain);
                        LOG.log(Level.INFO, "Processing user event successful, user {0} removed", userName);
                        break;
                    case updated:
                        userControl.update(
                                userName,
                                domain,
                                event.get(Content.fullname.id, null), //optional
                                asBool(event.get(Content.enabled.id, null), ENABLED_VALUE), //optional
                                asBool(event.get(Content.gal.id, null), VISIBLE_VALUE), //optional
                                asBool(event.get(Content.hasDomainRestriction.id, null), IS_ALLOWED_SEND_OUTSIDE)); //optional
                        LOG.log(Level.INFO, "Processing user event successful, user {0} updated", userName);
                        break;
                    case repair:
                        userControl.repair(userName, domain);
                        LOG.log(Level.INFO, "Processing user event successful, user {0} repaired", userName);
                        break;
                    default:
                        LOG.log(Level.WARNING, "Processing user event failed, unsupported operation {0}", operation);
                        event.fail("Processing user event failed, unsupported operation");
                }
            } else {
                LOG.log(Level.WARNING, "Processing user event failed, domain {0} not in sync", operation);
            }
        } catch (Exception e) {
            LOG.log(Level.WARNING, "Processing user event failed", e);
            event.fail("Processing user event failed");
        }
    }
    
    
    private Boolean asBool(String intValue, String trueValue) {
        Boolean result = null;
        if (intValue != null) {
            return trueValue.equals(intValue);
        }
        return result;
    }
}
