/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.core.cert;

import com.kerio.im.bean.Certificate;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;
import org.mockito.Matchers;

/**
 *
 * @author tskoda
 */
public class CertificateStoreTest {

    public CertificateStoreTest() {
    }

    @BeforeClass
    public static void setUpClass() {
    }

    @AfterClass
    public static void tearDownClass() {
    }

    @Before
    public void setUp() {
    }

    @After
    public void tearDown() {
    }

    /**
     * Test of put method, of class CertificateStore.
     */
    @Test
    public void testPut() {
        CertificateStore store = new CertificateStore("default");

        assertTrue(store.map().isEmpty());

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));

        store.put(cert, aliases, false);
        assertThat(store.map().size(), is(1));
        assertThat(store.map().get("alias1"), is(cert));

        store.put(cert, aliases, false);
        assertThat(store.map().size(), is(1));
        assertThat(store.map().get("alias1"), is(cert));

    }

    @Test
    public void testPutEmptyAlias() {
        CertificateStore store = new CertificateStore("default");

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>();

        store.put(cert, aliases, false);
        assertTrue(store.map().isEmpty());

    }

    @Test
    public void testPutMoreAlias() {
        CertificateStore store = new CertificateStore("default");

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1", "alias2", "alias3"));

        store.put(cert, aliases, false);
        assertThat(store.map().size(), is(3));
        assertThat(store.map().get("alias1"), is(cert));
        assertThat(store.map().get("alias2"), is(cert));
        assertThat(store.map().get("alias3"), is(cert));

    }

    @Test
    public void testPutMore() {
        CertificateStore store = new CertificateStore("default");

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));

        Certificate cert2 = new Certificate("cert1", true, null, null);
        Set<String> aliases2 = new HashSet<>(Arrays.asList("alias2"));

        store.put(cert, aliases, false);
        store.put(cert2, aliases2, false);

        assertThat(store.map().size(), is(2));
        assertThat(store.map().get("alias1"), is(cert));
        assertThat(store.map().get("alias2"), is(cert2));

    }

    @Test
    public void testPutDuplicateInvalid() {

        Certificate certInvalid = new Certificate("cert1", false, null, null);
        Set<String> aliasesInvalid = new HashSet<>(Arrays.asList("alias1", "alias2"));

        Certificate certValid = new Certificate("cert1", true, null, null);
        Set<String> aliasesValid = new HashSet<>(Arrays.asList("alias1", "alias3"));

        {
            CertificateStore store = new CertificateStore("default");

            store.put(certInvalid, aliasesInvalid, false);
            store.put(certValid, aliasesValid, false);

            assertThat(store.map().size(), is(3));
            assertThat(store.map().get("alias1"), is(certValid));
            assertThat(store.map().get("alias2"), is(certInvalid));
            assertThat(store.map().get("alias3"), is(certValid));
        }

        {
            CertificateStore store = new CertificateStore("default");

            store.put(certValid, aliasesValid, false);
            store.put(certInvalid, aliasesInvalid, false);

            assertThat(store.map().size(), is(3));
            assertThat(store.map().get("alias1"), is(certValid));
            assertThat(store.map().get("alias2"), is(certInvalid));
            assertThat(store.map().get("alias3"), is(certValid));
        }

    }

    @Test
    public void testPutDefault() {
        CertificateStore store = new CertificateStore("default");

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));

        store.put(cert, aliases, true);
        assertThat(store.map().size(), is(2));
        assertThat(store.map().get("alias1"), is(cert));
        assertThat(store.map().get("default"), is(cert));
    }

    /**
     * Test of removeById method, of class CertificateStore.
     */
    @Test
    public void testRemoveByIdNotExists() {
        CertificateStore store = new CertificateStore("default");

        Map<String, Certificate> result = store.removeById("identifier");
        assertTrue(result.isEmpty());

    }

    @Test
    public void testRemoveById() {
        CertificateStore store = new CertificateStore("default");

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1", "alias2"));
        Certificate cert2 = new Certificate("cert2", true, null, null);
        Set<String> aliases2 = new HashSet<>(Arrays.asList("alias3", "alias4"));

        store.put(cert, aliases, false);
        store.put(cert2, aliases2, false);

        assertThat(store.map().size(), is(4));

        Map<String, Certificate> result = store.removeById("cert2");

        assertThat(store.map().size(), is(2));

        assertThat(result.size(), is(2));
        assertTrue(result.containsKey("alias3"));
        assertNull(result.get("alias3"));
        assertTrue(result.containsKey("alias4"));
        assertNull(result.get("alias4"));

    }

    @Test
    public void testRemoveByIdDefault() {
        CertificateStore store = new CertificateStore("default");

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1", "alias2"));
        Certificate cert2 = new Certificate("cert2", true, null, null);
        Set<String> aliases2 = new HashSet<>(Arrays.asList("alias3", "alias4"));

        store.put(cert, aliases, false);
        store.put(cert2, aliases2, true);

        assertThat(store.map().size(), is(5)); //+1 default

        Map<String, Certificate> result = store.removeById("cert2");

        assertThat(store.map().size(), is(2));

        assertThat(result.size(), is(3));
        assertTrue(result.containsKey("alias3"));
        assertNull(result.get("alias3"));
        assertTrue(result.containsKey("alias4"));
        assertNull(result.get("alias4"));
        assertTrue(result.containsKey("default"));
        assertNull(result.get("default"));

    }

    /**
     * Test of add method, of class CertificateStore.
     */
    @Test
    public void testAdd() {

        CertificateStore store = new CertificateStore("default");

        assertTrue(store.map().isEmpty());

        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));

        Map<String, Certificate> addResult = store.add(cert, aliases);
        assertThat(store.map().size(), is(1));
        assertThat(store.map().get("alias1"), is(cert));
        assertThat(addResult.size(), is(1));
        assertThat(addResult.get("alias1"), is(cert));

    }

    @Test
    public void testAddAlreadyExists() {

        CertificateStore store = new CertificateStore("default");
        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));

        store.put(cert, aliases, true);

        assertThat(store.map().size(), is(2));

        Certificate addCertInvalid = new Certificate("cert2", false, null, null);

        {
            Map<String, Certificate> addResult = store.add(addCertInvalid, aliases);

            assertThat(store.map().size(), is(2));
            assertThat(store.map().get("alias1"), is(cert));
            assertTrue(addResult.isEmpty());
        }

        Certificate addCertValid = new Certificate("cert2", true, null, null);
        {
            Map<String, Certificate> addResult = store.add(addCertValid, aliases);

            assertThat(store.map().size(), is(2));
            assertThat(store.map().get("alias1"), is(addCertValid));
            assertThat(addResult.size(), is(1));
            assertThat(addResult.get("alias1"), is(addCertValid));
        }

        //after all default is not affected
        assertThat(store.map().get("default"), is(cert));
    }

    /**
     * Test of updateDefault method, of class CertificateStore.
     */
    @Test
    public void testUpdateDefaultNotExists() {
        CertificateStore store = new CertificateStore("default");
        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));

        assertTrue(store.map().isEmpty());

        Map<String, Certificate> updateDefaultResult = store.updateDefault("not exist");
        assertTrue(updateDefaultResult.isEmpty());
        assertTrue(store.map().isEmpty());

    }

    @Test
    public void testUpdateDefaultNotSet() {
        CertificateStore store = new CertificateStore("default");
        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));
        
        store.put(cert, aliases, false);

        assertThat(store.map().size(), is(1)); //no default set
        assertNull(store.map().get("default"));
        
        Map<String, Certificate> updateDefaultResult = store.updateDefault("cert1");
        
        assertThat(store.map().size(), is(2)); //no default set
        assertThat(store.map().get("default"), is(cert));
        
        assertThat(updateDefaultResult.size(), is(1)); //no default set
        assertThat(updateDefaultResult.get("default"), is(cert));

    }
    
    @Test
    public void testUpdateDefault() {
        CertificateStore store = new CertificateStore("default");
        Certificate cert = new Certificate("cert1", true, null, null);
        Set<String> aliases = new HashSet<>(Arrays.asList("alias1"));        
        store.put(cert, aliases, true);

        Certificate certNewDefault = new Certificate("certNewDefault", true, null, null);
        Set<String> aliasesNewDefault = new HashSet<>(Arrays.asList("alias2"));
        store.put(certNewDefault, aliasesNewDefault, false);
        
        assertThat(store.map().size(), is(3)); 
        assertThat(store.map().get("default"), is(cert));
        
        Map<String, Certificate> updateDefaultResult = store.updateDefault("certNewDefault");
        
        assertThat(store.map().size(), is(3)); 
        assertThat(store.map().get("default"), is(certNewDefault));
        
        assertThat(updateDefaultResult.size(), is(1)); 
        assertThat(updateDefaultResult.get("default"), is(certNewDefault));

    }

}
