/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import com.kerio.im.bean.User;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;

/**
 *
 * @author tskoda
 */
public class ArchiveBufferSchemaTest {

    private MemoryDerby derbyDB;

    @Before
    public void setUp() {
        this.derbyDB = new MemoryDerby();
    }

    @After
    public void tearDown() {
        this.derbyDB.destroy();
    }

    private int addParticipant(Connection db, String participant) throws SQLException {
        PreparedStatement addParticipant = db.prepareStatement(ArchiveBufferSchema.SQL.addParticipant.sql, Statement.RETURN_GENERATED_KEYS);
        try {
            addParticipant.setString(1, participant);
            addParticipant.execute();
            ResultSet generatedKeys = addParticipant.getGeneratedKeys();
            try {
                generatedKeys.next();
                return generatedKeys.getInt(1);
            } finally {
                TempDerby.close(generatedKeys);
            }
        } finally {
            TempDerby.close(addParticipant);
        }
    }

    private Map<Integer, String> getParticipant(Connection db, String participant1, String participant2) throws SQLException {
        PreparedStatement getParticipant = db.prepareStatement(ArchiveBufferSchema.SQL.getParticipants.sql);
        Map<Integer, String> result = new HashMap<Integer, String>(2);
        try {
            if (participant1 != null) {
                getParticipant.setString(1, participant1);
            } else {
                getParticipant.setNull(1, Types.VARCHAR);
            }

            if (participant2 != null) {
                getParticipant.setString(2, participant2);
            } else {
                getParticipant.setNull(2, Types.VARCHAR);
            }

            ResultSet rs = getParticipant.executeQuery();
            try {
                while (rs.next()) {
                    result.put(rs.getInt("id"), rs.getString("jid"));
                }
            } finally {
                TempDerby.close(rs);
            }
        } finally {
            TempDerby.close(getParticipant);
        }
        return result;
    }

    private int addRecord(Connection db, int sender, int recipient, String text) throws SQLException {
        PreparedStatement storeRecord = db.prepareStatement(ArchiveBufferSchema.SQL.storeRecord.sql);
        try {
            storeRecord.setInt(1, sender);
            storeRecord.setInt(2, recipient);
            storeRecord.setInt(3, ArchiveBufferSchema.RECORD_TYPE.message.toInt());
            storeRecord.setString(4, text);
            storeRecord.setNull(5, Types.VARCHAR);
            storeRecord.setNull(6, Types.BIGINT);
            return storeRecord.executeUpdate();
        } finally {
            TempDerby.close(storeRecord);
        }
    }

    private int addRecord(Connection db, int sender, int recipient, String fileName, long size) throws SQLException {
        PreparedStatement storeRecord = db.prepareStatement(ArchiveBufferSchema.SQL.storeRecord.sql);
        try {
            storeRecord.setInt(1, sender);
            storeRecord.setInt(2, recipient);
            storeRecord.setInt(3, ArchiveBufferSchema.RECORD_TYPE.file.toInt());
            storeRecord.setNull(4, Types.VARCHAR);
            storeRecord.setString(5, fileName);
            storeRecord.setLong(6, size);
            return storeRecord.executeUpdate();
        } finally {
            TempDerby.close(storeRecord);
        }
    }

    private void applySchema(String dbName) throws SQLException {
        Connection db = this.derbyDB.create(dbName);
        DerbyConnectionFactory cf = mock(DerbyConnectionFactory.class);
        when(cf.newConnection()).thenReturn(db);
        ArchiveBufferSchema.apply(cf);
    }

    @Test
    public void testApplyClose() throws Exception {
        Connection inMemoryDB = mock(Connection.class);
        Statement statement = mock(Statement.class);
        when(inMemoryDB.createStatement()).thenReturn(statement);

        DerbyConnectionFactory cf = mock(DerbyConnectionFactory.class);
        when(cf.newConnection()).thenReturn(inMemoryDB);

        ArchiveBufferSchema.apply(cf);

        verify(inMemoryDB, times(1)).close();
    }

    @Test
    public void testSQLSingleParticipant() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant"), is(1));
            Map<Integer, String> result = getParticipant(db, "test-participant", "test-participant-non-existent");
            assertThat(result.size(), is(1));
            assertThat(result.get(1), is("test-participant"));
        } finally {
            TempDerby.close(db);
        }
    }

    @Test(expected = SQLException.class)
    public void testSQLSingleParticipantTwice() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant"), is(1));
            addParticipant(db, "test-participant"); //should throw exception
            fail();
        } finally {
            TempDerby.close(db);
        }
    }

    @Test
    public void testSQLTwoParticipants() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            assertThat(addParticipant(db, "test-participant-2"), is(2));
            Map<Integer, String> result = getParticipant(db, "test-participant-1", "test-participant-2");
            assertThat(result.size(), is(2));
            assertThat(result.get(1), is("test-participant-1"));
            assertThat(result.get(2), is("test-participant-2"));

        } finally {
            TempDerby.close(db);
        }
    }

    @Test
    public void testSQLMaxParticipantLength() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            char[] longName = new char[ArchiveBufferSchema.LIMITS.JID.limit];
            Arrays.fill(longName, 'a');
            assertThat(addParticipant(db, new String(longName)), is(1));
        } finally {
            TempDerby.close(db);
        }
    }

    @Test(expected = SQLException.class)
    public void testSQLRecordRecipientDoesNotExist() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            addRecord(db, 0, 1, "text");
            fail();
        } finally {
            TempDerby.close(db);
        }
    }

    @Test(expected = SQLException.class)
    public void testSQLRecordSenderDoesNotExist() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            addRecord(db, 1, 0, "text");
            fail();
        } finally {
            TempDerby.close(db);
        }
    }

    @Test
    public void testSQLRecord() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            assertThat(addParticipant(db, "test-participant-2"), is(2));
            assertThat(addRecord(db, 1, 2, "text"), is(1));
            assertThat(addRecord(db, 2, 1, "filename", 5l), is(1));
        } finally {
            TempDerby.close(db);
        }
    }

    @Test
    public void testSQLRecordMaxTextLength() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            char[] longText = new char[ArchiveBufferSchema.LIMITS.TEXT.limit];
            Arrays.fill(longText, 'a');
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            assertThat(addRecord(db, 1, 1, new String(longText)), is(1));
        } finally {
            TempDerby.close(db);
        }
    }

    @Test
    public void testSQLRecordMaxFileSize() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            char[] longFilename = new char[ArchiveBufferSchema.LIMITS.FILENAME.limit];
            Arrays.fill(longFilename, 'a');
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            assertThat(addRecord(db, 1, 1, new String(longFilename), 10), is(1));
        } finally {
            TempDerby.close(db);
        }
    }

    @Test
    public void testSQLRecordMaxFileNameSize() throws SQLException {
        applySchema("test-archive-buffer");
        Connection db = this.derbyDB.create("test-archive-buffer");
        try {
            assertThat(addParticipant(db, "test-participant-1"), is(1));
            assertThat(addRecord(db, 1, 1, "filename", Long.MAX_VALUE), is(1));
        } finally {
            TempDerby.close(db);
        }
    }
}
