/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import java.io.File;
import java.io.IOException;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

/**
 *
 * @author tskoda
 */
public class DatabaseDestinationTest {

    private TempFolder parentTmp;
    private TempDerby derbyTmp;

    public DatabaseDestinationTest() {
    }

    @BeforeClass
    public static void setUpClass() {
    }

    @AfterClass
    public static void tearDownClass() {
    }

    @Before
    public void setUp() throws IOException {
        this.parentTmp = new TempFolder();
        this.derbyTmp = new TempDerby(this.parentTmp);
    }

    @After
    public void tearDown() {
        this.derbyTmp.destroy();
        this.parentTmp.destroy();
    }

    @Test
    public void testOpenNotExist() throws Exception {
        String dbName = "i_dont_exist";
        File dbFile = this.parentTmp.get(dbName);

        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(dbFile.exists(), is(false));

        db.open();

        assertThat(db.toPath(), is(dbFile.getAbsolutePath()));
        assertThat(db.exists(), is(false));
        assertThat(dbFile.exists(), is(false));
    }

    @Test(expected = DerbyDestination.DestinationExistException.class)
    public void testOpenExistFile() throws Exception {
        String dbName = "im_file";
        File dbFile = this.parentTmp.createFile(dbName);
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);
        assertThat(dbFile.isFile(), is(true));

        db.open(); //throws exception
    }

    @Test(expected = DerbyDestination.DestinationExistException.class)
    public void testOpenExistFolderEmpty() throws Exception {
        String dbName = "im_empty_folder";
        File dbFile = this.parentTmp.createFolder(dbName);
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(dbFile.isDirectory(), is(true));

        db.open(); //throws exception
    }

    @Test(expected = DerbyDestination.DestinationExistException.class)
    public void testOpenExistFolderNonEmpty() throws Exception {
        String dbName = "im_non_empty_folder";
        File dbFile = this.parentTmp.createFolder(dbName);
        File dbFileContent = new File(dbFile, "content");
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(dbFile.isDirectory(), is(true));
        assertThat(this.parentTmp.createFileCustom(dbFileContent), is(true));

        db.open();
    }

    @Test
    public void testOpenExistDerby() throws Exception {
        String dbName = "im_derby";
        File derbyFile = this.derbyTmp.create(dbName);
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(derbyFile.isDirectory(), is(true));
        assertThat(new File(derbyFile, "seg0").isDirectory(), is(true));
        assertThat(new File(derbyFile, "log").isDirectory(), is(true));
        assertThat(new File(derbyFile, "service.properties").isFile(), is(true));

        db.open();

        assertThat(db.exists(), is(true));
        assertThat(db.toPath(), is(derbyFile.getAbsolutePath()));
    }

    @Test(expected = DerbyDestination.DestinationExistException.class)
    public void testOpenExistDerbyIncomplete() throws Exception {
        String dbName = "im_derby_incomplete";
        File derbyFile = this.derbyTmp.create(dbName);
        assertThat(new File(derbyFile, "seg0").isDirectory(), is(true));
        assertThat(new File(derbyFile, "log").isDirectory(), is(true));
        assertThat(new File(derbyFile, "service.properties").isFile(), is(true));
        assertThat(new File(derbyFile, "service.properties").delete(), is(true));

        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);
        db.open();
    }

    @Test
    public void testRepairNotExist() throws Exception {
        String dbName = "i_dont_exist";
        File dbFile = this.parentTmp.get(dbName);
        File dbFileBackup = this.parentTmp.get(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START));
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);
        db.open();

        assertThat(dbFileBackup.exists(), is(false));
        assertThat(dbFile.exists(), is(false));

        db.repair();

        assertThat(dbFileBackup.exists(), is(false));
        assertThat(dbFile.exists(), is(false));
        assertThat(db.exists(), is(false));
    }

    @Test
    public void testRepairExistFile() throws Exception {
        String dbName = "im_file";
        File dbFile = this.parentTmp.createFile(dbName);
        File dbFileBackup = this.parentTmp.get(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START));
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);
        try {
            db.open();
            fail();
        } catch (DerbyDestination.DestinationExistException e) {
            //expected, lets test repair
        }

        assertThat(dbFileBackup.exists(), is(false));
        assertThat(dbFile.isFile(), is(true));

        db.repair();

        assertThat(dbFileBackup.isFile(), is(true));
        assertThat(dbFile.exists(), is(false));
        assertThat(db.exists(), is(false));
    }

    @Test
    public void testRepairExistFolderEmpty() throws Exception {
        String dbName = "im_empty_folder";
        File dbFile = this.parentTmp.createFolder(dbName);
        File dbFileBackup = this.parentTmp.get(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START));
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);
        try {
            db.open();
            fail();
        } catch (DerbyDestination.DestinationExistException e) {
            //expected, lets test repair
        }

        assertThat(dbFile.isDirectory(), is(true));
        assertThat(dbFileBackup.exists(), is(false));

        db.repair();

        assertThat(dbFile.exists(), is(false));
        assertThat(dbFileBackup.exists(), is(false));
        assertThat(db.exists(), is(false));
    }

    @Test
    public void testRepairExistFolderNonEmpty() throws Exception {
        String dbName = "im_non_empty_folder";
        File dbFile = this.parentTmp.createFolder(dbName);
        File dbFileContent = new File(dbFile, "content");
        File dbFileBackup = this.parentTmp.get(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START));
        File dbFileContentBackup = new File(dbFileBackup, "content");

        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(dbFile.isDirectory(), is(true));
        assertThat(this.parentTmp.createFileCustom(dbFileContent), is(true));
        assertThat(dbFileBackup.exists(), is(false));

        try {
            db.open();
            fail();
        } catch (DerbyDestination.DestinationExistException e) {
            //expected
        }

        db.repair();

        assertThat(dbFile.exists(), is(false));
        assertThat(dbFileBackup.isDirectory(), is(true));
        assertThat(dbFileContentBackup.isFile(), is(true));
        assertThat(db.exists(), is(false));
    }
//

    @Test
    public void testRepairDerby() throws Exception {
        String dbName = "im_derby";
        File derbyFile = this.derbyTmp.create(dbName);
        File derbyFileBackup = this.parentTmp.get(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START));
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);
        db.open(); //should not throw exception

        assertThat(derbyFileBackup.exists(), is(false));
        assertThat(derbyFile.isDirectory(), is(true));
        assertThat(new File(derbyFile, "seg0").isDirectory(), is(true));
        assertThat(new File(derbyFile, "log").isDirectory(), is(true));
        assertThat(new File(derbyFile, "service.properties").isFile(), is(true));

        db.repair();

        assertThat(db.exists(), is(false));
        assertThat(derbyFile.exists(), is(false));
        assertThat(derbyFileBackup.isDirectory(), is(true));
        assertThat(new File(derbyFileBackup, "seg0").isDirectory(), is(true));
        assertThat(new File(derbyFileBackup, "log").isDirectory(), is(true));
        assertThat(new File(derbyFileBackup, "service.properties").isFile(), is(true));
    }

    @Test
    public void testRepairIncrement() throws IOException {
        String dbName = "increment";
        File dbFile = this.parentTmp.createFile(dbName);
        File dbFileFirst = this.parentTmp.createFile(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START));
        File dbFileSecond = this.parentTmp.createFile(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START + 1));
        File expected = this.parentTmp.get(String.format(DerbyDestination.RENAME_FORMAT, dbName, DerbyDestination.RENAME_SEQUENCE_START + 2));
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(dbFile.isFile(), is(true));
        assertThat(dbFileFirst.isFile(), is(true));
        assertThat(dbFileSecond.isFile(), is(true));
        assertThat(expected.exists(), is(false));

        try {
            db.open(); //should not throw exception
            fail();
        } catch (DerbyDestination.DestinationExistException e) {
            //expected
        }

        db.repair();

        assertThat(dbFile.exists(), is(false));
        assertThat(db.exists(), is(false));
        assertThat(dbFileFirst.isFile(), is(true));
        assertThat(dbFileSecond.isFile(), is(true));
        assertThat(expected.isFile(), is(true));
    }

    @Test(expected = IOException.class)
    public void testExistsNotOpenned() throws Exception {
        String dbName = "custom";
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        db.exists(); //should throw an exception 
    }

    @Test(expected = IOException.class)
    public void testRepairNotOpenned() throws Exception {
        String dbName = "custom";
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        db.repair(); //should throw an exception
    }

    @Test
    public void testToPathNotOpenned() throws Exception {
        String dbName = "custom";
        File dbFile = new File(this.parentTmp.raw(), dbName);
        DerbyDestination db = new DerbyDestination(this.parentTmp.raw(), dbName);

        assertThat(db.toPath(), is(dbFile.getAbsolutePath()));
    }

}
