/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import com.kerio.im.core.api.service.MessageService;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.remote.CommandEvent;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.hamcrest.CoreMatchers;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import static org.mockito.Mockito.*;

/**
 *
 * @author tskoda
 */
public class EventGroupTest {

    private ServiceBus bus;
    private ServiceProcessor processor;
    private EventGroup group;
    private Map<String, String> response;

    public EventGroupTest() {
    }

    @Before
    public void before() {

        this.bus = mock(ServiceBus.class);
        this.processor = mock(ServiceProcessor.class);
        this.group = new EventGroup(this.bus, this.processor);
        this.response = new HashMap<String, String>();
    }

    private CommandEvent command(String type, Map<String, String> content) {
        return new CommandEvent("1", type, content, this.response, mock(MessageService.ResponseHandler.class));
    }

    private Map<String, String> map(String... pairs) {
        Map<String, String> result = new HashMap<String, String>();
        for (int i = 0; i + 1 < pairs.length; i = i + 2) {
            result.put(pairs[i], pairs[i + 1]);
        }
        return result;
    }

    @Test(expected = NullPointerException.class)
    public void testEventNull() {
        this.group.event(null);
    }

    @Test
    public void testEventUnknown() {
        CommandEvent event = command("unknown", new HashMap<String, String>());
        assertThat(event.getType(), CoreMatchers.is(CoreMatchers.not(EventGroup.MESSAGE)));
        this.group.event(event);

        verifyZeroInteractions(this.bus, this.processor);

    }

    @Test
    public void testEventMissingEvent() {
        CommandEvent event = command(EventGroup.MESSAGE, map("unknownElement", "unknownValue"));
        this.group.event(event);

        assertTrue(event.isFail());
        verifyZeroInteractions(this.bus, this.processor);
    }

    @Test
    public void testEventUnknownEvent() {
        CommandEvent event = command(EventGroup.MESSAGE, map(EventGroup.Content.event.id, "unknownValue"));
        this.group.event(event);

        assertTrue(event.isFail());
        verifyZeroInteractions(this.bus, this.processor);
    }

    @Test
    public void testEventChangedNoRunningService() {
        CommandEvent event = command(EventGroup.MESSAGE, map(EventGroup.Content.event.id, EventGroup.Event.changed.name()));
        when(this.processor.runningServices()).thenReturn(Collections.EMPTY_SET);

        this.group.event(event);

        assertFalse(event.isFail());
        ArgumentCaptor<CommandEvent> eventCaptor = ArgumentCaptor.forClass(CommandEvent.class);
        verify(this.bus, times(1)).publish(eventCaptor.capture());

        assertThat(eventCaptor.getValue().getId(), CoreMatchers.is(event.getId()));
        assertThat(eventCaptor.getValue().getType(), CoreMatchers.is(EventService.MESSAGE));
        assertThat(eventCaptor.getValue().get(EventService.Content.action.id), CoreMatchers.is(EventService.Operation.security.toString()));

    }

    @Test
    public void testEventChangedWithRunningService() {
        CommandEvent event = command(EventGroup.MESSAGE, map(EventGroup.Content.event.id, EventGroup.Event.changed.name()));
        Set<String> runningServices = new LinkedHashSet<String>(Arrays.asList("service1", "service2", "service3"));
        when(this.processor.runningServices()).thenReturn(runningServices);

        this.group.event(event);

        assertFalse(event.isFail());
        ArgumentCaptor<CommandEvent> eventCaptor = ArgumentCaptor.forClass(CommandEvent.class);
        verify(this.bus, times(4)).publish(eventCaptor.capture());
        
        List<CommandEvent> result = eventCaptor.getAllValues();
        
        //1. - 3. event: service change for each provided service
        assertThat(result.get(0).get(EventService.Content.action.id), CoreMatchers.is(EventService.Operation.changed.toString()));
        assertThat(result.get(0).get(EventService.Content.identifier.id), CoreMatchers.is("service1"));
        
        assertThat(result.get(1).get(EventService.Content.action.id), CoreMatchers.is(EventService.Operation.changed.toString()));
        assertThat(result.get(1).get(EventService.Content.identifier.id), CoreMatchers.is("service2"));
        
        assertThat(result.get(2).get(EventService.Content.action.id), CoreMatchers.is(EventService.Operation.changed.toString()));
        assertThat(result.get(2).get(EventService.Content.identifier.id), CoreMatchers.is("service3"));
        
        //4. event: service security changed
        assertThat(result.get(3).get(EventService.Content.action.id), CoreMatchers.is(EventService.Operation.security.toString()));

    }
}
