/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.user;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.service.AccessPolicyService;
import com.kerio.im.core.api.service.PacketType;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.configuration.StoreConfiguration;
import com.kerio.im.external.ExtService;
import com.kerio.im.external.ExtService.Settings;
import java.util.EnumSet;
import org.hamcrest.CoreMatchers;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

/**
 *
 * @author tskoda
 */
public class UserAccessPolicyTest {

    public static final String CONTROL_LOGIN = "remoteLogin";
    public static final String CONTROL_DOMAIN = "remoteDomain";
    StoreConfiguration conf;

    public UserAccessPolicyTest() {
    }

    @Before
    public void before() {
        this.conf = mock(StoreConfiguration.class);
        when(this.conf.get(GlobalKey.SERVICE_REMOTE_LOGIN)).thenReturn(CONTROL_LOGIN);
        when(this.conf.get(GlobalKey.SERVICE_REMOTE_DOMAIN)).thenReturn(CONTROL_DOMAIN);

    }

    /**
     * UserAccessPolicy in UNKNOWN state allows only incoming traffic from
     * control user when domains differs. user.
     */
    @Test
    public void testIncomingUnknown() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);

        assertThat(uac.getState(), CoreMatchers.is(UserAccessPolicy.State.UNKNOWN));

        //allow if domains match
        //allow if incomplete (any except from is null)
        assertTrue(allowInOut(uac, "from", "domain", "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, "domain", "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, null, "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", "domain", "to", null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", "domain", null, null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", null, null, null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, null, null, EnumSet.allOf(PacketType.class)));

        //allow if from is control
        assertTrue(allowInOut(uac, "from", "domain", CONTROL_LOGIN, CONTROL_DOMAIN, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, CONTROL_LOGIN, CONTROL_DOMAIN, CONTROL_LOGIN, CONTROL_DOMAIN, EnumSet.allOf(PacketType.class)));

        //deny if domains not match, even if it is from control user or control user name match or control domain match 
        assertTrue(denyInOut(uac, CONTROL_LOGIN, CONTROL_DOMAIN, "to", "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", "fromDomain", "to", "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", "fromDomain", CONTROL_LOGIN, "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", "fromDomain", "remote", CONTROL_DOMAIN, EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, CONTROL_LOGIN, "fromDomain", "to", "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", CONTROL_DOMAIN, "to", "toDomain", EnumSet.allOf(PacketType.class)));

    }

    /**
     * UserAccessPolicy in DISABLED state denies every traffic from different
     * domains. user.
     */
    @Test
    public void testIncomingDisabled() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);
        uac.setState(UserAccessPolicy.State.DISABLED);

        assertThat(uac.getState(), CoreMatchers.is(UserAccessPolicy.State.DISABLED));

        //allow if domains match
        //allow if incomplete (any except from is null)
        assertTrue(allowInOut(uac, "user", "domain", "user2", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, "domain", "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, null, "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", "domain", "to", null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", "domain", null, null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", null, null, null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, null, null, EnumSet.allOf(PacketType.class)));

        //deny if domains not match even for control user
        assertTrue(denyInOut(uac, "from", "fromDomain", CONTROL_LOGIN, CONTROL_DOMAIN, EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, CONTROL_LOGIN, CONTROL_DOMAIN, "to", "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", "fromDomain", "to", "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", "fromDomain", CONTROL_LOGIN, "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", "fromDomain", "remote", CONTROL_DOMAIN, EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, CONTROL_LOGIN, "fromDomain", "to", "toDomain", EnumSet.allOf(PacketType.class)));
        assertTrue(denyInOut(uac, "from", CONTROL_DOMAIN, "to", "toDomain", EnumSet.allOf(PacketType.class)));
    }

    /**
     * UserAccessPolicy in ENABLED, the traffic is allowed or denied if user is
     * allowed / is not allowed to send messages outside his/her domain. user.
     */
    @Test
    public void testIncomingEnabled() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);

        Users users = mock(Users.class);
        ServiceBus bus = mock(ServiceBus.class);

        Settings settings = mock(ExtService.Settings.class);
        when(settings.isSendOutsideEnabled()).thenReturn(Boolean.TRUE);

        ExtService extService = mock(ExtService.class);
        when(extService.getSettings()).thenReturn(settings);

        uac.setExtService(users, extService, bus);
        assertThat(uac.getState(), CoreMatchers.is(UserAccessPolicy.State.ENABLED));

        //allow if domain match
        //allow if incomplete
        assertTrue(allowInOut(uac, "user", "domain", "user2", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, "domain", "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, "to", "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, null, "domain", EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", "domain", "to", null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", "domain", null, null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, "from", null, null, null, EnumSet.allOf(PacketType.class)));
        assertTrue(allowInOut(uac, null, null, null, null, EnumSet.allOf(PacketType.class)));


        User allowed = mock(User.class);
        when(allowed.isAllowedSendOutside()).thenReturn(Boolean.TRUE);
        User denied = mock(User.class);
        when(denied.isAllowedSendOutside()).thenReturn(Boolean.FALSE);
        UserControl userControlMock = mock(UserControl.class);
        when(userControlMock.find("allowed")).thenReturn(allowed);
        when(userControlMock.find("denied")).thenReturn(denied);
        when(users.getControl("existing")).thenReturn(userControlMock);

        //allow for users from not managed domain
        assertTrue(allowInOut(uac, "from", "fromDomain", "to", "notExisting", EnumSet.allOf(PacketType.class)));
        //allow for users from managed with allow flag
        assertTrue(allowInOut(uac, "from", "fromDomain", "allowed", "existing", EnumSet.allOf(PacketType.class)));

        //deny for users who are not allowed to send emails outside their domins
        assertTrue(denyInOut(uac, "from", "fromDomain", "denied", "existing", EnumSet.allOf(PacketType.class)));
        //deny for unknown users
        assertTrue(denyInOut(uac, "from", "fromDomain", "unknown", "existing", EnumSet.allOf(PacketType.class)));

    }

    private boolean allowInOut(UserAccessPolicy uac, String from, String fromDomain, String to, String toDomain, EnumSet<PacketType> packetTypes) {
        for (PacketType type : packetTypes) {
            if (!uac.incoming(from, fromDomain, to, toDomain, type) || !uac.outgoing(from, fromDomain, to, toDomain, type)) {
                return false;
            }
        }

        return true;
    }

    private boolean denyInOut(UserAccessPolicy uac, String from, String fromDomain, String to, String toDomain, EnumSet<PacketType> packetTypes) {
        for (PacketType type : packetTypes) {
            if (uac.incoming(from, fromDomain, to, toDomain, type) || uac.outgoing(from, fromDomain, to, toDomain, type)) {
                return false;
            }
        }

        return true;
    }

    @Test
    public void testOutgoing() {
    }

    @Test
    public void testIsSameDomain() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);

        assertTrue(uac.isSameDomain("", "domain.org", "domain.org"));
        assertTrue(uac.isSameDomain("conference", "domain.org", "domain.org"));
        assertTrue(uac.isSameDomain("conference", "conference.domain.org", "domain.org"));
        assertTrue(uac.isSameDomain("conference", "domain.org", "conference.domain.org"));

        assertFalse(uac.isSameDomain("", "domain.org", "other.org"));
        assertFalse(uac.isSameDomain("domain", "domain.org", "other.org"));
        assertFalse(uac.isSameDomain("other", "domain.org", "other.org"));
        assertFalse(uac.isSameDomain("prefix", "another.domain.org", "domain.org"));
        assertFalse(uac.isSameDomain("prefix", "domain.org", "another.domain.org"));
    }

    @Test
    public void testIsSameDomainCase() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);

        assertTrue(uac.isSameDomain("", "domain.org", "DoMaIn.org"));
        assertTrue(uac.isSameDomain("conference", "domain.org", "DoMaIn.org"));
        assertTrue(uac.isSameDomain("conference", "CoNfErEnCe.domain.org", "DoMaIn.org"));
        assertTrue(uac.isSameDomain("conference", "domain.org", "CoNfErEnCe.DoMaIn.org"));
    }

    @Test
    public void testIncomingResponse() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);
        assertThat(uac.incomingResponse(), CoreMatchers.notNullValue());
    }

    @Test
    public void testOutgoingResponse() {
        UserAccessPolicy uac = new UserAccessPolicy(this.conf);
        assertThat(uac.outgoingResponse(), CoreMatchers.notNullValue());
    }
}
