/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.connect;

import com.kerio.im.common.error.ServiceException;
import com.kerio.im.external.ExtService;
import com.kerio.lib.json.api.connect.admin.iface.common.IpAddressGroups;
import com.kerio.lib.json.api.connect.admin.struct.GroupLimit;
import com.kerio.lib.json.api.connect.admin.struct.Service;
import com.kerio.lib.json.api.connect.admin.struct.common.IpAddressEntry;
import com.kerio.lib.json.api.connect.admin.struct.common.IpAddressGroup;
import com.kerio.lib.json.api.connect.admin.struct.common.SearchQuery;
import com.kerio.lib.json.api.connect.admin.struct.common.SubCondition;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class AddressFilter {

    private static final Logger LOG = Logger.getLogger(AddressFilter.class.getName());

    public static AddressFilter create(final IpAddressGroups ipAddressGroupsApi) {
        return new AddressFilter(readAllGroups(ipAddressGroupsApi));
    }

    private static Map<String, List<IpAddressEntry>> readAllGroups(final IpAddressGroups ipAddressGroupsApi) {
        Map<String, List<IpAddressEntry>> result = new HashMap<String, List<IpAddressEntry>>();
        SearchQuery groupSearchQuery = new SearchQuery();
        groupSearchQuery.setStart(0);
        IpAddressGroups.GetGroupListResult groups = ipAddressGroupsApi.getGroupList();
        for (IpAddressGroup group : groups.getGroups()) {
            result.put(group.getId(), readGroup(ipAddressGroupsApi, group.getId()));
        }
        return result;
    }

    private static List<IpAddressEntry> readGroup(final IpAddressGroups ipAddressGroupsApi, String groupId) {
        List<IpAddressEntry> result;
        SubCondition groupIdCondition = new SubCondition();
        groupIdCondition.setFieldName("groupId");
        groupIdCondition.setValue(groupId);
        SearchQuery groupQuery = new SearchQuery();
        groupQuery.setConditions(new SubCondition[]{groupIdCondition});
        IpAddressGroups.GetResult groups = ipAddressGroupsApi.get(groupQuery);
        if ((groups.getTotalItems() != null) && (groups.getTotalItems() > 0)) {
            result = Arrays.asList(groups.getList());
        } else {
            result = new LinkedList<IpAddressEntry>();
        }
        return result;
    }

    public interface ErrorHandler {
        void groupNotFound(String groupId) throws ServiceException;
    }

    final Map<String, List<IpAddressEntry>> ipAddressEntries;

    AddressFilter(Map<String, List<IpAddressEntry>> ipAddressEntries) {
        this.ipAddressEntries = ipAddressEntries;
    }

    public void apply(Service service, ExtService.AccessHandler handler, ErrorHandler errorHandler) {
        GroupLimit group = service.getGroup();
        if (group.getIsUsed()) {
            if (group.getIpGroup() != null) {
                apply(group.getIpGroup().getId(), true, handler, errorHandler);
            } else {
                errorHandler.groupNotFound(null);
            }
        }
    }

    public void apply(String groupId, ExtService.AccessHandler handler, ErrorHandler errorHandler) {
        apply(groupId, true, handler, errorHandler);
    }

    private void apply(String groupId, boolean enabled, ExtService.AccessHandler handler, ErrorHandler errorHandler) {
        if (this.ipAddressEntries.get(groupId) != null) {
            for (IpAddressEntry entry : this.ipAddressEntries.get(groupId)) {
                boolean entryEnabled = enabled && entry.getEnabled();
                switch (entry.getType()) {
                    case ChildGroup:
                        apply(entry.getChildGroupId(), entryEnabled, handler, errorHandler);
                        break;
                    case Host:
                        handler.filterHost(entry.getHost(), entryEnabled);
                        break;
                    case Network:
                        handler.filterMask(entry.getAddr1(), entry.getAddr2(), entryEnabled);
                        break;
                    case Range:
                        handler.filterRange(entry.getAddr1(), entry.getAddr2(), entryEnabled);
                        break;
                    case ThisMachine:
                        handler.filterLocalhost(entryEnabled);
                        break;
                    default:
                        LOG.log(Level.WARNING, "Received unsupported IP group type {0}", entry.getType());
                }
            }
        } else {
            errorHandler.groupNotFound(groupId); //error could be silently ignored or exception thrown
        }
    }

}
