/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.connect;

import com.kerio.lib.json.api.client.KApiInterface;
import com.kerio.lib.json.api.client.KJsonRpcClient;
import com.kerio.lib.json.api.client.KJsonRpcSession;
import java.lang.reflect.Method;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.json.JSONException;
import org.json.JSONObject;

/**
 *
 * @author tsoukup
 */
public abstract class ApiClient {

	private static final Logger LOG = Logger.getLogger(ApiClient.class.getName());
	private static final String HOSTNAME_FORMAT = "http%1$s://%2$s:%3$d";
	//

	protected static String hostname(String server, Integer port, Boolean secure) {
		return String.format(HOSTNAME_FORMAT, secure ? "s" : "", server, port);
	}
	//
	final KJsonRpcSessionPrivate session;
	private final KJsonRpcClientPrivate client;
    private final FileApiClient fileDownload;
	//

	abstract void login();

	abstract void logout();

	abstract String getName();

	//
	public ApiClient(String hostname, String path) {
		this.session = new KJsonRpcSessionPrivate(hostname, path);
		this.client = new KJsonRpcClientPrivate(session);
		this.fileDownload = new FileApiClient(this.session);
	}

	public KJsonRpcSession getSession() {
		return session;
	}

	public KJsonRpcClient getClient() {
		return client;
	}

    public FileApiClient getFileDownload() {
        return fileDownload;
    }
    
    public String getHostname() {
        return this.session.getHostname();
    }

	public <T extends KApiInterface> T getApi(Class<T> clazz) {
		return this.client.getProxy(clazz);
	}

	class KJsonRpcSessionPrivate extends KJsonRpcSession {

		private static final int SESSION_TIMEOUT = -32001;

		public KJsonRpcSessionPrivate(String hostname, String apiPath) {
			setHostname(hostname);
			setApiPath(apiPath);
		}

		public void reconnect() {
			LOG.log(Level.INFO, "{0} session timeout, reconnecting", ApiClient.this.getName());
			ApiClient.this.login();
		}

		@Override
		public JSONObject sendAndReceive(JSONObject message, Method method) {
			JSONObject result = super.sendAndReceive(message, method);
			if (isSessionTimeout(result)) {
				//TODO prevent reconnecting twice
				reconnect();
				result = super.sendAndReceive(message, method);
			}

			return result;
		}

		private boolean isSessionTimeout(JSONObject message) {
			boolean result = false;
			try {
				if (message.has("error")) { //detected error response
					JSONObject error = message.getJSONObject("error");
					int code = new Integer(error.has("code") ? error.getInt("code") : 0);
					if (code == SESSION_TIMEOUT) {
						result = true;
					}
				}
			} catch (JSONException ex) {
				LOG.log(Level.FINE, "Detecting session timeout failed, invalid error response");
			}
			return result;
		}
	}

	class KJsonRpcClientPrivate extends KJsonRpcClient {

		public KJsonRpcClientPrivate(KJsonRpcSession session) {
			super(session);
		}
	}
}
