/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.connect;

import com.kerio.im.bean.VCard;
import com.kerio.im.external.ExtService;
import com.kerio.lib.json.api.connect.webmail.iface.changes.Changes;
import com.kerio.lib.json.api.connect.webmail.iface.changes.Changes.GetFolderResult;
import com.kerio.lib.json.api.connect.webmail.iface.changes.Changes.GetFolderSyncKeyResult;
import com.kerio.lib.json.api.connect.webmail.iface.contacts.Contacts;
import com.kerio.lib.json.api.connect.webmail.iface.contacts.Contacts.GetResult;
import com.kerio.lib.json.api.connect.webmail.iface.session.Session;
import com.kerio.lib.json.api.connect.webmail.struct.changes.Change;
import com.kerio.lib.json.api.connect.webmail.struct.changes.ChangeType;
import com.kerio.lib.json.api.connect.webmail.struct.changes.ItemType;
import com.kerio.lib.json.api.connect.webmail.struct.contacts.Contact;
import com.kerio.lib.json.api.connect.webmail.struct.contacts.ContactType;
import com.kerio.lib.json.api.connect.webmail.struct.contacts.EmailAddress;
import com.kerio.lib.json.api.connect.webmail.struct.contacts.PhoneNumber;
import com.kerio.lib.json.api.connect.webmail.struct.contacts.PostalAddress;
import com.kerio.lib.json.api.connect.webmail.struct.common.ApiApplication;
import com.kerio.lib.json.api.connect.webmail.struct.common.CompareOperator;
import com.kerio.lib.json.api.connect.webmail.struct.common.SearchQuery;
import com.kerio.lib.json.api.connect.webmail.struct.common.SubCondition;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.codec.binary.Base64OutputStream;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.codec.digest.DigestUtils;

/**
 *
 * @author tskoda
 */
public class WebmailApi extends ApiClient {

    private static final Logger LOG = Logger.getLogger(WebmailApi.class.getName());
    private static final String API_PATH = "/webmail/api/jsonrpc/";
    public static final String DEF_IMAGE_CONTENT_TYPE = "image/jpeg";
    public static final byte[] NEW_LINE = {'\n'};
    public static final int MAX_LINE_LENGTH = 64;
    public static final boolean DO_ENCODING = true;

    private static final class PhotoBuffer extends OutputStream {

        private final ByteArrayOutputStream photo64;
        private final Base64OutputStream photo64converter;
        private final MessageDigest photoSHA1;

        public PhotoBuffer() throws NoSuchAlgorithmException, FileNotFoundException {
            this.photo64 = new ByteArrayOutputStream();
            /* aligning base64 output breaks contact photos on Messages / Yosemite, #CONNECT-41688 */
            //this.photo64converter = new Base64OutputStream(photo64, DO_ENCODING, MAX_LINE_LENGTH, NEW_LINE);
            this.photo64converter = new Base64OutputStream(photo64, DO_ENCODING, -1, null);
            this.photoSHA1 = DigestUtils.getSha1Digest();
        }

        @Override
        public void write(int i) throws IOException {
            this.photo64converter.write(i);
            this.photoSHA1.update((byte) i);
        }

        @Override
        public void flush() throws IOException {
            this.photo64converter.flush();
        }

        @Override
        public void close() throws IOException {
            this.photo64converter.close();
        }

        public byte[] getHash() {
            return this.photoSHA1.digest();
        }

        public byte[] getPhoto64() {
            return this.photo64.toByteArray();
        }
    }
    private final ApiApplication app;
    private final String username;
    private final String password;

    public WebmailApi(String username, String password, String server, Integer port, Boolean secure) {
        super(hostname(server, port, secure), API_PATH);

        this.app = new ApiApplication();
        this.app.setName("Instant Messaging Webmail API client");
        this.app.setVendor("Kerio Technologies, inc");
        this.app.setVersion("1.0");

        this.username = username;
        this.password = password;

    }

    @Override
    void login() {
        LOG.log(Level.FINE, "Connecting to WEBMAIL API on {0}", this.getSession().getHostname());
        getApi(Session.class).login(username, password, this.app);
    }

    @Override
    void logout() {
        getApi(Session.class).logout();
    }

    @Override
    String getName() {
        return this.app.getName();
    }

    public Runnable getPing() {
        return new WebmailApiPing(this);
    }

    private long syncKeyForFolder(String folder) {
        GetFolderSyncKeyResult folderSyncKey = getApi(Changes.class).getFolderSyncKey(folder);
        return folderSyncKey.getSyncKey() != null ? folderSyncKey.getSyncKey() : 0;
    }

    private String publicFolderPath(boolean isPublic, String domain) {
        return isPublic ? "#public/Contacts" : "#public@" + domain + "/Contacts";
    }

    public long getAllVCards(boolean isPublic, long syncKey, String domain, ExtService.VCardCache cache, VCard.VCardFactory factory) {

        String folderPath = publicFolderPath(isPublic, domain);
        long freshSyncKey;

        //on first run, cache has source set to null, so full sync is performed
        if (!folderPath.equals(cache.getSource())) {
            cache.clear(); //applies only for change in public folder type
            freshSyncKey = syncKeyForFolder(folderPath);
            fullSync(folderPath, domain, cache, factory);
            cache.setSource(folderPath);
        } else {
            freshSyncKey = syncKey;
        }

        return incrementSync(freshSyncKey, folderPath, domain, cache, factory);
    }

    public long incrementSync(long syncKey, String folder, String domain, ExtService.VCardCache cache, VCard.VCardFactory factory) {
        LOG.log(Level.FINE, "Refreshing vcard cache, performing incremental synchronization");
        long freshSyncKey = syncKeyForFolder(folder);

        if (syncKey < freshSyncKey) {
            int added = 0;
            int removed = 0;
            int changed = 0;
            int skipped = 0;
            GetFolderResult folderChangesResult = getApi(Changes.class).getFolder(folder, syncKey);
            for (Change change : folderChangesResult.getList()) {
                if ((change.getItemType() == ItemType.itContact) && (change.getItemId() != null)) {
                    if (change.getType() == ChangeType.chtCreated) {
                        if (!cache.containsId(change.getItemId())) {
                            Contact contact = getContact(change.getItemId());
                            if (contact != null) {
                                if (contact.getIsGalContact()) {
                                    try {
                                        String userLogin = getPrimaryEmail(contact, domain);
                                        if (userLogin != null) {
                                            cache.put(map(contact, factory.instance()), change.getItemId(), userLogin);
                                            added++;
                                        } else {
                                            LOG.log(Level.FINE,
                                                    "Skipping GAL contact {0} {1}. Contact is from different domain or default email address is missing.",
                                                    new Object[]{contact.getFirstName(), contact.getSurName()});
                                            skipped++;
                                        }
                                    } catch (Exception e) {
                                        LOG.log(Level.WARNING, "Failed to load new vcard", e);
                                        skipped++;
                                    }
                                }
                            } else {
                                LOG.log(Level.INFO, "Adding vcard for contact {0} failed, contact not found", change.getItemId());
                            }
                        } else {
                            LOG.log(Level.INFO, "Adding vcard for contact {0} ignored, contact already exist", change.getItemId());
                        }
                    } else if (change.getType() == ChangeType.chtDeleted) {
                        if (cache.containsId(change.getItemId())) {
                            cache.removeById(change.getItemId());
                            removed++;
                        }

                    } else if (change.getType() == ChangeType.chtModified) {
                        if (cache.containsId(change.getItemId())) {
                            Contact contact = getContact(change.getItemId());
                            if (contact != null) {
                                if (contact.getIsGalContact()) {
                                    try {
                                        //add whole new, do not update existing
                                        String userLogin = getPrimaryEmail(contact, domain);
                                        if (userLogin != null) {
                                            cache.put(map(contact, factory.instance()), change.getItemId(), userLogin);
                                            changed++;
                                        } else {
                                            LOG.log(Level.FINE,
                                                    "Skipping GAL contact {0} {1}. Contact is from different domain or default email address is missing.",
                                                    new Object[]{contact.getFirstName(), contact.getSurName()});
                                            skipped++;
                                        }
                                    } catch (Exception e) {
                                        LOG.log(Level.WARNING, "Failed to load modified vcard", e);
                                        skipped++;
                                    }
                                }
                            } else {
                                skipped++;
                                LOG.log(Level.INFO, "Updating vcard for contact {0} failed, contact not found", change.getItemId());
                            }
                        }
                    }
                }
            }
            LOG.log(Level.INFO, "Refreshing vcard cache successful. Added {0}, removed {1}, updated {2}, skipped {3}, total items in cache {4}.", new Object[]{added, removed, changed, skipped, cache.count()});
        } else if (syncKey == freshSyncKey) {
            LOG.log(Level.INFO, "Refreshing vcard cache successful, no changes detected. Watermark: {0}", new Object[]{syncKey, freshSyncKey});
        } else if (syncKey > freshSyncKey) {
            LOG.log(Level.WARNING, "Refreshing vcard cache not performed, received invalid watermark. Last: {0}, Current: {1}", new Object[]{syncKey, freshSyncKey});
        }

        return freshSyncKey;

    }

    public void fullSync(String folder, String domain, ExtService.VCardCache cache, VCard.VCardFactory factory) {
        LOG.log(Level.FINE, "Refreshing vcard cache, performing full synchronization");
        Contacts contactsApi = getApi(Contacts.class);
        SearchQuery query = new SearchQuery();
        SubCondition filterDomain = new SubCondition();
        filterDomain.setFieldName("email");
        filterDomain.setValue(domain);
        filterDomain.setComparator(CompareOperator.Like);

        query.setFields(new String[]{
                    "id",
                    "commonName",
                    "firstName",
                    "middleName",
                    "surName",
                    "nickName",
                    "phoneNumbers",
                    "emailAddresses",
                    "postalAddresses",
                    "urls",
                    "birthDay",
                    "companyName",
                    "departmentName",
                    "profession",
                    "photo",
                    "isGalContact",
                    "type"});

        query.setConditions(new SubCondition[]{filterDomain});
        GetResult contactList = contactsApi.get(new String[]{folder}, query);

        int added = 0;
        int skipped = 0;
        for (Contact contact : contactList.getList()) {
            String userLogin;
            if (contact.getIsGalContact() && contact.getType() == ContactType.ctContact && (userLogin = getPrimaryEmail(contact, domain)) != null) {
                try {
                    cache.put(map(contact, factory.instance()), contact.getId(), userLogin);
                    added++;
                } catch (Exception e) {
                    LOG.log(Level.WARNING, "Failed to load " + userLogin + " vcard", e);
                    skipped++;
                }
            }
        }

        LOG.log(Level.INFO, "Refreshing vcard cache successful. Added {0}, skipped {1}", new Object[]{added, skipped});

    }

    private Contact getContact(String contactId) {
        Contacts contactsApi = getApi(Contacts.class);
        Contact[] contactResult = contactsApi.getById(new String[]{contactId}).getResult();

        Contact result = null;
        if ((contactResult != null) && (contactResult.length == 1)) {
            result = contactResult[0];
        }
        return result;
    }

    private String getPrimaryEmail(Contact contact, String domain) {
        String result = null;
        //TODO change to regex
        String domainWithAmp = "@" + domain;
        for (EmailAddress contactEmail : contact.getEmailAddresses()) {
            //if (contactEmail.getPreferred() && 
            if (contactEmail.getAddress() != null && contactEmail.getAddress().toLowerCase().endsWith(domainWithAmp)) {
                result = contactEmail.getAddress();
                break;
            }
        }
        return result;
    }

    private VCard map(Contact contact, VCard vcard) throws MalformedURLException, IOException, NoSuchAlgorithmException {

        putValue(vcard, "FN", contact.getCommonName());
        putValue(vcard, "NICKNAME", contact.getNickName());

        VCard name = vcard.put("N");
        putValue(name, "FAMILY", contact.getSurName());
        putValue(name, "GIVEN", contact.getFirstName());
        putValue(name, "MIDDLE", contact.getMiddleName());
        putValue(name, "PREFIX", contact.getTitleBefore());
        putValue(name, "SUFFIX", contact.getTitleAfter());

        putValue(vcard, "SORT-STRING", contact.getSurName());

        putValue(vcard, "BDAY", contact.getBirthDay());

        if (contact.getPhoneNumbers() != null) {
            for (PhoneNumber pn : contact.getPhoneNumbers()) {
                VCard tel = vcard.put("TEL");
                switch (pn.getType()) {
                    case TypeWorkFax:
                        putTags(tel, "WORK", "FAX");
                        break;
                    case TypeWorkVoice:
                        putTags(tel, "WORK", "VOICE");
                        break;
                    case TypeHomeFax:
                        putTags(tel, "HOME", "FAX");
                        break;
                    case TypeHomeVoice:
                        putTags(tel, "HOME", "VOICE");
                        break;
                    case TypeIsdn:
                        putTags(tel, "ISDN");
                        break;
                    case TypeMobile:
                        putTags(tel, "CELL");
                        break;
                    case TypePrimary:
                        putTags(tel, "PREF");
                        break;
                }
                putValueEmpty(tel, "NUMBER", pn.getNumber());
            }
        }

        if (contact.getEmailAddresses() != null) {
            for (EmailAddress ea : contact.getEmailAddresses()) {
                VCard email = vcard.put("EMAIL");
                switch (ea.getType()) {
                    case EmailHome:
                        putTags(email, "HOME");
                        break;
                    case EmailWork:
                        putTags(email, "WORK");
                        break;
                }

                if (ea.getPreferred()) {
                    putTags(email, "PREF");
                }

                putValue(email, "USERID", ea.getAddress());
            }
        }

        if (contact.getPostalAddresses() != null) {
            for (PostalAddress postal : contact.getPostalAddresses()) {
                VCard address = vcard.put("ADR");
                switch (postal.getType()) {
                    case AddressHome:
                        putTags(address, "HOME");
                        break;
                    case AddressWork:
                        putTags(address, "WORK");
                        break;
                    default:
                }

                if (postal.getPreferred()) {
                    putTags(address, "PREF");
                }

                putValue(address, "EXTADD", postal.getExtendedAddress());
                putValue(address, "STREET", postal.getStreet());
                putValue(address, "LOCALITY", postal.getLocality());
                putValue(address, "CTRY", postal.getCountry());
                putValue(address, "REGION", postal.getState());
                putValue(address, "PCODE", postal.getZip());
                putValue(address, "POBOX", postal.getPobox());

            }
        }

        putValue(vcard, "TITLE", contact.getProfession());

        VCard org = vcard.put("ORG");
        putValue(org, "ORGNAME", contact.getCompanyName());
        putValue(org, "ORGUNIT", contact.getDepartmentName());

        VCard photo = vcard.put("PHOTO");

        if ((contact.getPhoto() != null) && (contact.getPhoto().getUrl() != null) && (!contact.getPhoto().getUrl().isEmpty())) {
            PhotoBuffer photoBuffer = new PhotoBuffer();
            String contentType = getFileDownload().fileContent(this.getSession().getHostname() + contact.getPhoto().getUrl(), photoBuffer);
            putValue(photo, "TYPE", contentType != null ? contentType : DEF_IMAGE_CONTENT_TYPE);
            putPhoto(photo, "BINVAL", photoBuffer.getPhoto64(), photoBuffer.getHash());
        } else {
            putTags(photo, "BINVAL");
        }

        putValue(vcard, "URL", contact.getUrls().length > 0 ? contact.getUrls()[0].getUrl() : null);
        putValue(vcard, "NOTE", contact.getComment());

        return vcard;
    }

    private void putTags(VCard vcard, String... tags) {
        for (String tag : tags) {
            if (tag != null) {
                vcard.put(tag);
            }
        }
    }

    private void putValue(VCard vcard, String tag, String value) {
        if ((value != null) && (!value.isEmpty())) {
            vcard.put(tag, value);
        }
    }

    private void putPhoto(VCard vcard, String tag, byte[] value, byte[] hash) throws UnsupportedEncodingException {
        if ((value != null) && (value.length > 0)) {
            vcard.photo(tag, new String(value, "UTF8"), Hex.encodeHexString(hash));
        }
    }

    private void putValueEmpty(VCard vcard, String tag, String value) {
        if ((value != null) && (!value.isEmpty())) {
            vcard.put(tag, value);
        } else {
            vcard.put(tag);
        }
    }

    static class WebmailApiPing implements Runnable {

        private static final Logger LOG = Logger.getLogger(WebmailApiPing.class.getName());
        private final ApiClient apiClient;

        private WebmailApiPing(ApiClient apiClient) {
            this.apiClient = apiClient;
        }

        @Override
        public void run() {
            LOG.fine("Webmail API session ping");
            this.apiClient.getApi(Session.class).whoAmI();
        }
    }
}
