/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.im;

import com.kerio.im.external.IMContactID;
import com.kerio.im.external.IMConversation;
import com.kerio.im.external.IMMessage;
import com.kerio.im.external.IMPresence;
import com.kerio.im.external.ImResponseEnvelope;
import com.kerio.lib.json.api.client.KApiInterface;
import com.kerio.lib.json.api.client.KJsonRpcSession;
import com.kerio.lib.json.api.im.iface.Server;
import com.kerio.lib.json.api.im.iface.Server.RegisterResult;
import com.kerio.lib.json.api.im.iface.Server.SendMessageOnBehalfResult;
import com.kerio.lib.json.api.im.iface.Server.SubscribeGlobalPresenceResult;
import com.kerio.lib.json.api.im.iface.im;
import com.kerio.lib.json.api.im.iface.im.SendMessageResult;
import com.kerio.lib.json.api.im.iface.im.SubscribeConversationsResult;
import com.kerio.lib.json.api.im.iface.im.SubscribeMessagesResult;
import com.kerio.lib.json.api.im.iface.im.SubscribePresenceResult;
import com.kerio.lib.json.api.im.struct.Conversation;
import com.kerio.lib.json.api.im.struct.ConversationEvent;
import com.kerio.lib.json.api.im.struct.Message;
import com.kerio.lib.json.api.im.struct.MessageEvent;
import com.kerio.lib.json.api.im.struct.Presence;
import com.kerio.lib.json.api.im.struct.Status;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author oleopold
 */
public class ImApi {

    private static final Logger LOG = Logger.getLogger(ImApi.class.getName());
    private static final String API_PATH = "/jsonrpc/1/";
    private static final String HOSTNAME_FORMAT = "http%1$s://%2$s:%3$d";
    private static int sessionidCount = 1;

    private ImJsonRpcSession session;
    private final ImJsonRpcClient client;

    protected static String hostname(String server, Integer port, Boolean secure) {
        return String.format(HOSTNAME_FORMAT, secure ? "s" : "", server, port);
    }

    public KJsonRpcSession getSession() {
        return session;
    }

    public void setRegisterId(String registerid) {
        session.setRegisterid(registerid);
    }

    public ImApi(String server, Integer port, Boolean secure, String user, String registerid) {
        this.session = new ImJsonRpcSession(hostname(server, port, secure), API_PATH, user, "XMPP-BRIDGE-SESSION-" + Integer.toString(sessionidCount++), registerid);
        this.client = new ImJsonRpcClient(session);
    }

    public ImApi(String server, Integer port, Boolean secure, String registerid) {
        this(server, port, secure, ".imadmin", registerid);
    }

    void login() {
        LOG.log(Level.FINE, "Connecting to IM API on {0}", this.getSession().getHostname());
    }

    void logout() {
    }

    private <T extends KApiInterface> T getApi(Class<T> clazz) {
        return this.client.getProxy(clazz);
    }

    public String register(String ip, Integer port, String user, String password) {
        Server server = getApi(Server.class);
        RegisterResult result = server.register(ip, port, user, password, true);
        return result.getId();
    }

    void ping(String id) {
        Server server = getApi(Server.class);
        server.ping(id);
    }

    synchronized public ImResponseEnvelope<List<IMPresence>> subscribeGlobalPresence(Integer requestId) {
        Server server = getApi(Server.class);
        client.setRequestId(requestId);
        SubscribeGlobalPresenceResult apiResult = server.subscribeGlobalPresence();
        List<IMPresence> result = new ArrayList<>();
        if (apiResult.getList() != null) {
            for (Presence p : apiResult.getList()) {
                result.add(convertJsonPresence(p));
            }
        }
        return new ImResponseEnvelope(session.getSessionid(), client.getLastRequestId(), result);
    }

    private IMPresence convertJsonPresence(Presence p) throws IllegalArgumentException {
        IMContactID contactId = new IMContactID(p.getContactId());
        return new IMPresence(contactId, toImStatus(p.getStatus()), p.getText());
    }

    synchronized public ImResponseEnvelope<List<IMPresence>> subscribePresence() {
        im imClient = getApi(im.class);
        SubscribePresenceResult apiResult = imClient.subscribePresence();
        List<IMPresence> result = new ArrayList<>();
        if (apiResult.getList() != null) {
            for (Presence p : apiResult.getList()) {
                result.add(convertJsonPresence(p));
            }
        }
        return new ImResponseEnvelope(session.getSessionid(), client.getLastRequestId(), result);
    }

    IMPresence.Status toImStatus(Status status) {
        switch (status) {
            case available:
                return IMPresence.Status.available;
            case offline:
                return IMPresence.Status.offline;
            case dnd:
                return IMPresence.Status.dnd;
            case away:
                return IMPresence.Status.away;
            case invisible:
                return IMPresence.Status.invisible;
            default:
                throw new AssertionError(status.name());

        }
    }

    Status toJsonStatus(IMPresence.Status status) {
        switch (status) {
            case available:
                return Status.available;
            case offline:
                return Status.offline;
            case dnd:
                return Status.dnd;
            case away:
                return Status.away;
            case invisible:
                return Status.invisible;
            default:
                throw new AssertionError(status.name());
        }
    }

    synchronized public void setPresence(IMPresence.Status status, String text) {
        im imClient = getApi(im.class);
        imClient.setPresence(toJsonStatus(status), text);
    }

    synchronized public ImResponseEnvelope<List<IMConversation>> subscribeConversations(Integer requestId) {
        im imClient = getApi(im.class);
        client.setRequestId(requestId);
        SubscribeConversationsResult result = imClient.subscribeConversations();
        List<IMConversation> convertedResult = new ArrayList<>();
        if (result.getList() != null) {
            for (Conversation conversation : result.getList()) {
                try {
                    convertedResult.add(convertImConversation(conversation));
                } catch (IllegalArgumentException e) {

                }
            }
        }
        return new ImResponseEnvelope(session.getSessionid(), client.getLastRequestId(), convertedResult);
    }

    private List<IMContactID> convertContactIds(String[] ids) throws IllegalArgumentException {
        ArrayList<IMContactID> result = new ArrayList<>();
        if (ids != null) {
            for (String id : ids) {
                result.add(new IMContactID(id));
            }
        }
        return result;
    }

    private IMConversation convertImConversation(Conversation conversation) throws IllegalArgumentException {
        return new IMConversation(
                conversation.getConversationId(),
                conversation.getReceivedUnreadCount(),
                conversation.getReceivedLastReadId(),
                convertContactIds(conversation.getContacts()),
                toImConversationEvent(conversation.getEvent()),
                conversation.getMuted(),
                conversation.getLastActivity(),
                conversation.getSentLastReadId(),
                conversation.getSentLastDeliveredId());
    }

    private IMConversation.ConversationEvent toImConversationEvent(ConversationEvent event) {
        if (event == null) {
            return IMConversation.ConversationEvent.Created;
        }
        switch (event) {
            case Created:
                return IMConversation.ConversationEvent.Created;
            case Updated:
                return IMConversation.ConversationEvent.Updated;
            case Delivered:
                return IMConversation.ConversationEvent.Delivered;
            case Read:
                return IMConversation.ConversationEvent.Read;
            default:
                throw new AssertionError(event.name());
        }
    }

    synchronized public ImResponseEnvelope<List<IMMessage>> subscribeMessages(Integer conversationId, Integer currentMessageId, Integer requestId) {
        im imClient = getApi(im.class);
        client.setRequestId(requestId);
        SubscribeMessagesResult result = imClient.subscribeMessages(conversationId, currentMessageId == null ? 0 : currentMessageId , 0);
        List<IMMessage> convertedResult = new ArrayList<>();
        if (result.getList() != null) {
            for (Message c : result.getList()) {
                try {
                    convertedResult.add(new IMMessage(c.getMessageId(), c.getText(), toImMessageEvent(c.getEvent()), c.getTo(), new IMContactID(c.getFrom()), c.getTime()));
                } catch (IllegalArgumentException e) {
                    LOG.log(Level.WARNING, "SubscribeMessages: Invalid message response.", e);
                }
            }
        }
        return new ImResponseEnvelope(session.getSessionid(), client.getLastRequestId(), convertedResult);
    }

    private IMMessage.MessageEvent toImMessageEvent(MessageEvent event) {
        if (event == null) {
            return IMMessage.MessageEvent.inactive;
        }
        switch (event) {
            case active:
                return IMMessage.MessageEvent.active;
            case inactive:
                return IMMessage.MessageEvent.inactive;
            default:
                throw new AssertionError(event.name());
        }
    }

    synchronized public Integer sendMessage(IMMessage message, Boolean markAsRead) {
        im imClient = getApi(im.class);
        Message m = new Message();
        m.setEvent(toJsonMessageEvent(message.getEvent()));
        m.setText(message.getText());
        m.setTo(message.getTo());
        m.setFrom(message.getFrom().getFullName());
        SendMessageResult result = imClient.sendMessage(m, markAsRead);
        return result.getMessageId();
    }

    private MessageEvent toJsonMessageEvent(IMMessage.MessageEvent event) {
        switch (event) {
            case active:
                return MessageEvent.active;
            case inactive:
                return MessageEvent.inactive;
            default:
                throw new AssertionError(event.name());
        }
    }

    IMConversation createConversation(List<IMContactID> contacts) {
        im imClient = getApi(im.class);
        String[] array = contacts.stream().map(m -> m.getFullName()).toArray(size -> new String[size]);
        return convertImConversation(imClient.createConversation(array).getConversation());
    }

    int generateRequestId() {
        return client.generateRequestId();
    }

    void readConversation(Integer conversationId, Integer msgIdRead) {
        im imClient = getApi(im.class);
        imClient.readConversation(conversationId, msgIdRead);
    }

    public synchronized Integer sendMessageOnBehalf(IMMessage message, boolean markAsRead) {
        Server server = getApi(Server.class);
        Message m = new Message();
        m.setEvent(toJsonMessageEvent(message.getEvent()));
        m.setText(message.getText());
        m.setTo(message.getTo());
        m.setFrom(message.getFrom().getFullName());
        SendMessageOnBehalfResult result = server.sendMessageOnBehalf(m, markAsRead);
        return result.getMessageId();
    }

    public synchronized IMConversation createConversationOnBehalf(IMContactID initiator, List<IMContactID> participants) {
        Server server = getApi(Server.class);
        String[] array = participants.stream().map(m -> m.getFullName()).toArray(size -> new String[size]);
        return convertImConversation(server.createConversationOnBehalf(initiator.getFullName(), array).getConversation());
    }
}
