/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.im;

import com.kerio.im.common.Threads;
import com.kerio.im.common.TypedProperties;
import com.kerio.im.connect.ConnectKey;
import com.kerio.im.external.IMContactID;
import com.kerio.im.external.IMConversation;
import com.kerio.im.external.IMMessage;
import com.kerio.im.external.IMPresence;
import com.kerio.im.external.ISingnalHandler;
import com.kerio.im.external.ImResponseEnvelope;
import com.kerio.im.external.ImService;
import com.kerio.im.external.UserImClient;
import java.lang.reflect.UndeclaredThrowableException;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author oleopold
 */
public class ImServiceImpl implements ImService {
    private static final Logger LOG = Logger.getLogger(ImServiceImpl.class.getName());

    private static final String SERVER_DEFAULT = "localhost";
    public static final int PORT_DEFAULT = 4043;
    public static final boolean SECURE_DEFAULT = false;
    private static final String INVALID_REGISTER_ID = "invalid";

    private final Integer pingInterval = 20;
    private final String userName;
    private final String password;
    private final ImApi imApi;
    private ScheduledExecutorService scheduler;
    private final String listenIp = "127.0.0.1";
    private final Integer listenPort = 4044;
    private final String serverIp;
    private final Integer serverPort;
    private final Boolean secure;
    private String registerid;

    ImServiceImpl(TypedProperties global, Map<String, String> specific) {
        registerid = INVALID_REGISTER_ID;
        userName = "xmpp.bridge." + UUID.randomUUID().toString();
        password = "xmpp.bridge.password." + UUID.randomUUID().toString();
        serverIp = TypedProperties.string(ConnectKey.IMAPI_IP, specific, SERVER_DEFAULT);
        serverPort = TypedProperties.integer(ConnectKey.IMAPI_PORT, specific, PORT_DEFAULT);
        secure = TypedProperties.bool(ConnectKey.IMAPI_SECURE, specific, SECURE_DEFAULT);
        imApi = new ImApi(
                serverIp,
                serverPort,
                secure,
                registerid
        );
    }

    @Override
    public String getUserName() {
        return userName;
    }

    @Override
    public String getPassword() {
        return password;
    }

    @Override
    public synchronized void register(ISingnalHandler watcher) {
        registerid = imApi.register(listenIp, listenPort, userName, java.util.Base64.getEncoder().encodeToString(password.getBytes()));
        imApi.setRegisterId(registerid);
        scheduler = Executors.newScheduledThreadPool(1, Threads.create("[im service] session keep alive"));
        ApiPing pingTask = getPing(watcher);
        ScheduledFuture pingTaskHandler = scheduler.scheduleAtFixedRate(pingTask, pingInterval, pingInterval, TimeUnit.SECONDS);
        pingTask.setHandler(pingTaskHandler);
        LOG.log(Level.FINE, "Started ping task with watcher {0}.", watcher.toString());
    }

    @Override
    public synchronized void unregister() {
        registerid = INVALID_REGISTER_ID;
        imApi.setRegisterId(registerid);
        if (scheduler != null) {
            scheduler.shutdownNow();
            try {
                if (!scheduler.awaitTermination(10, TimeUnit.SECONDS)) {
                    LOG.log(Level.FINE, "Im API session ping does not finish in time.");
                }
            } catch (InterruptedException ex) {
                LOG.log(Level.FINE, "Waiting for ping thread was interrupted.", ex);
            }
        }
        scheduler = null;
    }

    private void ping() {
        imApi.ping(registerid);
    }

    @Override
    public ImResponseEnvelope<List<IMPresence>> subscribeGlobalPresence(Integer requestId) {
        return imApi.subscribeGlobalPresence(requestId);
    }

    @Override
    public UserImClient createUserImClient(IMContactID user) {
        return new UserImClientImpl(new ImApi(
                serverIp,
                serverPort,
                secure,
                user.getFullName(),
                registerid
        ));
    }

    private ApiPing getPing(ISingnalHandler watcher) {
        return new ApiPing(this, watcher);
    }

    @Override
    public Integer generateRequestId() {
        return imApi.generateRequestId();
    }

    @Override
    public void sendMessageOnBehalf(IMMessage imMessage, boolean markAsRead) {
        imApi.sendMessageOnBehalf(imMessage, markAsRead);
    }

    @Override
    public IMConversation createConversationOnBehalf(IMContactID initiator, List<IMContactID> participants) {
        return imApi.createConversationOnBehalf(initiator, participants);
    }

    static class ApiPing implements Runnable {

        private static final Logger LOG = Logger.getLogger(ApiPing.class.getName());
        private final ImServiceImpl imService;
        private final ISingnalHandler watcher;
        private ScheduledFuture handler;

        private ApiPing(ImServiceImpl imService, ISingnalHandler watcher) {
            this.imService = imService;
            this.watcher = watcher;
            this.handler = null;
        }

        @Override
        public synchronized void run() {
            try {
                LOG.log(Level.FINE, "Im API session ping(watcher: {0})", watcher.toString());
                this.imService.ping();
            } catch (UndeclaredThrowableException ex) {
                LOG.log(Level.SEVERE, "IM Ping ended with error.", ex);
            } catch (Throwable ex) {
                LOG.log(Level.SEVERE, "Ping error occured with watcher " + watcher.toString() + ", IM bridge service will be restarted.", ex);
                watcher.signalRestartNeeded();
                if (handler != null) {
                    handler.cancel(false);
                }
            }
        }

        public synchronized void setHandler(ScheduledFuture handler) {
            this.handler = handler;
        }
    }
}
