/*
 * Copyright 2015 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.rpcserver;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.Arrays;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Map;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServlet;

/**
 *
 * @author tskoda
 */
public class RpcServlet extends HttpServlet {

    private static final Logger LOG = Logger.getLogger(RpcServlet.class.getName());
    private static final int UNKNOWN_ID = -1;
    
    public static final String SESSION_ID_COOKIE_NAME = "SESSIONID";

    private final Map<String, RpcMethod> handlers;
    private final ObjectMapper mapper;
    
    private final SessionManager sessionManager;

    public RpcServlet(ObjectMapper json, Map<String, RpcMethod> handlers) {
        this.mapper = json;
        this.handlers = handlers;
        this.sessionManager = new SessionManager();
    }

    @Override
    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        this.doGet(req, resp);
    }

    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        try {
            String charEnc = req.getCharacterEncoding() == null ? "UTF-8": req.getCharacterEncoding();
            java.util.Scanner s = new java.util.Scanner(new InputStreamReader(req.getInputStream(), charEnc)).useDelimiter("\\A");
            String cc = s.hasNext() ? s.next() : "";
            
            RequestEnvelope content = this.mapper.readValue(cc, RequestEnvelope.class);
            try {
                RpcMethod handler = this.handlers.get(content.getMethod());
                if (handler != null) {
                    try {
                        handler.execute(rpcRequest(req, content), rpcResponse(resp, content.getId()), sessionManager);
                    } catch (NotConfiguredException | UnsupportedOperationException e) {
                        LOG.log(Level.WARNING, "Failed to process rpc request, service is waiting for configuration", e);
                        rpcResponse(resp, content.getId()).error(new RpcError(-1, "Service is waiting for configuration"), HttpServletResponse.SC_SERVICE_UNAVAILABLE);
                    }
                } else {
                    LOG.log(Level.FINE, "Failed to process rpc request, no handler for method \"{0}\"", content.getMethod());
                    LOG.log(Level.FINE, "Called method \"{0}\" with data: {1}", new Object[]{content.getMethod(), content.getParams().toString()});
                    rpcResponse(resp, content.getId()).error(new RpcError(-1, "Method " + content.getMethod() + " not found"), HttpServletResponse.SC_NOT_FOUND);
                }
            } catch (IllegalArgumentException e) {
                LOG.log(Level.FINE, "Failed to process rpc request, incorrect JSON format", e);
                rpcResponse(resp, content.getId()).error(new RpcError(-1, "Failed to process rpc request, incorrect JSON format."), HttpServletResponse.SC_BAD_REQUEST);
            }
        } catch (JsonProcessingException e) {
            LOG.log(Level.FINE, "Failed to process rpc request, incorrect JSON format", e);
            rpcResponse(resp, UNKNOWN_ID).error(new RpcError(-1, "Failed to process rpc request, incorrect JSON format."), HttpServletResponse.SC_BAD_REQUEST);
        } catch (Exception e) {
            LOG.log(Level.SEVERE, "Failed to process rpc request, internal server error", e);
            rpcResponse(resp, UNKNOWN_ID).error(new RpcError(-1, "Failed to process rpc request, internal server error."), HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        }
    }

    private RpcRequestImpl rpcRequest(HttpServletRequest req, RequestEnvelope content) {
        String sessionID = null;
        if (req.getCookies() != null) {
            Optional<Cookie> c = Arrays.asList(req.getCookies()).stream().filter((t) -> t.getName().equals(SESSION_ID_COOKIE_NAME)).findFirst();
            sessionID = c.isPresent() ? c.get().getValue() : null;            
        }        
        return new RpcRequestImpl(content, req, this.mapper, sessionID, req.getHeader("X-Token"));
    }

    private RpcResponseImpl rpcResponse(HttpServletResponse resp, int id) {
        return new RpcResponseImpl(resp, id, this.mapper);
    }
}
