/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase;

import com.kerio.im.core.api.service.AuthService;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.server.xmppsession.SessionManager;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.NoConnectionIdException;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.XMPPSession;

/**
 *
 * @author tskoda
 */
public class SessionProvider {

    private static final Logger LOG = Logger.getLogger(SessionProvider.class.getName());
    private static final byte[] NO_PASS = null;
    private final SessionManager sessionManagerComponent;
    private final AuthService authService;

    public SessionProvider(SessionManager sessionManagerComponent, AuthService authService) {
        this.sessionManagerComponent = sessionManagerComponent;
        this.authService = authService;
    }

    public List<UserResource> getUserResources(BareJID jid) throws NotAuthorizedException {
        List<XMPPResourceConnection> activeResources = getActiveResources(jid);
        List<UserResource> result = new ArrayList<UserResource>();

        for (XMPPResourceConnection con : activeResources) {
            result.add(new UserResourceImpl(con));
        }

        return result;
    }

    public UserData getUserData(BareJID jid) {
        return new UserDataImpl(getParentSession(jid));
    }

    private List<XMPPResourceConnection> getActiveResources(BareJID jid) throws NotAuthorizedException {
        XMPPSession connection = getParentSession(jid);
        List<XMPPResourceConnection> result;

        if (connection != null) {
            result = connection.getActiveResources();
        } else {
            LOG.log(Level.FINEST, "No session for {0} found", jid);
            result = Collections.EMPTY_LIST;
        }

        return result;
    }

    
    public void disconnect(BareJID jid) throws NotAuthorizedException {

        for (XMPPResourceConnection connection : getActiveResources(jid)) {
            try {
                try {
                    AuthService.Credentials credentials  = new AuthService.Credentials(
                        jid.getLocalpart(),
                        jid.getDomain(),
                        NO_PASS);
                    this.authService.disconnected(credentials);
                } finally {
                    LOG.log(Level.FINEST, "Disconnecting {0}, session {1}", new Object[]{jid, connection.getSessionId()});
                    connection.logout(); 
                }
            } catch (NotAuthorizedException e) {
                //connection is not authorized yet, cant disconnect
                LOG.log(Level.INFO, "Disconnecting " + jid.toString() + " failed, user not authorized yet", e);
            }
        }
    }
    
    public boolean isOnline(BareJID jid) throws NotAuthorizedException {
        boolean result = false;
        XMPPSession connection = getParentSession(jid);

        if (connection != null) {
            List<XMPPResourceConnection> activeSessions = connection.getActiveResources();
            if ((activeSessions != null) && (activeSessions.size() > 0)) {
                LOG.log(Level.FINEST, "User {0} is online", jid);
                result = true;
            } else {
                LOG.log(Level.FINEST, "User {0} is NOT online", jid);
            }
        } else {
            LOG.log(Level.FINEST, "User {0} is NOT online", jid);
        }

        return result;
    }

    public XMPPSession getParentSession(BareJID jid) {
        return this.sessionManagerComponent.getSession(jid);
    }
    
    public interface UserResource {

        boolean isAuthorized();

        String nextStanzaId();

        BareJID getBareJID() throws NotAuthorizedException;

        JID getJID() throws NotAuthorizedException;

        JID getConnectionId() throws NoConnectionIdException;
        
        XMPPResourceConnection raw();
    }

    public interface UserData {

        Object getCommonSessionData(String key);

        void putCommonSessionData(String key, Object value);

        Object removeCommonSessionData(String key);

    }

    public class UserDataImpl implements UserData {

        private final XMPPSession session;

        public UserDataImpl(XMPPSession session) {
            this.session = session;
        }

        @Override
        public Object getCommonSessionData(String key) {
            return this.session.getCommonSessionData(key);
        }

        @Override
        public void putCommonSessionData(String key, Object value) {
            this.session.putCommonSessionData(key, value);
        }

        @Override
        public Object removeCommonSessionData(String key) {
            return this.session.removeCommonSessionData(key);
        }
    }

    public class UserResourceImpl implements UserResource {

        private final XMPPResourceConnection connection;

        public UserResourceImpl(XMPPResourceConnection connection) {
            this.connection = connection;
        }

        @Override
        public boolean isAuthorized() {
            return this.connection.isAuthorized();
        }

        @Override
        public String nextStanzaId() {
            return this.connection.nextStanzaId();
        }

        @Override
        public BareJID getBareJID() throws NotAuthorizedException {
            return this.connection.getBareJID();
        }

        @Override
        public JID getJID() throws NotAuthorizedException {
            return this.connection.getJID();
        }

        @Override
        public JID getConnectionId() throws NoConnectionIdException {
            return this.connection.getConnectionId();
        }
        
        @Override
        public XMPPResourceConnection raw() {
            return this.connection;
        }
    }
}
