/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.auth;

import com.kerio.im.core.api.service.AuthService;
import com.kerio.im.core.api.service.UserService;
import com.kerio.im.tigase.ClientResolver;
import com.kerio.im.tigase.bridge.xmpp.IAuthServiceProvider;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.xmpp.IXmppConnectionManager;
import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.sasl.Sasl;
import javax.security.sasl.SaslException;
import javax.security.sasl.SaslServer;
import tigase.db.AuthRepository;
import tigase.db.AuthorizationException;
import tigase.db.DBInitException;
import tigase.db.TigaseDBException;
import tigase.db.UserExistsException;
import tigase.db.UserNotFoundException;
import tigase.util.Base64;
import tigase.util.TigaseStringprepException;
import tigase.xmpp.BareJID;
import tigase.xmpp.XMPPResourceConnection;

/**
 *
 * @author tskoda
 */
public class ConnectAuthenticator implements AuthRepository {

    private static final Logger LOG = Logger.getLogger(ConnectAuthenticator.class.getName());
    //private static final String[] non_sasl_mechs = {"password"};
    private static final String[] sasl_mechs = {"PLAIN"};
    private UserService userService;
    private AuthService authService;
    private ClientResolver clientResolver;

    public ConnectAuthenticator() {
    }

    @Override
    public void initRepository(String resource_uri, Map<String, String> params) throws DBInitException {   
    }

    public void setAuthService(AuthService authService) {
        this.authService = authService;
    }
    
    public void setClientSession(ClientResolver clientResolver) {
        this.clientResolver = clientResolver;
    }
    
    @Override
    public String getResourceUri() {
        return "connection/im/core";
    }

    @Override
    public long getUsersCount() {
        long result = 0;

//        if (userService != null) {
//            result = userService.usersCount();
//        }

        return result;
    }

    @Override
    public long getUsersCount(String domain) {
        long result = 0;

        if (userService != null) {
            result = userService.usersCount(domain);
        }

        return result;
    }

    @Override
    @Deprecated
    public boolean plainAuth(BareJID user, String password) throws UserNotFoundException, TigaseDBException, AuthorizationException {
        throw new AuthorizationException("Plain authentication is not supported");
    }

    @Override
    @Deprecated
    public boolean digestAuth(BareJID user, String digest, String id, String alg) throws UserNotFoundException, TigaseDBException, AuthorizationException {
        //LOG.log(Level.INFO, "Digest auth: {0}, {1}, {3}", new Object[]{user, digest, id, alg});
        throw new AuthorizationException("Digest authentication is not supported");
        /* return true; */
    }

    private SaslServer getSaslServer(Map<String, Object> authProps) throws SaslException {
        SaslServer ss = (SaslServer) authProps.get("SaslServer");

        if (ss == null) {
            Map<String, String> saslProps = new TreeMap<String, String>();

            saslProps.put(Sasl.QOP, "auth");
            ss = Sasl.createSaslServer(
                    (String) authProps.get(MACHANISM_KEY), "xmpp",
                    (String) authProps.get(SERVER_NAME_KEY),
                    saslProps,
                    new ConnectCallbackHandler(authProps));
            authProps.put("SaslServer", ss);
        }

        return ss;
    }

    private boolean saslAuth(Map<String, Object> authProps) throws UserNotFoundException, TigaseDBException, AuthorizationException {
        LOG.log(Level.FINER, "Performing SASL authorization");
        try {
            SaslServer saslServer = getSaslServer(authProps);
            String authDataB64 = (String) authProps.get(DATA_KEY);
            byte[] authData = ((authDataB64 != null) ? Base64.decode(authDataB64) : new byte[0]);
            byte[] challenge = saslServer.evaluateResponse(authData);
            String challengeB64 = (((challenge != null) && (challenge.length > 0)) ? Base64.encode(challenge) : null);

            authProps.put(RESULT_KEY, challengeB64);

            return saslServer.isComplete();

        } catch (SaslException e) {
            throw new AuthorizationException("Sasl exception.", e);
        }
    }

    @Override
    public boolean otherAuth(Map<String, Object> authProps) throws UserNotFoundException, TigaseDBException, AuthorizationException {
        boolean result = false;
        String connectionId = (String) authProps.get(XMPPResourceConnection.CONNECTION_RESOURCE_ID);
        IAuthServiceProvider fakeConnectionsManager = clientResolver.getLookup().lookup(IMBridge.class).getConnectionManager();
        if (fakeConnectionsManager != null && fakeConnectionsManager.getAuthServiceForConnection(connectionId) != null) {
            try {
                authProps.put(USER_ID_KEY, BareJID.bareJIDInstance((String) authProps.get(DATA_KEY)));
                return true;
            } catch (TigaseStringprepException ex) {
                            LOG.log(Level.SEVERE, "Error occured while getting BareJID", ex);
                return false;
            }
        }
        String proto = (String) authProps.get(PROTOCOL_KEY);

        if (proto.equals(PROTOCOL_VAL_SASL)) {
            String mech = (String) authProps.get(MACHANISM_KEY);

            if (mech.equals("PLAIN")) {
                result = saslAuth(authProps);
            } else {
                throw new AuthorizationException("Mechanism is not supported: " + mech);
            }
        } else {
            throw new AuthorizationException("Protocol is not supported: " + proto);
        }

        return result;
    }

    @Override
    public void queryAuth(Map<String, Object> authProps) {

        String protocol = (String) authProps.get(PROTOCOL_KEY);

        if (protocol.equals(PROTOCOL_VAL_SASL)) {
            authProps.put(RESULT_KEY, sasl_mechs);
        } else {
            authProps.put(RESULT_KEY, new String[0]);
        }
    }

    @Override
    public void logout(BareJID user) throws UserNotFoundException, TigaseDBException {
        //throw new TigaseDBException("Not supported.");
    }

    /* these methods will not be implemented */
    @Override
    public void addUser(BareJID user, String password) throws UserExistsException, TigaseDBException {
        throw new TigaseDBException("Not supported.");
    }

    @Override
    public void removeUser(BareJID user) throws UserNotFoundException, TigaseDBException {
        throw new TigaseDBException("Not supported.");
    }

    @Override
    public void updatePassword(BareJID user, String password) throws UserNotFoundException, TigaseDBException {
        throw new TigaseDBException("Not supported.");
    }

    private class ConnectCallbackHandler implements CallbackHandler {

        private Map<String, Object> authProps;

        ConnectCallbackHandler(Map<String, Object> authProps) {
            this.authProps = authProps;
        }

        @Override
        public void handle(Callback[] clbcks) throws IOException, UnsupportedCallbackException {
            if (clbcks == null) {
                return;
            }

            for (Callback callback : clbcks) {
                if (callback instanceof ConnectCallback) {
                    ConnectCallback cc = (ConnectCallback) callback;

                    String userName = cc.getUserName().trim();
                    String domain = (String) authProps.get(REALM_KEY);
                    byte[] password = cc.getPassword();
                    BareJID userBareJID = BareJID.bareJIDInstanceNS(userName, domain);

                    String connectionId = (String) authProps.get(XMPPResourceConnection.CONNECTION_RESOURCE_ID);
                    
                    AuthService.Credentials credentials = new AuthService.Credentials(
                            userName, 
                            domain, 
                            password);
              

                    boolean result = authService.auth(credentials, clientResolver.resolve(connectionId));

                    if (result) {
                        authProps.put(USER_ID_KEY, userBareJID);
                    }

                    cc.setAuthenticated(result);
                } else {
                    throw new UnsupportedCallbackException(callback);
                }
            }
        }
    }
}
