/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.IM;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.kerio.im.common.Lookup;
import com.kerio.im.external.IMContactID;
import com.kerio.im.external.IMEvent;
import com.kerio.im.external.IMPresence;
import com.kerio.im.external.ISingnalHandler;
import com.kerio.im.external.ImService;
import com.kerio.im.external.ImServiceFactory;
import com.kerio.im.methods.Login;
import com.kerio.im.methods.Logout;
import com.kerio.im.methods.Publish;
import com.kerio.im.methods.WhoAmI;
import com.kerio.im.rpcserver.Config;
import com.kerio.im.rpcserver.JettyServer;
import com.kerio.im.tigase.bridge.IIMSynchronizationService;
import com.kerio.im.tigase.bridge.IMSubscribctionsMamager;
import com.kerio.im.tigase.bridge.communicator.IMtoXmppChannel;
import com.kerio.im.tigase.bridge.conversation.ConversationManager;
import com.kerio.im.tigase.bridge.xmpp.EventQueue;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.xmpp.IMEventsProcessor;
import com.kerio.im.tigase.bridge.xmpp.IXmppConnection;
import com.kerio.im.tigase.bridge.xmpp.IXmppPacketProcessor;
import com.kerio.im.tigase.bridge.xmpp.XmppConnectionManager;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.kerio.im.tigase.bridge.xmpp.IXmppConnectionManager;
import com.kerio.im.tigase.bridge.xmpp.VariableSleeper;
import java.util.List;
import java.util.Map;

/**
 *
 * @author vbystricky
 */
public class IMSynchronizationService implements IXmppConnectionManager, IIMSynchronizationService {

    private static final Logger LOGGER = Logger.getLogger(IMSynchronizationService.class.getName());
    private final XmppConnectionManager xmppConnectionManager;
    private final ConversationManager conversationManager;
    private final IMEventsProcessor imEventsProcessor;
    private final XmppPacketsProcessor xmppPacketsProcessor;
    private final JettyServer server;
    private final ImService imService;
    private final IMConnectionManager imConnectioManager;
    private final IMSubscribctionsMamager subscribctionsMamager;
    private final IDomainInfoProvider domainInfoProvider;
    private final EventQueue<IMEvent> eventQueue;
    private List<UserPresence> state;

    public IMSynchronizationService(IDomainInfoProvider domainInfoProvider, XmppConnectionManager xmppConnectionManager, Map<String, String> config) {
        eventQueue = new EventQueue<>();
        ImServiceFactory factory = Lookup.service(ImServiceFactory.class);
        this.imService = factory.getService(null, config);
        this.xmppConnectionManager = xmppConnectionManager;
        this.conversationManager = new ConversationManager();
        this.subscribctionsMamager = new IMSubscribctionsMamager();
        this.domainInfoProvider = domainInfoProvider;

        this.imEventsProcessor = new IMEventsProcessor(this, eventQueue);
        this.imConnectioManager = new IMConnectionManager(imService, eventQueue, subscribctionsMamager);
        this.xmppPacketsProcessor = new XmppPacketsProcessor(this);
        ObjectMapper mapper = new ObjectMapper();
        server = new JettyServer(mapper, new Publish(mapper, eventQueue, subscribctionsMamager), new Login(imService.getUserName(), imService.getPassword()), new Logout(), new WhoAmI());
        this.state = null;
    }

    private void startJetty() throws InterruptedException {
        boolean jettyStarted = false;
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Starting jetty.");
        VariableSleeper sleeper = new VariableSleeper();
        while (!jettyStarted) {
            try {
                server.start(new Config(4044, "0.0.0.0"));
                jettyStarted = true;
            } catch (Throwable ex) {
                LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Unable to start reverse channel web server.", ex);
                try {
                    server.shutdown();
                } catch (Exception ex1) {
                    LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Unable to stop reverse channel web server.", ex1);
                }
                sleeper.sleep();
            }
        }
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Jetty started.");
    }

    private void stopJetty() {
        try {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Stopping jetty.");
            server.shutdown();
        } catch (Exception ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Unable to stop reverse channel web server.", ex);
        }
    }

    private void startIMService(ISingnalHandler watcher) throws InterruptedException {
        boolean imServiceStarted = false;
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Starting IM service.");
        VariableSleeper sleeper = new VariableSleeper();
        while (!imServiceStarted) {
            try {
                eventQueue.lock();
                imService.register(watcher);
                Integer requestId = imService.generateRequestId();
                IMContactID imadminCid = new IMContactID(".imadmin", "");
                subscribctionsMamager.addSubscription(requestId, IMPresence.class, imadminCid);
                List<IMPresence> presenceList = imService.subscribeGlobalPresence(requestId).getResponse();
                for (IMPresence presence : presenceList) {
                    if (presence.getStatus() != IMPresence.Status.offline && presence.getStatus() != IMPresence.Status.invisible) {
                        eventQueue.enqueueUnchecked(new IMEvent(presence, imadminCid));
                    }
                }
                imServiceStarted = true;
            } catch (RuntimeException ex) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Starting IM service failed.", ex);
                subscribctionsMamager.clear();
                imService.unregister();
                sleeper.sleep();
            } finally {
                eventQueue.unlock();
            }
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "IM service started.");
        }
    }

    public void start(ISingnalHandler watcher) throws InterruptedException {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Starting IMSynchronizationService.");
        startJetty();
        startIMService(watcher);
        applyState();
        imEventsProcessor.start(watcher);
    }

    public void restart(ISingnalHandler watcher) throws InterruptedException {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Restarting IMSynchronizationService.");
        clear();
        start(watcher);
    }

    private String stateToString(List<UserPresence> state) {
        String result = "";
        result = state.stream().map((tmp) -> tmp.toString() + System.lineSeparator()).reduce(result, String::concat);
        return result;
    }

    private void applyState() throws InterruptedException {
        if (state == null) {
            return;
        }

        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Applying stored presence state: {0}", stateToString(state));
        for (UserPresence tmp : state) {
            IMConnection connection = imConnectioManager.createConnection(tmp.getUserID());
            for (IMConnection.ClientStatus clientStatus : tmp.getClientStatuses()) {
                connection.setPresence(clientStatus.getStatus(), clientStatus.getClientID());
            }
        }
        state = null;
    }

    private void clear() {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Reseting IMSynchronizationService state.");
        imEventsProcessor.stop();
        xmppConnectionManager.destroyAllConnections();
        state = imConnectioManager.getUserPresences();
        imConnectioManager.destroyAllConnections(true);
        imService.unregister();
        stopJetty();
        subscribctionsMamager.clear();
        conversationManager.clear();
        eventQueue.clear();
    }

    public void stop() {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Stopping IMSynchronizationService.");
        try {
            imConnectioManager.destroyAllConnections(false);
        } catch (RuntimeException e) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Error when closing im connections", e);
        }
        imEventsProcessor.stop();
        xmppConnectionManager.destroyAllConnections();
        imService.unregister();
        state = null;
        stopJetty();
    }

    @Override
    public void closeXmppConnection(String connectionId) {
        IXmppConnection connecetion = xmppConnectionManager.getConnectionById(connectionId);
        if (connecetion != null) {
            connecetion.close();
            xmppConnectionManager.destroyConnection(connecetion);
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Connection {0} closed.", connectionId);
        } else {
            LOGGER.log(Level.WARNING, IMBridge.LOG_PREFIX + "Connection {0} does not exist.", connectionId);
        }
    }

    @Override
    public com.kerio.im.tigase.bridge.IM.IMConnectionManager IMConnectionManager() {
        return imConnectioManager;
    }

    @Override
    public XmppConnectionManager XmppConnectionManager() {
        return xmppConnectionManager;
    }

    @Override
    public ConversationManager ConversationManager() {
        return conversationManager;
    }

    @Override
    public IDomainInfoProvider DomainInfoProvider() {
        return domainInfoProvider;
    }

    @Override
    public IMSubscribctionsMamager IMSubscribctionsMamager() {
        return subscribctionsMamager;
    }

    public IXmppPacketProcessor IXmppPacketProcessor() {
        return xmppPacketsProcessor;
    }
}
