/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.IM;

import com.kerio.im.common.TypedProperties;
import com.kerio.im.connect.ConnectKey;
import com.kerio.im.tigase.bridge.communicator.IMtoXmppChannel;
import com.kerio.im.tigase.bridge.xmpp.IAuthServiceProvider;
import com.kerio.im.tigase.bridge.xmpp.XmppConnectionManager;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author vbystricky
 */
public class IMSynchronizationServiceManager {

    private static final Logger LOGGER = Logger.getLogger(IMSynchronizationServiceManager.class.getName());
    private static final Boolean ENABLED_DEFAULT = true;
    private IMtoXmppChannel communicator;
    private IDomainInfoProvider domainInfoProvider;
    private IMSynchronizationService service;
    private XmppConnectionManager xmppConnectionManager;
    private boolean initilized = false;
    private IMSynchronizationWatcher watcher = new IMSynchronizationWatcher();
    private Map<String, String> config;
    private boolean isImEnabled = true;

    public synchronized void initialize(IMtoXmppChannel communicator, IDomainInfoProvider domainInfoProvider, String SESSION_MANAGER_ID) {
        this.communicator = communicator;
        this.domainInfoProvider = domainInfoProvider;
        this.xmppConnectionManager = new XmppConnectionManager(communicator, SESSION_MANAGER_ID);
        initilized = true;
    }

    public synchronized IMSynchronizationService service() throws DisabledIMException {
        if (!isImEnabled) {
            throw new DisabledIMException();
        }
        if (service != null && watcher.isRestartNeeded()) {
            restart();
        }
        return service;
    }

    public synchronized void start(Map<String, String> config) throws IllegalStateException, DisabledIMException {
        if (!initilized) {
            throw new IllegalStateException("IM Service can't be started before it was initilized.");
        }
        if (config == null) {
            throw new IllegalStateException("IM Service can't be started without configuration.");
        }
        if (service != null) {
            throw new IllegalStateException("IM service is already started.");
        }
        this.isImEnabled = TypedProperties.bool(ConnectKey.IMAPI_ENABLED, config, ENABLED_DEFAULT);
        if (!isImEnabled) {
            throw new DisabledIMException();
        }
        this.config = config;
        start();
    }

    private void start() throws IllegalStateException, DisabledIMException {
        try {
            service = new IMSynchronizationService(domainInfoProvider, xmppConnectionManager, config);
            watcher = new IMSynchronizationWatcher();
            service.start(watcher);
        } catch (InterruptedException ex) {
            LOGGER.log(Level.SEVERE, null, ex);
            if (service != null) {
                stop();
            }
        }
    }

    public synchronized void stop() throws IllegalStateException, DisabledIMException {
        if (!isImEnabled) {
            throw new DisabledIMException();
        }
        if (service == null) {
            throw new IllegalStateException("IM service is already stopped.");
        }
        service.stop();
        service = null;
    }

    public synchronized void restart() throws IllegalStateException, DisabledIMException {
        if (!isImEnabled) {
            throw new DisabledIMException();
        }
        try {
            if (!initilized) {
                throw new IllegalStateException("IM Service can't be restarted before it was initilized.");
            }
            if (config == null) {
                throw new IllegalStateException("IM Service can't be restarted without configuration.");
            }
            if (service == null) {
                start();
            } else {
                watcher = new IMSynchronizationWatcher();
                service.restart(watcher);
            }
        } catch (InterruptedException ex) {
            LOGGER.log(Level.SEVERE, null, ex);
        }
    }

    public synchronized IMSynchronizationWatcher getWatcher() {
        return watcher;
    }

    public IAuthServiceProvider authServiceProvider() {
        return xmppConnectionManager;
    }
}
