/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.IM;

import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.xmpp.XmppIdGenerator;
import java.util.logging.Level;
import java.util.logging.Logger;
import static tigase.conf.Configurable.DEF_BOSH_NAME;
import static tigase.conf.Configurable.DEF_C2S_NAME;
import tigase.server.Packet;
import tigase.xml.Element;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;

/**
 *
 * @author vbystricky
 */
public class PacketClassifier {

    private static final Logger LOGGER = Logger.getLogger(PacketClassifier.class.getName());
    private final IDomainInfoProvider domainInfo;

    private String getPacketID(Packet packet) {
        return (packet.getStanzaId() == null) ? "" : packet.getStanzaId();
    }

    public PacketClassifier(IDomainInfoProvider domainInfo) {
        this.domainInfo = domainInfo;
    }

    private boolean isLoopDetected(Packet packet) {
        if (packet.getStanzaFrom().getResource() != null) {
            if (XmppIdGenerator.isResourceIdInRightFormat(packet.getStanzaFrom().getResource())) {
                return true;
            }
        }
        if (packet.getStanzaId() != null) {
            if (XmppIdGenerator.isPacketIdInRightFormat(packet.getStanzaId())) {
                return true;
            }
        }
        return false;
    }

    public boolean isSupportedChatMessage(Packet packet) {
        JID sender = packet.getStanzaFrom();
        JID recipient = packet.getStanzaTo();
        if (sender == null || sender.getBareJID() == null || sender.getLocalpart() == null || sender.getDomain() == null) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Chat packet with id {0} has empty sender.", getPacketID(packet));
            return false;
        }
        if (recipient == null || recipient.getBareJID() == null || recipient.getLocalpart() == null || recipient.getDomain() == null) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Chat packet with id {0} has empty recipient.", getPacketID(packet));
            return false;
        }

        if (!domainInfo.isLocalDomain(sender.getDomain())) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Chat packet with id {0} has external sender {1}", new Object[]{getPacketID(packet), sender});
            return false;
        }
        if (!domainInfo.isLocalDomain(recipient.getDomain())) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Chat packet with id {0} has external recipient {1}", new Object[]{getPacketID(packet), recipient});
            return false;
        }

        if (isLoopDetected(packet)) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Loop detected. Chat packet with id {0} origins from IM sender {1}", new Object[]{getPacketID(packet), sender});
            return false;
        }
        return true;
    }

    public boolean isSupportedPresencePacket(Packet packet) {
        if (packet.getStanzaFrom() == null || packet.getStanzaFrom().getBareJID() == null || packet.getStanzaFrom().getLocalpart() == null || packet.getStanzaFrom().getDomain() == null) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Presence packet with id {0} has empty sender.", getPacketID(packet));
            return false;
        }
        if (!domainInfo.isLocalDomain(packet.getStanzaFrom().getDomain())) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Presence packet with id {0} has external sender {1}", new Object[]{getPacketID(packet), packet.getStanzaFrom()});
            return false;
        }

        if (isLoopDetected(packet)) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Loop detected. Presence packet with id {0} origins from IM sender {1}", new Object[]{getPacketID(packet), packet.getStanzaFrom()});
            return false;
        }
        return true;
    }

    public boolean hasMessage(Packet packet) {
        return packet.getElement().findChild("message/body") != null;
    }

    public boolean isHeadingToResource(Packet packet) {
        JID to = packet.getStanzaTo();
        if (to == null) {
            return false;
        }
        return to.getResource() != null && !to.getResource().isEmpty();
    }

    public boolean isPresencePacket(Packet packet) {
        if (packet != null && (packet.getType() == null || packet.getType() == StanzaType.available || packet.getType() == StanzaType.unavailable)) {
            return packet.getElemName() != null && packet.getElemName().equalsIgnoreCase("presence");
        } else {
            return false;
        }
    }

    public boolean isHeadingToIMBridge(Packet packet) {
        if (packet.getPacketTo() == null || packet.getPacketTo().getLocalpart() == null) {
            return false;
        }
        return packet.getPacketTo().getLocalpart().equals(IMBridge.IM_BRIDGE_NAME);
    }

    public boolean hasNonNullMessageBody(Packet packet) {
        return packet.getElemCData("message/body") != null;
    }

    public boolean isPacketHeadingToClient(Packet packet) {
        if (packet.getPacketTo() == null || packet.getPacketTo().getLocalpart() == null) {
            return false;
        }
        return packet.getPacketTo().getLocalpart().equals(DEF_C2S_NAME);
    }

    public boolean isPacketDeliveredOnlyToIM(Packet packet) {
        return packet.getAttribute(Element.RECIEVED_BY_MULTIPLE_CLIENTS) == null && packet.getAttribute(Element.IM_BRIDGE_COPY) == null;
    }
}
