/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge;

import com.kerio.im.common.TypedProperties;
import com.kerio.im.connect.ConnectKey;
import com.kerio.im.core.api.event.StartupFinishedEvent;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.IM.PacketClassifier;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Map;
import java.util.Queue;
import java.util.TreeMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.NonAuthUserRepository;
import tigase.db.TigaseDBException;
import tigase.server.Packet;
import tigase.util.TigaseStringprepException;
import tigase.xml.Element;
import static tigase.xml.Element.PROCESSED_BY_IM_BRIDGE_SYNC;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPPacketFilterIfc;
import tigase.xmpp.XMPPProcessor;
import tigase.xmpp.XMPPResourceConnection;
import static tigase.xml.Element.RECIEVED_BY_MULTIPLE_CLIENTS;

/**
 *
 * @author vbystricky
 */
public class IMBridgePlugin extends XMPPProcessor implements XMPPPacketFilterIfc {

    public static final String LOG_PREFIX = "IM Bridge Plugin: ";
    private static final Logger LOGGER = Logger.getLogger(IMBridgePlugin.class.getName());
    public static final String ID = "im-bridge-synchronizator";
    private final AtomicBoolean enabled = new AtomicBoolean(true);

    @Override
    public String id() {
        return ID;
    }

    @Override
    public void init(Map<String, Object> settings) throws TigaseDBException {
        super.init(settings);
        ServiceBus bus = ((ServiceBus) settings.get(ServiceBus.class.getName()));
        if (bus != null) {
            bus.subscribe(StartupFinishedEvent.class, new StartupListener(this));
        }
    }

    public void disable() {
        enabled.set(false);
    }

    private boolean isDomainLocal(String domain, XMPPResourceConnection session) {
        if (domain.equals("kerio-connect.private")) {	//ignore special kerio-connect.private domain used to comunicate with Kerio Connet
            return false;
        }
        return session.isLocalDomain(domain, false);
    }

    private Packet getPacketoForIMGateway(final Packet originalPacket, XMPPResourceConnection session, boolean removeStanzaTo) throws TigaseStringprepException {
        if (session.getSMComponentId() == null || session.getSMComponentId().getDomain() == null) {
            LOGGER.log(Level.WARNING, LOG_PREFIX + "Null session manager component ID");
            return null;
        }
        String domain = session.getSMComponentId().getDomain();
        Packet result = Packet.packetInstance(originalPacket.getElement().clone(), originalPacket.getStanzaFrom(), removeStanzaTo ? null : originalPacket.getStanzaTo());
        result.setPacketTo(JID.jidInstance(BareJID.bareJIDInstance(IMBridge.IM_BRIDGE_NAME, domain)));
        result.setAttribute(Element.IM_BRIDGE_COPY, "1");
        LOGGER.log(Level.FINE, LOG_PREFIX + "Created packet copy {0}", result.toStringFull());
        return result;
    }

    private boolean wasAlreadySynchronized(Packet packet) {
        return packet.getAttribute(PROCESSED_BY_IM_BRIDGE_SYNC) != null;
    }

    private String computePacketHash(Packet packet) {
        String data;
        data = packet.getStanzaId();
        if (data != null) {
            return data;
        }
        data = packet.getElemCData("message/body");
        if (data == null) {
            return "";
        }

        MessageDigest md;
        try {
            md = MessageDigest.getInstance("SHA-256");
        } catch (NoSuchAlgorithmException ex) {
            return "";
        }
        return new String(md.digest(data.getBytes()), StandardCharsets.UTF_8);
    }

    void addPacketForIMGateway(ArrayList<Packet> samePackets, ArrayList<Packet> result, XMPPResourceConnection session, boolean removeStanzaTo, PacketClassifier classifier) {
        Packet packetToIMBridge = samePackets.stream().filter(outPacket -> classifier.isHeadingToIMBridge(outPacket)).findFirst().orElse(null);
        if (packetToIMBridge != null) {
            LOGGER.log(Level.FINE, LOG_PREFIX + "Found chat packet already heading to IM Bridge");
            if (samePackets.size() > 1) {
                packetToIMBridge.setAttribute(RECIEVED_BY_MULTIPLE_CLIENTS, "1");
                LOGGER.log(Level.FINE, LOG_PREFIX + "Found packets heading to multiple clients including IM Bridge.");
            }
        } else {
            try {
                Packet tmp = getPacketoForIMGateway(samePackets.get(0), session, removeStanzaTo);
                if (tmp != null) {
                    result.add(tmp);
                }
            } catch (TigaseStringprepException ex) {
                Logger.getLogger(IMBridgePlugin.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    private ArrayList<Packet> createIMPackets(ArrayList<Packet> packets, XMPPResourceConnection session, PacketClassifier classifier) {
        ArrayList<Packet> result = new ArrayList<>();
        if (packets.size() == 1) {
            addPacketForIMGateway(packets, result, session, false, classifier);
            return result;
        }

        TreeMap<String, ArrayList<Packet>> categorizedPackets = new TreeMap<>();
        packets.stream().forEach((outputPacket) -> {
            String key = computePacketHash(outputPacket);
            ArrayList<Packet> tmp = categorizedPackets.get(key);
            if (tmp == null) {
                tmp = new ArrayList<>();
                categorizedPackets.put(key, tmp);
            }
            tmp.add(outputPacket);
        });
        categorizedPackets.values().stream().forEach((samePackets) -> {
            addPacketForIMGateway(samePackets, result, session, false, classifier);
        });
        return result;
    }

    private String getPresence(Packet packet) {
        if (packet.getType() != null && packet.getType() == StanzaType.unavailable) {
            return "offline";
        }

        String presence = packet.getElemCData("presence/show");
        if (presence == null) {
            if (packet.getType() == null || packet.getType() == StanzaType.available) {
                presence = "available";
            }
        }
        LOGGER.log(Level.FINE, LOG_PREFIX + "Found presence status {0}.", presence);
        return presence;
    }

    private Packet processPressencePacket(Packet packet, XMPPResourceConnection session) {
        if (getPresence(packet) != null) {
            try {
                return getPacketoForIMGateway(packet, session, true);
            } catch (TigaseStringprepException ex) {
                Logger.getLogger(IMBridgePlugin.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return null;
    }

    private void markAllResultsAsProcessed(Queue<Packet> results) {
        results.stream().forEach((outputPacket) -> {
            outputPacket.setAttribute(PROCESSED_BY_IM_BRIDGE_SYNC, "1");
        });
    }

    @Override
    public void filter(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results) {
        if (!enabled.get()) {
            return;
        }
        if (session == null) {
            return;
        }
        PacketClassifier classifier = new PacketClassifier((String domain) -> isDomainLocal(domain, session));
        LOGGER.log(Level.FINE, LOG_PREFIX + "Processing packet {0} expanded in {1} output packets.", new Object[]{packet, results.size()});
        ArrayList<Packet> synchronizationCandidates = new ArrayList<>();
        results.stream().filter((outputPacket) -> (outputPacket.getType() == StanzaType.chat
                && (classifier.isPacketHeadingToClient(outputPacket) || classifier.isHeadingToIMBridge(outputPacket))
                && classifier.isSupportedChatMessage(outputPacket)
                && classifier.hasNonNullMessageBody(outputPacket)
                && !wasAlreadySynchronized(outputPacket))).forEach((supportedChatPacket) -> {
            LOGGER.log(Level.FINE, LOG_PREFIX + "Found chat packet to sync {0}.", supportedChatPacket);
            synchronizationCandidates.add(supportedChatPacket);
            supportedChatPacket.setAttribute(PROCESSED_BY_IM_BRIDGE_SYNC, "1");
        });

        ArrayList<Packet> packetsToIMgateway = createIMPackets(synchronizationCandidates, session, classifier);
        packetsToIMgateway.stream().forEach((IMGatewayPacket) -> {
            results.add(IMGatewayPacket);
        });

        Packet presenceResult = null;
        for (Packet outputPacket : results) {
            if (!classifier.isPresencePacket(outputPacket)) {
                continue;
            }
            if (wasAlreadySynchronized(outputPacket)) {
                continue;
            }
            if (classifier.isHeadingToResource(outputPacket)) {
                LOGGER.log(Level.INFO, "Skipping presence packet heading to resource {0}.", outputPacket.toStringFull());
                continue;
            }

            LOGGER.log(Level.FINE, LOG_PREFIX + "Presence packet {0}", outputPacket.toStringFull());
            if ((presenceResult = processPressencePacket(outputPacket, session)) != null) {
                break;
            }
        }

        if (presenceResult != null) {
            results.add(presenceResult);
            markAllResultsAsProcessed(results);
        }
    }

    private class StartupListener implements ServiceBus.EventListener<StartupFinishedEvent> {

        private final IMBridgePlugin plugin;

        public StartupListener(IMBridgePlugin plugin) {
            this.plugin = plugin;
        }

        @Override
        public void event(StartupFinishedEvent event) {
            Boolean enabled = TypedProperties.bool(ConnectKey.IMAPI_ENABLED, event.getContent(), true);
            if (enabled != null && enabled == false) {
                plugin.disable();
            }
        }
    }
}
