/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.xmpp;

import com.kerio.im.tigase.bridge.communicator.IMtoXmppChannel;
import com.kerio.im.tigase.bridge.communicator.PacketResponse;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.server.Command;
import tigase.server.Packet;
import tigase.util.TigaseStringprepException;
import tigase.xml.Element;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;

/**
 *
 * @author jzacek
 */
public class XmppConnection implements IXmppConnection {

    private static final Logger LOGGER = Logger.getLogger(XmppConnection.class.getName());
    public static final int TIMEOUT = 60 * 1000;

    private final IMtoXmppChannel communicator;
    private final String SESSION_MANAGER_ID;
    private final String user;
    private final String domain;
    private final String connectionID;
    private final String resource;

    public XmppConnection(IMtoXmppChannel communicator, String user, String domain, String SESSION_MANAGER_ID) {
        this.communicator = communicator;
        this.user = user;
        this.domain = domain;
        this.resource = XmppIdGenerator.generateResourceID();
        this.connectionID = XmppIdGenerator.generateConnectionID(user, domain);
        this.SESSION_MANAGER_ID = SESSION_MANAGER_ID;
    }

    private boolean elementAttributeEquals(Element element, String attributeName, String attributeValue) {
        if (element == null || element.getAttribute(attributeName) == null) {
            return false;
        }
        return element.getAttribute(attributeName).equalsIgnoreCase(attributeValue);
    }

    private boolean elementNameEquals(Element element, String name) {
        if (element == null || element.getName() == null) {
            return false;
        }
        return element.getName().equalsIgnoreCase(name);
    }

    private boolean jidEquals(JID jid1, JID jid2) {
        if (jid1 == null || jid2 == null) {
            return false;
        }
        return jid1.equals(jid2);
    }

    private boolean sendOpenStreamPacket() throws InterruptedException {
        IMtoXmppChannel.ID responseID = communicator.newID();
        String packetId = getPacketID();
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Sending open stream packet with ID {0}.", packetId);
        return communicator
                .addResponse(responseID, PacketResponse.init("Open Stream Response (ID: " + packetId + ")")
                        .add((p) -> jidEquals(p.getPacketFrom(), getSessionManagerJID()))
                        .add((p) -> jidEquals(p.getPacketTo(), getConnectionJID()))
                        .add((p) -> elementNameEquals(p.getElement(), "iq"))
                        .add((p) -> elementAttributeEquals(p.getElement(), "id", packetId))
                        .add((p) -> elementAttributeEquals(p.getElement(), "type", "result")))
                .sendMessage(getStreamOpenPacket(packetId))
                .waitForResponse(responseID, TIMEOUT);
    }

    private boolean sendAuthPacket() throws TigaseStringprepException, InterruptedException {
        IMtoXmppChannel.ID responseID = communicator.newID();
        String packetId = getPacketID();
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Sending auth packet with ID {0}.", packetId);
        return communicator
                .addResponse(responseID, PacketResponse.init("Auth Response (ID: " + responseID + ")")
                        .add((p) -> jidEquals(p.getPacketFrom(), getSessionManagerJID()))
                        .add((p) -> jidEquals(p.getPacketTo(), getConnectionJID()))
                        .add((p) -> elementNameEquals(p.getElement(), "success")))
                .sendMessage(getAuthPacket(packetId))
                .waitForResponse(responseID, TIMEOUT);
    }

    private boolean sendBindPacket() throws TigaseStringprepException, InterruptedException {
        IMtoXmppChannel.ID responseID = communicator.newID();
        String packetId = getPacketID();
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Sending resource bind packet with ID {0}.", packetId);
        return communicator
                .addResponse(responseID, PacketResponse.init("Respource Bind Response (ID: " + responseID + ")")
                        .add((p) -> jidEquals(p.getPacketFrom(), getSessionManagerJID()))
                        .add((p) -> jidEquals(p.getPacketTo(), getConnectionJID()))
                        .add((p) -> elementNameEquals(p.getElement(), "iq"))
                        .add((p) -> elementAttributeEquals(p.getElement(), "type", "result")))
                .sendMessage(getResourceBindPacket(packetId))
                .waitForResponse(responseID, TIMEOUT);
    }

    @Override
    public boolean open() throws TigaseStringprepException, InterruptedException {
        if (!sendOpenStreamPacket()) {
            return false;
        }
        if (!sendAuthPacket()) {
            return false;
        }
        if (!sendBindPacket()) {
            return false;
        }
        return true;
    }

    @Override
    public boolean setPresence(XmppPresence presence, String text) throws TigaseStringprepException {
        String packetId = getPacketID();
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Sending {0} presence packet with ID {1}.", new Object[]{presence, packetId});
        communicator
                .sendMessage(getPresencePacket(presence, text, packetId));
        return true;
    }

    @Override
    public boolean sendMessage(JID receiver, String text) throws TigaseStringprepException {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Sending message from {0} to {1}.", new Object[]{getUserJID(), receiver});
        communicator
                .sendMessage(PacketFactory.getMessagePacket(getUserJID(), receiver, text, getPacketID(), getConnectionJID(), getSessionManagerJID()));
        return true;
    }

    @Override
    public void close() {

        String packetId = getPacketID();
        Packet packet = Command.STREAM_CLOSED_UPDATE.getPacket(
                getConnectionJID(),
                getSessionManagerJID(),
                StanzaType.set, packetId
        );
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Sending close stream packet with ID {0} for stream.", new Object[]{packetId, getConnectionJID()});

        communicator
                .sendMessage(packet);
    }

    @Override
    public String getConnectionID() {
        return connectionID;
    }

    private JID getConnectionJID() {
        return JID.jidInstanceNS(communicator.getComponentId().getBareJID(), getConnectionID());
    }

    @Override
    public JID getUserJID() {
        JID res = null;
        try {
            res = JID.jidInstance(getUser(), getDomain(), getResource());
        } catch (TigaseStringprepException ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Unable to create User JID.", ex);
        }
        return res;
    }

    @Override
    public String getFullUserName() {
        return getUser() + "@" + getDomain();
    }

    @Override
    public String getUser() {
        return user;
    }

    @Override
    public String getDomain() {
        return domain;
    }

    @Override
    public String getResource() {
        return resource;
    }

    private String getPacketID() {
        return XmppIdGenerator.generatePacketID();
    }

    private JID getSessionManagerJID() {
        return JID.jidInstanceNS(SESSION_MANAGER_ID);
    }

    private Packet getStreamOpenPacket(String id) {
        Packet streamOpen = Command.STREAM_OPENED.getPacket(
                getConnectionJID(),
                getSessionManagerJID(),
                StanzaType.set,
                id,
                Command.DataType.submit
        );
        Command.addFieldValue(streamOpen, "session-id", getConnectionID());
        Command.addFieldValue(streamOpen, "hostname", getDomain());
        Command.addFieldValue(streamOpen, "xml:lang", "en");
        return streamOpen;
    }

    private Packet getAuthPacket(String id) throws TigaseStringprepException {
        Element elem = new Element("auth");
        elem.addAttribute("mechanism", "PLAIN");
        elem.addAttribute("id", id);
        elem.setXMLNS("urn:ietf:params:xml:ns:xmpp-sasl");
        elem.setCData(getFullUserName());

        Packet login = Packet.packetInstance(elem);
        login.setPacketFrom(getConnectionJID());
        login.setPacketTo(getSessionManagerJID());
        return login;
    }

    private Packet getResourceBindPacket(String id) throws TigaseStringprepException {
        Element elemRESOURCE = new Element("resource");
        elemRESOURCE.setCData(getResource());
        elemRESOURCE.setDefXMLNS("urn:ietf:params:xml:ns:xmpp-bind");

        Element elemBIND = new Element("bind");
        elemBIND.addChild(elemRESOURCE);
        elemBIND.setXMLNS("urn:ietf:params:xml:ns:xmpp-bind");

        Element elemIQ = new Element("iq");
        elemIQ.addAttribute("type", "set");
        elemIQ.addAttribute("id", id);
        elemIQ.setXMLNS("jabber:client");
        elemIQ.addChild(elemBIND);

        Packet bind = Packet.packetInstance(elemIQ);
        bind.setPacketFrom(getConnectionJID());
        bind.setPacketTo(getSessionManagerJID());
        return bind;
    }

    private Packet getPresencePacket(XmppPresence presenceStatus, String text, String id) throws TigaseStringprepException {
        Element elemSHOW = null;
        if (presenceStatus != XmppPresence.AVAILABLE) {
            elemSHOW = new Element("show");
            elemSHOW.setCData(presenceStatus.toString());
        }

        Element elemSTATUS = null;
        if (text != null) {
            elemSTATUS = new Element("status");
            elemSTATUS.setCData(text);
        }

        Element elemPRESENCE = new Element("presence");
        elemPRESENCE.setXMLNS("jabber:client");
        if (elemSHOW != null) {
            elemPRESENCE.addChild(elemSHOW);
        }
        if (elemSTATUS != null) {
            elemPRESENCE.addChild(elemSTATUS);
        }
        elemPRESENCE.addAttribute("id", id);

        Packet presence = Packet.packetInstance(elemPRESENCE, getUserJID(), null);
        presence.setPacketFrom(getConnectionJID());
        presence.setPacketTo(getSessionManagerJID());
        return presence;
    }
}
