/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.xmpp;

import com.kerio.im.core.api.service.AuthService;
import com.kerio.im.tigase.bridge.communicator.IMtoXmppChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.util.TigaseStringprepException;
import tigase.xmpp.JID;

/**
 *
 * @author jzacek
 */
public class XmppConnectionManager implements IAuthServiceProvider {

    private static final Logger LOGGER = Logger.getLogger(XmppConnectionManager.class.getName());

    private final Map<String, IXmppConnection> connectionListByID = new HashMap<>();
    private final Map<String, IXmppConnection> connectionListByUserName = new HashMap<>();
    private final String SESSION_MANAGER_ID;
    private final IMtoXmppChannel communicator;

    public XmppConnectionManager(IMtoXmppChannel communicator, String SESSION_MANAGER_ID) {
        this.communicator = communicator;
        this.SESSION_MANAGER_ID = SESSION_MANAGER_ID;
    }

    public synchronized IXmppConnection getOfllineDeliveryConnection(String user, String domain) throws TigaseStringprepException {
        return new OfflineDeliveryXmppConnection(JID.jidInstance(user, domain, null), SESSION_MANAGER_ID, communicator);
    }

    public IXmppConnection createConnection(String user, String domain) throws TigaseStringprepException, XMPPConnectionOpenFailed, InterruptedException {
        IXmppConnection newConnection = new XmppConnection(this.communicator, user, domain, SESSION_MANAGER_ID);
        synchronized (this) {
            registerConnection(newConnection);
        }
        if (!newConnection.open()) {
            destroyConnection(newConnection);
            throw new XMPPConnectionOpenFailed();
        }

        return newConnection;
    }

    public synchronized void destroyAllConnections() {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Closing all xmpp conections");
        connectionListByID.values().stream().forEach((connection) -> {
            connection.close();
        });
        connectionListByID.clear();
        connectionListByUserName.clear();
    }

    public synchronized void destroyConnection(IXmppConnection connection) {
        connection.close();
        removeConnection(connection);
    }

    public synchronized IXmppConnection getConnectionByUserName(String userName) {
        return connectionListByUserName.get(userName);
    }

    public synchronized IXmppConnection getConnectionById(String id) {
        return connectionListByID.get(id);
    }

    @Override
    public AuthService.Connection getAuthServiceForConnection(String id) {
        IXmppConnection connection = getConnectionById(id);
        if (connection == null) {
            return null;
        }

        //TODO enter better values
        return new AuthService.Connection(
                666,
                true,
                "127.0.0.1",
                667,
                false
        );
    }

    public IMtoXmppChannel getCommunicator() {
        return communicator;
    }

    private void registerConnection(IXmppConnection connection) {
        connectionListByID.put(connection.getConnectionID(), connection);
        connectionListByUserName.put(connection.getFullUserName(), connection);
    }

    private void removeConnection(IXmppConnection connection) {
        connectionListByID.remove(connection.getConnectionID());
        connectionListByUserName.remove(connection.getFullUserName());
    }
}
