/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.configuration;

import com.kerio.im.common.Task;
import com.kerio.im.common.error.Code;
import com.kerio.im.core.api.CoreService;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.api.event.ShutdownEvent;
import com.kerio.im.core.api.service.ConfigurationService;
import com.kerio.im.tigase.ClientResolver;
import com.kerio.im.tigase.ComponentLookup;
import com.kerio.im.tigase.ConnectComponent;
import com.kerio.im.tigase.MessageResourceSelector;
import com.kerio.im.tigase.RepositoryManager;
import com.kerio.im.tigase.SessionProvider;
import com.kerio.im.tigase.auth.ConnectAuthenticator;
import com.kerio.im.tigase.bridge.IMBridgePlugin;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.xmpp.ResourceValidatorImpl;
import com.kerio.im.tigase.configuration.task.AddBoshHeaders;
import com.kerio.im.tigase.configuration.task.AddCertificatePaths;
import com.kerio.im.tigase.configuration.task.AddCrossDomainPolicy;
import com.kerio.im.tigase.configuration.task.AddDatabaseRoot;
import com.kerio.im.tigase.configuration.task.ApplyConnectComponentProperties;
import com.kerio.im.tigase.configuration.task.ApplyTigaseProperties;
import com.kerio.im.tigase.configuration.task.CheckDatabase;
import com.kerio.im.tigase.configuration.task.InitializeAMP;
import com.kerio.im.tigase.configuration.task.InitializeBindResource;
import com.kerio.im.tigase.configuration.task.InitializeCipherFilter;
import com.kerio.im.tigase.configuration.task.InitializeDisconnectPlugin;
import com.kerio.im.tigase.configuration.task.InitializeDomainFilter;
import com.kerio.im.tigase.configuration.task.InitializeDynamicRosters;
import com.kerio.im.tigase.configuration.task.InitializePingPlugin;
import com.kerio.im.tigase.configuration.task.InitializePorts;
import com.kerio.im.tigase.configuration.task.InitializeSASL;
import com.kerio.im.tigase.configuration.task.InitializeTLS;
import com.kerio.im.tigase.configuration.task.InitializeVCard;
import com.kerio.im.tigase.configuration.task.InitializeVCardFilter;
import com.kerio.im.tigase.configuration.task.InitiliazeIMBridgePlugin;
import com.kerio.im.tigase.configuration.task.SubscribeToShutdownEvent;
import com.kerio.im.tigase.plugin.ConversationListenerFilter;
import com.kerio.im.tigase.plugin.DisconnectPlugin;
import com.kerio.im.tigase.roster.IDynamicDomainRoster;
import com.kerio.im.tigase.roster.IPersonalRosterListener;
import com.kerio.im.xmpp.api.XmppException;
import java.io.File;
import java.lang.ref.WeakReference;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.conf.Configurable;
import tigase.conf.ConfigurationException;
import tigase.conf.ConfiguratorAbstract;
import tigase.db.RepositoryFactory;
import tigase.db.TigaseDBException;
import tigase.db.UserRepository;
import tigase.db.jdbc.JDBCRepository;
import tigase.server.MessageRouter;
import tigase.server.QueueType;
import tigase.server.ServerComponent;
import tigase.server.xmppserver.S2SConnectionManager;
import tigase.server.xmppsession.SessionManager;

/**
 *
 * @author tskoda
 */
public class ConnectConfigurator extends ConfiguratorAbstract implements RepositoryManager {

    private static final Logger LOG = Logger.getLogger(ConnectConfigurator.class.getName());
    private final ComponentLookup lookup;
    private final CoreService coreService;
    private final File dataDirectory;
    private final IDynamicDomainRoster dynamicRoster;
    private final IPersonalRosterListener personalRosterListener;

    public ConnectConfigurator(final CoreService coreService, final ComponentLookup lookup, IDynamicDomainRoster dynamicRoster, IPersonalRosterListener personalRosterListener) {
        this.lookup = lookup;
        this.coreService = coreService;
        this.dataDirectory = coreService.configuration().get(GlobalKey.DATA_DIR);
        this.dynamicRoster = dynamicRoster;
        this.personalRosterListener = personalRosterListener;
    }

    @Override
    public void init(String[] args) throws ConfigurationException, TigaseDBException {

        try {
            PropertySetter propertySetter = new PropertySetter();
            PluginConfigurator pluginConfigurator = new PluginConfigurator(super.configRepo);
            ConfigurationService conf = this.coreService.configuration();

            run(new SubscribeToShutdownEvent(this.coreService.bus()), new CloseProcessor());
            run(new AddDatabaseRoot(), conf);
            run(new ApplyTigaseProperties(conf), propertySetter);
            run(new CheckDatabase(conf), null);
            run(new ApplyConnectComponentProperties(), propertySetter);
            run(new AddCertificatePaths(conf), propertySetter);
            run(new AddBoshHeaders(conf), propertySetter);
            run(new AddCrossDomainPolicy(conf), propertySetter);
            //TODO use plugin configurator
            run(new InitializePorts(), super.configRepo);
            run(new InitializeDisconnectPlugin(pluginConfigurator), this.coreService.auth());
            run(new InitializeTLS(new ClientResolver(this.lookup)), pluginConfigurator);
            run(new InitializeCipherFilter(), this.coreService.securityService());
            run(new InitializeSASL(pluginConfigurator), new ClientResolver(this.lookup));
            run(new InitializeVCard(pluginConfigurator), this.coreService);
            run(new InitializeBindResource(pluginConfigurator), new ResourceValidatorImpl());
            run(new InitializeDomainFilter(pluginConfigurator), this.coreService);
            run(new InitializePingPlugin(this.lookup, this.coreService), pluginConfigurator);
            run(new InitializeVCardFilter(this.coreService), pluginConfigurator);
            run(new InitializeDynamicRosters(this.dynamicRoster, this.personalRosterListener), pluginConfigurator);
            run(new InitializeAMP(new MessageResourceSelector()), pluginConfigurator);
            run(new InitiliazeIMBridgePlugin(pluginConfigurator), this.coreService);
            super.init(args);

        } catch (Exception e) {
            throw new XmppException("Tigase initialization failed", e, Code.INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    public void setProperties(Map<String, Object> props) {

        try {
            super.setProperties(props);

            ConnectAuthenticator userAuthenticator = (ConnectAuthenticator) super.getAuthRepository();
            userAuthenticator.setAuthService(coreService.auth());
            userAuthenticator.setClientSession(new ClientResolver(lookup));

            AdditionalPropertySetter connectComponent = new AdditionalPropertySetter(ConnectComponent.COMPONENT_NAME);
            connectComponent.put(ComponentLookup.class.getName(), lookup)
                    .put(ClientResolver.class.getName(), new ClientResolver(lookup))
                    .put(CoreService.class.getName(), coreService)
                    .apply();

            AdditionalPropertySetter imBridgeComponent = new AdditionalPropertySetter(IMBridge.IM_BRIDGE_NAME);
            imBridgeComponent.put(CoreService.class.getName(), coreService)
                    .apply();

            AdditionalPropertySetter disconnectPlugin = new AdditionalPropertySetter(DisconnectPlugin.ID);
            disconnectPlugin.put(CoreService.class.getName(), coreService).apply();

        } catch (Exception e) {
            throw new XmppException("Error while configuring connect component", e, Code.INTERNAL_SERVER_ERROR);
        }

    }

    public ConnectComponent getConnectComponent() {
        return this.lookup.lookup(ConnectComponent.class);
    }

    public JDBCRepository getUserRepository(String domain) {
        UserRepository userRepo = super.getUserRepository().getRepoNoDef(domain);
        if (userRepo == null) {
            throw new XmppException("Failed to obtain repository for domain {0}", Code.RESOURCE_ERROR, domain);
        }
        return (JDBCRepository) super.getUserRepository().getRepoNoDef(domain);
    }

    @Override
    public void initializationCompleted() {
        super.initializationCompleted();

        //add conversation filter to session manager, pass only messages & filetransfers sent by user, and send/received group chat messages
        this.lookup.lookup(SessionManager.class).putFilter(new ConversationListenerFilter(
                coreService.conversations(),
                new ConversationListenerFilter.ChatMessage(ConversationListenerFilter.BLOCK_ON_BEHALF_OF_RECIPIENT, ConversationListenerFilter.BLOCK_DELAYED),
                new ConversationListenerFilter.FileTransfer(ConversationListenerFilter.BLOCK_ON_BEHALF_OF_RECIPIENT, ConversationListenerFilter.BLOCK_DELAYED),
                new ConversationListenerFilter.GroupChatMessage()),
                QueueType.OUT_QUEUE);

        //add conversation filter to client connection manager, out gueue handles packets sent FROM foreign user
        this.lookup.lookup(S2SConnectionManager.class).putFilter(new ConversationListenerFilter(
                coreService.conversations(),
                new ConversationListenerFilter.ChatMessage(),
                new ConversationListenerFilter.GroupChatMessage(),
                new ConversationListenerFilter.FileTransfer()),
                QueueType.OUT_QUEUE);

        //add conversation filter to client connection manager, out gueue handles packets sent FROM foreign user
        this.lookup.lookup(S2SConnectionManager.class).putFilter(new ConversationListenerFilter(
                coreService.conversations(),
                new ConversationListenerFilter.GroupChatMessage()),
                QueueType.IN_QUEUE);
    }

    @Override
    public Configurable getComponent(String name) {
        return super.getComponent(name);
    }

    @Override
    public String newPacketId(String prefix) {
        return super.newPacketId(prefix);
    }

    @Override
    public void release() {
        super.release();
    }

    @Override
    public boolean addComponent(ServerComponent component) {
        lookup.addComponent(component);
        return super.addComponent(component);

    }

    @Override
    public boolean deleteComponent(ServerComponent component) {
        lookup.deleteComponent(component);
        return super.deleteComponent(component);

    }

    @Override
    public void addRepository(String domain) {
        if (super.getUserRepository().getRepoNoDef(domain) != null) {
            throw new RepositoryManager.DuplicateException(domain);
        }

        String databaseType = String.format(TigaseKey.USER_DB.value);
        String databaseConnectionString = String.format(TigaseKey.USER_DB_URI.value, this.dataDirectory.getAbsolutePath(), domain);

        try {
            UserRepository repository = RepositoryFactory.getUserRepository(databaseType, databaseConnectionString, Collections.EMPTY_MAP);
            super.getUserRepository().addRepo(domain, repository);
        } catch (Exception e) {
            throw new XmppException("Adding user repository failed", e, Code.INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    public void removeRepository(String domain) {

        if (super.getUserRepository().getRepoNoDef(domain) == null) {
            throw new RepositoryManager.NotFoundException(domain);
        }

        //TODO we should close it and shutdown it and remove it from repository factory or do not use it at all
        super.getUserRepository().removeRepo(domain);

    }

    @Override
    public boolean exist(String domain) {
        return super.getUserRepository().getRepoNoDef(domain) != null;
    }

    private <T> void run(Task<T> task, T parameter) {
        task.run(parameter);
    }

    private AdditionalPropertySetter config(String component) {
        return new AdditionalPropertySetter(component);
    }

    public SessionProvider getSessions() {
        return new SessionProvider(this.lookup.lookup(SessionManager.class), this.coreService.auth());
    }

    private class AdditionalPropertySetter {

        private final Map<String, Object> values;
        private final String component;

        public AdditionalPropertySetter(String component) {
            this.values = new HashMap<String, Object>();
            this.component = component;
        }

        public AdditionalPropertySetter put(String key, Object value) {
            this.values.put(key, value);
            return this;
        }

        public void apply() throws ConfigurationException {
            ConnectConfigurator.this.putProperties(component, values);
        }
    }

    private class PropertySetter implements PropertyProcessor {

        @Override
        public void put(String key, String value) {
            ConnectConfigurator.this.processProperty(key, value);
        }
    }

    private class CloseProcessor implements ServiceBus.EventListener<ShutdownEvent> {

        @Override
        public void event(ShutdownEvent event) {

            try {
                ConnectComponent connectComponent = lookup.lookup(ConnectComponent.class);
                if (connectComponent != null) {
                    connectComponent.release();
                } else {
                    LOG.log(Level.WARNING, "Can not close connect component - component not found");
                }
            } catch (Exception e) {
                LOG.log(Level.SEVERE, "Error occured while closing message router component", e);
            }

            try {

                MessageRouter messageRouter = lookup.lookup(MessageRouter.class);

                if (messageRouter != null) {
                    messageRouter.release();
                } else {
                    LOG.log(Level.WARNING, "Can not close message router - component not found");
                }
            } catch (Exception e) {
                LOG.log(Level.SEVERE, "Error occured while closing message router component", e);
            }

            try {
                List<WeakReference<? extends ServerComponent>> list = lookup.getAll();
                for (ListIterator<WeakReference<? extends ServerComponent>> it = list.listIterator(list.size()); it.hasPrevious();) {
                    WeakReference<? extends ServerComponent> component = it.previous();
                    ServerComponent sComponent = component.get();

                    if ((sComponent != null) && !(sComponent instanceof MessageRouter) && (sComponent != this)) {
                        try {
                            sComponent.release();
                        } catch (Exception e) {
                            LOG.log(Level.SEVERE, "Error occured while closing component: " + sComponent != null ? sComponent.getName() : "unknown", e);
                        }
                    }
                }
            } catch (Exception e) {
                LOG.log(Level.SEVERE, "Error occured while closing components", e);
            }

//            try {
//                //if the possible error would happen before advance user repository initialization, but after real user repository initialization
//                if ((userRepository == null) && (ConnectConfigurator.super.user_repository != null)) {
//                    userRepository = new TigaseDatabase((JDBCRepository) user_repository);
//                }
//
//                if (userRepository != null) {
//                    userRepository.bye();
//                }
//            } catch (Exception e) {
//                LOG.log(Level.SEVERE, "Error occured while closing user repository", e);
//            }
            try {
                ConnectConfigurator.this.release();
            } catch (Exception e) {
                LOG.log(Level.SEVERE, "Error occured while closing connect configuration component", e);
            }
        }
    }
}
