/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.derby;

import com.kerio.im.common.error.Code;
import java.io.File;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Logger;
import static com.kerio.im.tigase.derby.DerbyDatabase.*;
import com.kerio.im.tigase.derby.version.Version;
import com.kerio.im.xmpp.api.XmppException;
import java.util.logging.Level;
import org.apache.derby.jdbc.EmbeddedDataSource40;
import tigase.db.jdbc.JDBCRepository;

/**
 *
 * @author tskoda
 */
//TODO make it more generic
public class DerbyChecker {

    private static final Logger LOG = Logger.getLogger(DerbyChecker.class.getName());
    private static final String PROCEDURE_EXISTS = "X0Y68";
    public static final String SELECT_TABLE = "select count(TABLEID) from SYS.SYSTABLES where TABLENAME = UPPER(?)";
    public static final String TIGASE_CONF_USER_EXIST = "{call TigGetUserDBUid('db-properties') }";
    public static final String SCHEMA_RECORD_EXIST = "values TigGetDBProperty('schema-version') ";
    private final String dbPath;
    private final EmbeddedDataSource40 derbyDataSource;

    //TODO add username and password
    public DerbyChecker(File dbRoot, String dbName) {
        this(new File(dbRoot, dbName).getAbsolutePath());
    }

    DerbyChecker(String dbName) {

        this.dbPath = dbName;
        this.derbyDataSource = new EmbeddedDataSource40();
        this.derbyDataSource.setDatabaseName(this.dbPath);
    }

    public boolean check() {
        LOG.log(Level.FINE, "Checking database on location {0}", this.dbPath);
        boolean result = false;
        try {
            checkIfExists();
            Connection connection = checkIfBootable();
            try {
                checkSchemaVersion(connection);
                checkAllTablesExist(connection);
                checkAllProceduresExist(connection);
                checkDataExists(connection);
                result = true;
                LOG.log(Level.INFO, "Checking database on location {0} successful", this.dbPath);
            } finally {
                close(connection);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to check database {0}", e, Code.RESOURCE_ERROR, this.dbPath);
        }

        return result;
    }

    protected void checkIfExists() {
        LOG.finer("Checking if database folder exists and is accessible");

        File dbDirectory = new File(this.dbPath);

        if (!dbDirectory.exists()) {
            throw new NotExistException(this.dbPath);
        } else if (!dbDirectory.isDirectory()
                || !dbDirectory.canRead()
                || !dbDirectory.canWrite()) {
            throw new XmppException("Database folder {0} not accessible", Code.RESOURCE_ERROR, this.dbPath);
        }
    }

    protected Connection checkIfBootable() throws SQLException {
        LOG.finer("Checking if database is bootable");

        Connection result = this.derbyDataSource.getConnection();

        return result;
    }

    protected void checkSchemaVersion(final Connection connection) throws SQLException {
        Version version = checkSchema(connection);
        switch (version) {
            case UNDEFINED:
                throw new XmppException("Failed to check database {0}, database has uknown schema ", Code.RESOURCE_ERROR, this.dbPath);
            case CURRENT:
                LOG.log(Level.FINE, "Database schema version {0} is correct", version);
                break;
            default:
                throw new ObsoleteSchemaException(version);
        }
    }

    protected void checkAllTablesExist(final Connection connection) throws SQLException {
        LOG.finer("Checking if all default tables exists");
        PreparedStatement statement = null;

        try {
            statement = connection.prepareStatement(SELECT_TABLE);

            for (TABLE table : TABLE.values()) {
                if (!checkTableExists(table.name, statement)) {
                    throw new XmppException("Table {0} in database {1} not found", Code.RESOURCE_ERROR, table.name, this.dbPath);
                }
            }
        } finally {
            close(statement);
        }
    }

    protected boolean checkTableExists(String table, PreparedStatement statement) throws SQLException {

        boolean result = false;

        statement.setString(1, table);
        ResultSet resultSet = null;

        try {
            resultSet = statement.executeQuery();

            if (resultSet.next()) {
                int count = resultSet.getInt(1);
                if (count == 1) {
                    result = true;
                }
            }
        } finally {
            close(resultSet);
        }

        return result;

    }

    protected void checkAllProceduresExist(final Connection connection) throws SQLException {
        LOG.finer("Checking if all default procedures exists");

        for (PROCEDURE procedure : PROCEDURE.values()) {
            if (!checkProcedureExists(procedure.create, connection)) {
                throw new XmppException("Procedure {0} in database {1} not found", Code.RESOURCE_ERROR, procedure.name(), this.dbPath);
            }
        }
    }

    protected boolean checkProcedureExists(String createProcedure, Connection connection) throws SQLException {
        boolean result = false;

        Statement statement = null;

        try {
            statement = connection.createStatement();
            statement.execute(createProcedure);

        } catch (SQLException e) {
            if (e.getSQLState().equals(PROCEDURE_EXISTS)) {
                result = true;
            } else {
                throw e;
            }
        } finally {
            close(statement);
        }

        return result;
    }

    protected void checkDataExists(final Connection connection) throws SQLException {
        LOG.finer("Checking if all default data exists");

        Statement statement = connection.createStatement();
        try {
            ResultSet dbPropsUserResult = statement.executeQuery(TIGASE_CONF_USER_EXIST);
            try {
                if (!dbPropsUserResult.next()) {
                    throw new XmppException("System user in database {0} not found", Code.RESOURCE_ERROR, this.dbPath);
                }
            } finally {
                close(dbPropsUserResult);
            }

            ResultSet schemaResult = statement.executeQuery(SCHEMA_RECORD_EXIST);
            try {
                boolean result = false;
                if (schemaResult.next()) {
                    String schemaVersion = schemaResult.getString(1);
                    result = ((schemaVersion != null) && (!schemaVersion.isEmpty()));
                }

                if (!result) {
                    throw new XmppException("Incorrect schema version in database {0}", Code.RESOURCE_ERROR, this.dbPath);
                }
            } finally {
                close(schemaResult);
            }

        } finally {
            close(statement);
        }
    }

    private Version checkSchema(Connection connection) throws SQLException {

        Version result = Version.UNDEFINED;

        Statement schemaStatement = connection.createStatement();
        try {
            ResultSet schemaResult = schemaStatement.executeQuery(JDBCRepository.DERBY_GETSCHEMAVER_QUERY);
            try {
                if (schemaResult.next()) {
                    result = Version.fromString(schemaResult.getString(1));
                }
            } finally {
                close(schemaResult);
            }
        } finally {
            close(schemaStatement);
        }

        return result;
    }

    private void close(ResultSet resultset) {
        if (resultset != null) {
            try {
                resultset.close();
            } catch (Exception e) {
                LOG.warning("Can not close ResultSet");
            }
        }
    }

    private void close(Statement statement) {
        if (statement != null) {
            try {
                statement.close();
            } catch (Exception e) {
                LOG.warning("Can not close Statement");
            }
        }
    }

    private void close(Connection connection) {
        if (connection != null) {
            try {
                connection.close();
            } catch (Exception e) {
                LOG.warning("Can not close Connection");
            }
        }
    }

    public static class ObsoleteSchemaException extends XmppException {

        private static final String MESSAGE = "Obsolete message schema found. Expected {0}, found {1}.";

        private final Version found;

        public ObsoleteSchemaException(Version found) {
            super(MESSAGE, Code.RESOURCE_ERROR, Version.CURRENT, found);
            this.found = found;
        }

        public Version getFound() {
            return found;
        }
    }

    public static class NotExistException extends XmppException {

        private static final String MESSAGE = "Database folder {0} not exist";

        public NotExistException(String path) {
            super(MESSAGE, Code.RESOURCE_ERROR, path);
        }
    }
}
