/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.derby;

import com.kerio.im.common.FileUtil;
import com.kerio.im.common.error.Code;
import com.kerio.im.tigase.derby.DerbyDatabase.DATA;
import com.kerio.im.tigase.derby.DerbyDatabase.PROCEDURE;
import com.kerio.im.tigase.derby.DerbyDatabase.TABLE;
import com.kerio.im.tigase.derby.DerbyDatabase.SCHEMA;
import com.kerio.im.tigase.derby.version.Version;
import com.kerio.im.xmpp.api.XmppException;
import java.io.File;
import java.io.IOException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.derby.jdbc.EmbeddedDataSource40;

/**
 *
 * @author tskoda
 */
public class DerbyHandler {

    private static final String LEGACY_OFFLINE_DELETE = "delete from app.tig_pairs p where p.nid in (select n.nid from app.tig_nodes n where n.node = 'msgoffline')";
    //private static final String LEGACY_OFFLINE_COUNT = "select count(n.nid) from tig_pairs p, tig_nodes n where p.NID = n.nid and n.node = 'msgoffline'";
    private static final String AMP_TABLE = "MSG_HISTORY";
    private static final String AMP_OFFLINE_DELETE = "delete from app.msg_history h";
    //private static final String AMP_OFFLINE_COUNT = "select count(o.MSG_ID) from APP.MSG_HISTORY o";

    private static final Logger LOG = Logger.getLogger(DerbyHandler.class.getName());
    private static final String CREATE = "create";
    private final String dbPath;
    private final String dbName;

    public DerbyHandler(File dbRoot, String dbName) {
        dbPath = new File(dbRoot, dbName).getAbsolutePath();
        this.dbName = dbName;
    }

    public String getDbPath() {
        return dbPath;
    }

    public void create() {

        try {
            this.createDatabase();
        } catch (SQLException e) {
            throw new XmppException("Creating new database {0} failed", e, Code.RESOURCE_ERROR, this.dbPath);
        }
    }

    public boolean backup(String path) {

        try {
            this.backupDatabase(path);
            return true;
        } catch (SQLException e) {
            String[] params = {dbName, path};
            LOG.log(Level.SEVERE, "BACKUP: XMPP database backup failed for domain '{0}'", dbName);
            LOG.log(Level.INFO, "Database backup failed for domain '{0}'; backup path '{1}'", params);
            return false;
        }
    }

    public boolean remove() {
        boolean result = true;
        try {
            File derbyFolder = new File(this.dbPath);
            if (derbyFolder.exists()) {
                FileUtil.deleteFolderWithFiles(new File(derbyFolder, "seg0"));
                FileUtil.deleteFolderWithFiles(new File(derbyFolder, "tmp"));
                FileUtil.deleteFolderWithFiles(new File(derbyFolder, "log"));
                FileUtil.deleteFile(derbyFolder, "service.properties");
                FileUtil.deleteFile(derbyFolder, "BACKUP.HISTORY");
                FileUtil.deleteFile(derbyFolder, "dbex.lck");
                FileUtil.deleteFile(derbyFolder, "db.lck");
                result = FileUtil.deleteFolder(derbyFolder); //expects folder to be empty
            }
        } catch (IOException e) {
            throw new XmppException("Failed to remove domain database", e, Code.RESOURCE_ERROR);
        }
        return result;
    }

    public void upgradeSchema(Version found) {
        Version latestSchema = Version.UNDEFINED;
        try {
            EmbeddedDataSource40 ds = new EmbeddedDataSource40();
            ds.setDatabaseName(dbPath);
            Connection connection = ds.getConnection();
            try {
                Statement statement = connection.createStatement();
                try {
                    connection.setAutoCommit(false);
                    switch (found) {
                        case V_5_2: {
                            //upgrade to V_5_3
                            LOG.log(Level.INFO, "Upgrading schema from 5.2 to 5.3");
                            latestSchema = Version.V_5_3;

                            LOG.log(Level.INFO, "Creating personalized contact table");
                            statement.execute(TABLE.PERSONAL_ROSTER.create);
                        }
                        //no break, we should continuously upgrade to the latest!
                        case V_5_3: {
                            //upgrade to V_5_4
                            latestSchema = Version.V_5_4;
                            LOG.log(Level.INFO, "Upgrading schema from 5.3 to 5.4");

                            LOG.log(Level.INFO, "Cleaning legacy offline message storage");
                            int legacyDeleted = statement.executeUpdate(LEGACY_OFFLINE_DELETE);
                            LOG.log(Level.INFO, "Cleaning legacy offline message storage successful, deleted {0} records", legacyDeleted);

                            LOG.log(Level.INFO, "Cleaning amp offline message storage");
                            ResultSet tables = connection.getMetaData().getTables(null, null, AMP_TABLE, null);
                            try {
                                if (tables.next()) {
                                    int ampDeleted = statement.executeUpdate(AMP_OFFLINE_DELETE);
                                    LOG.log(Level.INFO, "Cleaning amp offline message storage successful, deleted {0} records", ampDeleted);
                                } else {
                                    LOG.log(Level.INFO, "Cleaning amp offline message storage skipped, amp table not created yet");
                                }
                            } finally {
                                close(tables);
                            }
                        }
                        //no break, we should continuously upgrade to the latest!
                        case V_5_4: {
                            //nothing, this is currently the latest version
                            break;
                        }
                        default:
                            //all other variants are not supported
                            throw new XmppException("Upgrading {0} database schema from {1} to {2} failed, upgrade not implemented.", Code.RESOURCE_ERROR, this.dbPath, found, Version.CURRENT);
                    }
                    LOG.log(Level.INFO, "Finalizing schema upgrade to {0}", latestSchema);
                    statement.execute(SCHEMA.UPDATE_DB_SCHEMA.create(latestSchema.toString()));

                } finally {
                    close(statement);
                }

                //update schema to target
                connection.commit();
                LOG.log(Level.INFO, "Upgrading schema record from {0} to {1} successful", new Object[]{found, latestSchema});

            } finally {
                //rollback after commit should make no harm
                rollback(connection);
                close(connection);
            }
        } catch (SQLException e) {
            throw new XmppException("Upgrading {0} database schema from {1} to {2} failed", e, Code.RESOURCE_ERROR, this.dbPath, found, latestSchema);
        }
    }

    protected void backupDatabase(String path) throws SQLException {
        String[] params = {dbName, path};

        LOG.log(Level.INFO, "Database backup started for domain '{0}'; backup path '{1}'", params);

        EmbeddedDataSource40 ds = new EmbeddedDataSource40();
        ds.setDatabaseName(dbPath);
        Connection connection = ds.getConnection();

        try {
            connection.setAutoCommit(false);
            CallableStatement statement = connection.prepareCall("CALL SYSCS_UTIL.SYSCS_BACKUP_DATABASE(?)");

            try {
                statement.setString(1, path);
                statement.execute();
                LOG.log(Level.INFO, "Database backup finished succesfully for domain '{0}'", dbName);
            } finally {
                close(statement);
            }
        } finally {
            close(connection);
        }
    }

    protected void createDatabase() throws SQLException {

        LOG.log(Level.FINE, "Creating new database on path {0}", this.dbPath);
        EmbeddedDataSource40 ds = new EmbeddedDataSource40();
        ds.setDatabaseName(this.dbPath);
        ds.setCreateDatabase(CREATE);
        Connection connection = ds.getConnection();

        try {
            connection.setAutoCommit(false);
            Statement statement = connection.createStatement();
            try {
                LOG.finer("Creating tables");
                for (TABLE table : TABLE.values()) {
                    statement.execute(table.create);

                    for (String index : table.indexes) {
                        statement.execute(index);
                    }
                }

                LOG.finer("Creating procedures");
                for (PROCEDURE procedure : PROCEDURE.values()) {
                    statement.execute(procedure.create);
                }

                LOG.finer("Inserting default data");
                for (DATA data : DATA.values()) {
                    statement.execute(data.insert);
                }

                statement.close();
                connection.commit();
                LOG.log(Level.FINE, "Creating new database on path {0} successful", this.dbPath);
            } finally {
                close(statement);
            }

        } finally {
            close(connection);
        }
    }

    private void close(ResultSet resultSet) {
        if (resultSet != null) {
            try {
                resultSet.close();
            } catch (Exception e) {
                LOG.warning("Can not close ResultSet");
            }
        }
    }

    private void close(Statement statement) {
        if (statement != null) {
            try {
                statement.close();
            } catch (Exception e) {
                LOG.warning("Can not close Statement");
            }
        }
    }

    private void rollback(Connection connection) {
        if (connection != null) {
            try {
                connection.rollback();
            } catch (Exception e) {
                LOG.warning("Can not rollback Connection");
            }
        }
    }

    private void close(Connection connection) {
        if (connection != null) {
            try {
                connection.close();
            } catch (Exception e) {
                LOG.warning("Can not close Connection");
            }
        }
    }
}
