/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.plugin;

import com.kerio.im.common.error.Code;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.tigase.ComponentLookup;
import com.kerio.im.tigase.ConnectComponent;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.event.PingEvent;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.NonAuthUserRepository;
import tigase.db.TigaseDBException;
import tigase.server.Iq;
import tigase.server.Packet;
import tigase.util.TigaseStringprepException;
import tigase.xml.Element;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPPreprocessorIfc;
import tigase.xmpp.XMPPProcessor;
import tigase.xmpp.XMPPResourceConnection;

/**
 *
 * @author tskoda
 */
public class PingPlugin extends XMPPProcessor implements XMPPPreprocessorIfc, ServiceBus.EventListener<PingEvent> {

    public static final String ID = "connect-ping";
    private static final Element PING_ELEMENT = new Element("ping", new String[]{"xmlns"}, new String[]{"urn:xmpp:ping"});
    private static final Logger LOG = Logger.getLogger(PingPlugin.class.getName());
    private static final String ELEMENT_PING_RESPONSE = "iq";
    private static final String XMLNS_PING_RESPONSE = "jabber:client";
    private static final String[] ELEMENTS = {ELEMENT_PING_RESPONSE};
    private static final String[] NAMESPACES = {XMLNS_PING_RESPONSE};
    private static final boolean PASS = false;
    private static final boolean BLOCK = true;
    /**
     *
     */
    private ServiceBus bus;
    /**
     *
     */
    private ComponentLookup lookup;

    public PingPlugin() {
    }

    @Override
    public String id() {
        return ID;
    }

    @Override
    public String[] supElements() {
        return ELEMENTS;
    }

    @Override
    public String[] supNamespaces() {
        return NAMESPACES;
    }

    @Override
    public void init(Map<String, Object> settings) throws TigaseDBException {
        if (settings.containsKey(ServiceBus.class.getName()) && (this.bus == null)) {
            this.bus = (ServiceBus) settings.get(ServiceBus.class.getName());
            this.bus.subscribe(PingEvent.class, this);
        }

        if (settings.containsKey(ComponentLookup.class.getName()) && (this.lookup == null)) {
            this.lookup = (ComponentLookup) settings.get(ComponentLookup.class.getName());
        }
    }

    @Override
    public boolean preProcess(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) {
        boolean result = PASS;
        if ((session != null) && session.isServerSession() && (packet.getType() != null) && (packet.getType() == StanzaType.result) && isPingId(packet.getElement())) {
            LOG.log(Level.FINE, "Ping response from {0} received", packet.getStanzaFrom());
            try {
                PingEvent pingEvent = from(packet);
                this.bus.publish(pingEvent);

            } catch (Exception e) {
                LOG.log(Level.INFO, "Ping response processing failed", e);
            } finally {
                result = BLOCK;
            }
        }
        return result;
    }

//    @Override
//    public void processFromUserToServerPacket(JID connectionId, Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
//    }
//
//    @Override
//    public void processToUserPacket(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
//        if ((packet.getProcessorsIds().size() == 1) && (packet.wasProcessedBy(ID))) {
//            super.processToUserPacket(packet, session, repo, results, settings);
//        }
//    }
//
//    @Override
//    public void processFromUserOutPacket(JID connectionId, Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
//        //only if it wasnt processed by other iq plugin
//        if ((packet.getProcessorsIds().size() == 1) && (packet.wasProcessedBy(ID))) {
//            super.processFromUserOutPacket(connectionId, packet, session, repo, results, settings);
//        }
//    }
//
//    @Override
//    public void processServerSessionPacket(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
//        if ((packet.getType() == StanzaType.result) && (isPingId(packet.getElement()))) {
//            LOG.log(Level.FINE, "Ping response from {0} received", packet.getStanzaFrom());
//            try {
//                PingEvent result = from(packet);
//                this.bus.publish(result);
//            } catch (Exception e) {
//                LOG.log(Level.INFO, "Ping response processing failed", e);
//            }
//        }
//    }
    private boolean isPingId(Element element) {
        String id;
        boolean result = false;
        if ((element != null) && ((id = element.getAttribute("id")) != null)) {
            result = PingEvent.isPingId(id);
        }

        return result;
    }

    @Override
    public void event(PingEvent event) {
        if (event.getType() == PingEvent.Type.REQUEST) {
            try {
                ConnectComponent sender = this.lookup.lookup(ConnectComponent.class);
                if (sender != null) {
                    Packet pingPacket = from(event);
                    sender.processOutPacket(pingPacket);
                } else {
                    LOG.log(Level.INFO, "Failed to send ping request, packet sender is null");
                }
            } catch (TigaseStringprepException e) {
                throw new XmppException("Server ping request failed, either sender {0} or recipient {1} is invalid", Code.COMMAND_ERROR, new Object[]{event.getSender(), event.getRecipient()});
            }
        }
    }

    private Packet from(PingEvent event) throws TigaseStringprepException {
        Element iq = new Element(
                "iq",
                new Element[]{PING_ELEMENT},
                new String[]{"from", "to", "id", "type"},
                new String[]{event.getSender(), event.getRecipient(), event.getId(), "get"});

        return new Iq(iq);
    }

    private PingEvent from(Packet packet) {
        String sender = packet.getStanzaFrom().toString();
        String recipient = packet.getStanzaTo().toString();
        String id = packet.getElement().getAttribute("id");

        return new PingEvent(PingEvent.Type.RESPONSE_SUCCESS, id, sender, recipient);
    }
}
