/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.plugin;

import com.kerio.im.core.api.CoreService;
import com.kerio.im.core.api.event.StartupFinishedEvent;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.api.service.VCardService;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.NonAuthUserRepository;
import tigase.db.TigaseDBException;
import tigase.server.Packet;
import tigase.xml.Element;
import tigase.xmpp.Authorization;
import tigase.xmpp.JID;
import tigase.xmpp.NoConnectionIdException;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.PacketErrorTypeException;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.impl.VCardTemp;

/**
 *
 * @author tskoda
 */
public class VCardPlugin extends VCardTemp {

    private static final Logger LOG = Logger.getLogger(VCardPlugin.class.getName());
    public static final String PLUGIN_ID = "connect-vcard";
    private VCardService<VCardWrapper> vcardService;

    @Override
    public void init(Map<String, Object> settings) throws TigaseDBException {

        super.init(settings);
        CoreService coreService = init(CoreService.class.getName(), settings);
        coreService.bus().subscribe(StartupFinishedEvent.class, new StartupListener(coreService));
        //TODO listen for VCardEvent and publish

    }

    private <T> T init(String id, Map<String, Object> settings) throws TigaseDBException {
        Object vcardManagerObject = settings.get(id);
        if (vcardManagerObject != null) {
            return (T) vcardManagerObject;
        } else {
            throw new TigaseDBException("VCardPlugin initialization failed, " + id + " not found");
        }
    }

    @Override
    public String id() {
        return PLUGIN_ID;
    }

    /**
     * This method is called when the owner of vcard wants to manipulate with
     * it. If gal is enabled we allow to full vcard manipulation otherwise we
     * forbid updates and force clients to use vcard from gal.
     */
    @Override
    public void processFromUserToServerPacket(JID connectionId, Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
        if (controlled()) {
            try {
                 JID userJID = session.getJID();
                switch (packet.getType()) {
                    case set:
                        LOG.log(Level.INFO, "Ignoring 'set own vcard' request received from {0}, vcard updates not allowed", userJID);
                        results.offer(Authorization.NOT_ALLOWED.getResponseMessage(packet, "VCard update not allowed", false));
                        //TODO do we need to republish the original one? NO
                        break;
                    case get:
                       LOG.log(Level.FINEST, "Processing 'get own vcard' request received from {0}", userJID);
                        
                        VCardWrapper vcard = this.vcardService.get(userJID.getLocalpart(), userJID.getDomain());
                        if (vcard != null) {
                            Packet result = getVCardPacket(packet, vcard);
                            result.setPacketTo(session.getConnectionId());
                            results.offer(result);
                            LOG.log(Level.INFO, "Processing 'get own vcard' request received from {0} successful", userJID);
                        } else {
                            results.offer(Authorization.ITEM_NOT_FOUND.getResponseMessage(packet, "User not found", false));
                            LOG.log(Level.INFO, "Processing 'get own vcard' request received from {0} failed, vcard not found", userJID);
                        }

                        break;
                }
            } catch (NoConnectionIdException e) {
                LOG.log(Level.INFO, "Processing 'own vcard' request failed, invalid client xmpp request", e);
            } catch (NotAuthorizedException e) {
                LOG.log(Level.INFO, "Processing 'own vcard' request failed", e);
                results.offer(Authorization.NOT_AUTHORIZED.getResponseMessage(packet, "You must authorize session first.", true));
            } catch (Exception e) {
                LOG.log(Level.INFO, "Processing 'own vcard' request failed", e);
                results.offer(Authorization.INTERNAL_SERVER_ERROR.getResponseMessage(packet, "Failed to retrieve vcard.", true));
            }
        } else {
            super.processFromUserToServerPacket(connectionId, packet, session, repo, results, settings);
        }
    }

    /**
     * This method is called when someone from foreign domain ask for vcard of
     * user on this server.
     */
    @Override
    public void processNullSessionPacket(Packet packet, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {

        if (controlled()) {
            try {
                if (packet.getType() == StanzaType.get) {
                    LOG.log(Level.INFO, "Ignoring 'get {0} vcard' request received from {1}, vcards are not sent outside domain", new Object[]{packet.getTo(), packet.getFrom()});
//                    JID userJID = packet.getStanzaTo();
//                    User user = new User(userJID.getLocalpart(), userJID.getDomain(), null, false);
//
//                    VCardWrapper wrapper = VCardWrapper.instance();
//                    if (this.vcardService.fillVCard(user, wrapper)) {
//                        Packet result = getVCardPacket(packet, wrapper);
//                        results.offer(result);
//                    } else {
//                        results.offer(Authorization.ITEM_NOT_FOUND.getResponseMessage(packet, "User not found", true));
//                    }
                    results.offer(Authorization.ITEM_NOT_FOUND.getResponseMessage(packet, "User not found", false));
                }
            } catch (Exception e) {
                LOG.log(Level.INFO, "Processing vcard request failed", e);
                results.offer(Authorization.INTERNAL_SERVER_ERROR.getResponseMessage(packet, "Failed to retrieve vcard", true));
            }

        } else {
            super.processNullSessionPacket(packet, repo, results, settings);
        }
    }

    /**
     * Loccaly logged in user is requesting someone elses vcard. Someone else
     * could be from the same domain or another on different server.
     */
    @Override
    public void processFromUserOutPacket(JID connectionId, Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
        if (controlled()) {
            if (session.isLocalDomain(packet.getStanzaTo().getDomain(), false)) {
                // This is a local user so we can quickly get his vCard from the database
                try {
                    JID userJID = packet.getStanzaTo();
                    
                    LOG.log(Level.FINEST, "Processing 'get {0} vcard' request received from {1}", new Object[]{userJID, session.getJID()});
                    
                    VCardWrapper vcard = this.vcardService.get(userJID.getLocalpart(), userJID.getDomain());
                    if (vcard != null) {
                        Packet result = getVCardPacket(packet, vcard);
                        result.setPacketTo(connectionId);
                        results.offer(result);
                        LOG.log(Level.INFO, "Processing 'get {0} vcard' request received from {1} successful", new Object[]{userJID, session.getJID()});
                    } else {
                        LOG.log(Level.INFO, "Processing 'get {0} vcard' request received from {1} failed, user not found", new Object[]{userJID, session.getJID()});
                        results.offer(Authorization.ITEM_NOT_FOUND.getResponseMessage(packet, "User not found", false));
                    }
                } catch (Exception e) {
                    LOG.log(Level.INFO, "Processing vcard request failed", e);
                    results.offer(Authorization.INTERNAL_SERVER_ERROR.getResponseMessage(packet, "Failed to retrieve vcard.", true));
                }
            } else {
                // Else forward the packet to a remote server
                results.offer(packet.copyElementOnly());
            }

        } else {
            super.processFromUserOutPacket(connectionId, packet, session, repo, results, settings);
        }
    }

    @Override
    public void processToUserPacket(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) throws PacketErrorTypeException {
        super.processToUserPacket(packet, session, repo, results, settings);
    }

    @Override
    public void processServerSessionPacket(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) {
        super.processServerSessionPacket(packet, session, repo, results, settings);
    }

    private Packet getVCardPacket(Packet original, VCardWrapper wrapper) {

        Packet result;

        if (wrapper.isEmpty()) {
            result = original.okResult((Element) null, 1);
        } else {
            result = original.okResult((Element) null, 0);
            result.getElement().addChild(wrapper.root());
        }

        result.setPacketFrom(null);
        result.setPacketTo(null);
        return result;
    }

    private boolean controlled() {
        synchronized(this) {
            return this.vcardService != null && this.vcardService.enabled();
        }
    }

    private class StartupListener implements ServiceBus.EventListener<StartupFinishedEvent> {

        private final CoreService coreService;

        public StartupListener(CoreService coreService) {
            this.coreService = coreService;
        }

        @Override
        public void event(StartupFinishedEvent event) {
            synchronized(VCardPlugin.this) {
                VCardPlugin.this.vcardService = coreService.vcards();
            }
        }
    }
}
