/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.TigaseDBException;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.impl.roster.RosterAbstract;

/**
 *
 * @author tskoda
 */
public class DynamicDomainRoster implements IDynamicDomainRoster {

    private static final Logger LOG = Logger.getLogger(DynamicDomainRoster.class.getName());

    private final DefaultRoster defaultRoster;
    private final PersonalRoster personalRoster;
    private final SessionUtil sessionUtil;
    private final ElementMapper mapper;

    public DynamicDomainRoster(DefaultRoster defaultRoster, PersonalRoster personalRoster, SessionUtil sessionUtil, ElementMapper mapper) {
        this.defaultRoster = defaultRoster;
        this.personalRoster = personalRoster;
        this.sessionUtil = sessionUtil;
        this.mapper = mapper;
    }

    @Override
    public DefaultRoster getDefaultRoster() {
        return defaultRoster;
    }

    @Override
    public PersonalRoster getPersonalRoster() {
        return personalRoster;
    }

    @Override
    public void setItemExtraData(Element item) {
        String jid = item.getAttribute("jid");
        LOG.log(Level.FINE, "Ignoring set item extra data, user {0}, item {1}", new Object[]{jid, item});
    }

    @Override
    public Element getItemExtraData(Element item) {
        return null;
    }

    @Override
    public void init(Map<String, Object> props) {
    }

    @Override
    public void init(String par) {
    }

    @Override
    public RosterAbstract.SubscriptionType getBuddySubscription(XMPPResourceConnection session, JID buddy) throws NotAuthorizedException, TigaseDBException {

        BareJID ownerJID = this.sessionUtil.ownerOf(session);
        RosterAbstract.SubscriptionType result = RosterAbstract.SubscriptionType.none;
        
        PersonalRoster.Buddy find = this.personalRoster.of(ownerJID).find(buddy.getBareJID());

        if (((find == null) || (!find.isDeleted())) && (manages(session, buddy))) {
            result = RosterAbstract.SubscriptionType.both;
        }

        LOG.log(Level.FINE, "Reading subscription between {0} and {1} successfull, subscription is {2}", new Object[]{ownerJID, buddy, result});

        return result;
    }

    //do not consider private roster - see Presence.processInProbe
    @Override
    public boolean manages(XMPPResourceConnection session, JID buddy) throws NotAuthorizedException, TigaseDBException {
        boolean result = false;

        BareJID ownerJID = this.sessionUtil.ownerOf(session);

        //we manages only people from ownerOf domain, other contacts should be stored in persistent roster, so we return false
        String ownerDomain = ownerJID.getDomain();
        if (this.sessionUtil.localDomain(session)) {
            result = this.defaultRoster.roster(ownerDomain).find(ownerJID) != null //owner is managed
                    && this.defaultRoster.roster(ownerDomain).find(buddy.getBareJID()) != null; //buddy is in owners roster
        }

        LOG.log(Level.FINE, "Resolving buddy {0} presence in {1}''s roster successful, result is {2}", new Object[]{buddy, ownerJID, result});
        
        return result;
    }

    //Element is further considered to have both subscription, do not return deleted
    @Override
    public Element getBuddyItem(XMPPResourceConnection session, JID buddy) throws NotAuthorizedException {

        BareJID ownerJID = this.sessionUtil.ownerOf(session);
        Element result = null;
        RosterContent defaultContent = this.defaultRoster.roster(ownerJID.getDomain());
        RosterContent.Buddy def = defaultContent.find(buddy.getBareJID());
        String state = "not found";

        if ((defaultContent.find(ownerJID) != null) && (def !=null)) {
            PersonalRoster.Buddy personal = this.personalRoster.of(ownerJID).find(buddy.getBareJID());

            //only if default user exists
            if (personal == null) {
                result = def.getElement();
                state = "found in default";
            } else if (personal.getFlag() == PersonalRoster.Buddy.FLAG.updated) {
                result = mapper.value(def.getUser(), personal);
                state = "found in personal";
            } else {
                state = "not found - " + personal.getFlag();
            }
        }

        LOG.log(Level.FINE, "Looking for buddy {0} in {1}''s roster successful, buddy {2} ", new Object[]{buddy, ownerJID, state});

        return result;
    }

    @Override
    public JID[] getBuddies(XMPPResourceConnection session) throws NotAuthorizedException {
        BareJID ownerJID = this.sessionUtil.ownerOf(session);
        JID[] result = new JID[0];

        //this roster holds only visible and enabled, make new roster content otherwise content of default roster could be changed on merge
        RosterContent defaultContent = this.defaultRoster.roster(ownerJID.getDomain());

        if (defaultContent.find(ownerJID) != null) {
            //this roster contains all manually added, doesn't contains all manually removed and applies all name and groups change
            Map<BareJID, PersonalRoster.Buddy> personalizedContent = this.personalRoster.of(ownerJID).asMap();
            //do merge
            result = defaultContent.asArray(ownerJID, personalizedContent);

            LOG.log(Level.FINE,
                    "Returning dynamic roster array for user {0}, default size {1}, personal size {2}, result size {3}",
                    new Object[]{ownerJID, defaultContent.size(), personalizedContent.size(), result.length});
        } else {

            LOG.log(Level.FINE,
                    "Returning dynamic roster array for user {0}, array is empty as user is not in gal",
                    ownerJID);
        }

        return result;
    }

    @Override
    public List<Element> getRosterItems(XMPPResourceConnection session) throws NotAuthorizedException {
        List<Element> result = Collections.EMPTY_LIST;
        //this roster contains all manually added, doesn't contains all manually removed and applies all name and groups change
        BareJID ownerJID = this.sessionUtil.ownerOf(session);

        //this roster holds only visible and enabled, make new roster content otherwise content of default roster could be changed on merge
        RosterContent defaultContent = this.defaultRoster.roster(ownerJID.getDomain());

        if (defaultContent.find(ownerJID) != null) {
            Map<BareJID, PersonalRoster.Buddy> personalizedContent = this.personalRoster.of(ownerJID).asMap();
            //do merge
            result = defaultContent.asList(ownerJID, personalizedContent);
            LOG.log(Level.FINE,
                    "Returning dynamic roster list for user {0}, default size {1}, personal size {2}, result size {3}",
                    new Object[]{ownerJID, defaultContent.size(), personalizedContent.size(), result.size()});
        } else {
            LOG.log(Level.FINE,
                    "Returning dynamic roster list for user {0}, list is empty as user is not in gal",
                    ownerJID);
        }

        return result;
    }
}
