/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.xmpp.api.XmppException;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.TigaseDBException;
import tigase.xml.XMLUtils;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.impl.roster.RosterAbstract;

/**
 *
 * @author tskoda
 */
public class PersonalRosterListener implements IPersonalRosterListener {

    private static final Logger LOG = Logger.getLogger(PersonalRosterListener.class.getName());

    private final PersonalRoster personalRoster;
    private final DefaultRoster defaultRoster;
    private final GroupService groupService;
    private final SessionUtil userResolver;

    public PersonalRosterListener(PersonalRoster personalRoster, DefaultRoster defaultRoster, GroupService groupService) {
        this(personalRoster, defaultRoster, groupService, new SessionUtil());
    }

    PersonalRosterListener(PersonalRoster personalRoster, DefaultRoster defaultRoster, GroupService groupService, SessionUtil userResolver) {
        this.personalRoster = personalRoster;
        this.defaultRoster = defaultRoster;
        this.groupService = groupService;
        this.userResolver = userResolver;
    }

    @Override
    public void removeBuddy(XMPPResourceConnection connection, JID buddy) throws NotAuthorizedException, TigaseDBException {
        BareJID ownerJID = this.userResolver.ownerOf(connection);
        if (this.defaultRoster.roster(ownerJID.getDomain()).find(buddy.getBareJID()) != null) {
            try {
                this.personalRoster.of(ownerJID).markRemoved(buddy.getBareJID());
                LOG.log(Level.INFO, "Marking {0} as REMOVED from {1}''s automatic roster successful", new Object[]{buddy.getBareJID(), ownerJID});
            } catch (XmppException e) {
                throw new TigaseDBException("Marking " + buddy.getBareJID() + " as REMOVED from " + ownerJID + "'s automatic roster failed", e);
            }
        } else {
            LOG.log(Level.INFO, "Marking {0} as REMOVED from {1}''s automatic roster failed, no such user", new Object[]{buddy, ownerJID});
        }
    }

    @Override
    public void addBuddy(XMPPResourceConnection connection, JID buddy, String name, String[] gr) throws NotAuthorizedException, TigaseDBException {
        String nameUnescaped = (name != null)? XMLUtils.unescape(name): null;
        String[] groupsUnescaped = (gr != null)? new String[gr.length]: null;
        
        if (groupsUnescaped != null) {
            for (int i = 0; i < groupsUnescaped.length; i++) {
                groupsUnescaped[i] = gr[i] != null? XMLUtils.unescape(gr[i]): null;
            }
        }
        
        addBuddyUnescaped(connection, buddy, nameUnescaped, groupsUnescaped);
    }
    
    private void addBuddyUnescaped(XMPPResourceConnection connection, JID buddy, String name, String[] gr) throws NotAuthorizedException, TigaseDBException {
        BareJID ownerJID = this.userResolver.ownerOf(connection);
        BareJID buddyJID = buddy.getBareJID();
        RosterContent.Buddy defaultBuddy = this.defaultRoster.roster(ownerJID.getDomain()).find(buddy.getBareJID());

        if (defaultBuddy != null) {
            
            String nameToUpdate = null;
            if ((name != null) && (!name.equals(defaultBuddy.getUser().getFullName()))) {
                nameToUpdate = name;
            }

            String[] groupsToUpdate = null;
            if (gr == null) {
                groupsToUpdate = new String[]{null};
            } else if (!Arrays.equals(gr, new String[]{this.groupService.group(buddyJID.toString())})) {
                groupsToUpdate = gr;
            }

            if (nameToUpdate == null && groupsToUpdate == null) {
                try {
                    this.personalRoster.of(ownerJID).remove(buddyJID);
                    LOG.log(Level.INFO, "Modifying {0} in {1}''s roster successful, buddy is managed by automatic roster now", new Object[]{buddyJID, ownerJID});
                } catch (XmppException e) {
                    throw new TigaseDBException("Modifying " + buddy.getBareJID() + " in " + ownerJID + "'s roster failed", e);
                }

            } else {
                try {
                    this.personalRoster.of(ownerJID).update(buddyJID, nameToUpdate, groupsToUpdate);
                    LOG.log(Level.INFO, "Modifying {0} in {1}''s roster successful", new Object[]{buddyJID, ownerJID});
                } catch (XmppException e) {
                    throw new TigaseDBException("Modifying " + buddy.getBareJID() + " in " + ownerJID + "'s roster failed", e);
                }
            }

        } else {
            LOG.log(Level.INFO, "Modifying {0} in {1}''s roster failed, no such user", new Object[]{buddy, ownerJID});
        }
    }

    @Override
    public void added(XMPPResourceConnection connection, JID buddy) throws NotAuthorizedException, TigaseDBException {
        BareJID ownerJID = this.userResolver.ownerOf(connection);
        try {
            //buddy is expected not to be in default roster
            this.personalRoster.of(ownerJID).markAdded(buddy.getBareJID());
            LOG.log(Level.INFO, "Marking {0} as added to {1}''s roster successful", new Object[]{buddy.getBareJID(), ownerJID});
        } catch (XmppException e) {
            LOG.log(Level.INFO, "Marking " + buddy.getBareJID() + " as ADDED to " + ownerJID + "'s roster failed", e);
            //do not report it outside, we dont want to block addind the user
        }
    }

    @Override
    public void removed(XMPPResourceConnection connection, JID buddy) throws NotAuthorizedException, TigaseDBException {
        BareJID ownerJID = this.userResolver.ownerOf(connection);
        try {
            this.personalRoster.of(ownerJID).remove(buddy.getBareJID());
            LOG.log(Level.INFO, "Removing {0} from {1}''s personal roster successful", new Object[]{buddy.getBareJID(), ownerJID});
        } catch (XmppException e) {
           LOG.log(Level.INFO, "Removing " + buddy.getBareJID() + " from " + ownerJID + "'s personal roster failed", e);
            //do not report it outside, we dont want to block addind the user
        }
    }

    @Override
    public void setBuddySubscription(XMPPResourceConnection connection, RosterAbstract.SubscriptionType subscriptionType, JID buddy) throws NotAuthorizedException, TigaseDBException {
        //do nothing, it is repair mechanism of wrong subscription, we dont store any subscription, it is counted either both or none
    }

}
