/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.derby;

import com.kerio.im.tigase.derby.DerbyDatabase.DATA;
import com.kerio.im.tigase.derby.DerbyDatabase.PROCEDURE;
import com.kerio.im.tigase.derby.DerbyDatabase.TABLE;
import com.kerio.im.tigase.derby.DerbyDatabase.SCHEMA;
import com.kerio.im.tigase.derby.version.Version;
import com.kerio.im.xmpp.api.XmppException;
import java.io.File;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Arrays;
import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

/**
 *
 * @author tskoda
 */
@RunWith(JUnit4.class)
public class DerbyCheckerTest {

    @Rule
    public ExpectedException exception = ExpectedException.none();

    public DerbyCheckerTest() {
    }

    @Test
    public void testCheckIfExists() throws Exception {
        File tempDirectory = DerbyUtilFile.createTempFolder();

        try {

            DerbyChecker derbyDb = new DerbyChecker(tempDirectory.getParentFile(), tempDirectory.getName());
            derbyDb.checkIfExists(); //throws exception if not
        } finally {
            DerbyUtilFile.deleteDerbyFolder(tempDirectory);
        }
    }

    @Test
    public void testCheckIfExists_NotExists() throws Exception {
        File tempDirectory = DerbyUtilFile.createTempFolder();

        try {
            DerbyChecker derbyDb = new DerbyChecker(tempDirectory, "doesnotexist");
            this.exception.expect(DerbyChecker.NotExistException.class);
            derbyDb.checkIfExists(); //throws NotExistException
        } finally {
            DerbyUtilFile.deleteDerbyFolder(tempDirectory);
        }
    }

    @Test
    public void testCheckIfExists_NotDirectory() throws Exception {
        File tempFile = File.createTempFile("dbs_", null);

        try {
            DerbyChecker derbyDb = new DerbyChecker(tempFile.getParentFile(), tempFile.getName());
            this.exception.expect(XmppException.class);
            derbyDb.checkIfExists();
        } finally {
            tempFile.delete();
        }
    }

    @Test
    public void testCheckIfBootable() throws Exception {
        File tempDirectory = DerbyUtilFile.createTempFolder();
        File derbyDirectory = new File(tempDirectory, "testDB");

        try {
            DerbyUtilFile.createDatabase(derbyDirectory);

            DerbyChecker checker = new DerbyChecker(tempDirectory, derbyDirectory.getName());
            Connection connection = checker.checkIfBootable();
            try {
                assertNotNull(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }

        } finally {

            DerbyUtilFile.shutdown(derbyDirectory);
            DerbyUtilFile.deleteDerbyFolder(derbyDirectory);
            DerbyUtilFile.deleteFolderWithFiles(tempDirectory);
        }
    }

    @Test
    public void testCheckIfBootable_NonExistent() throws Exception {
        File tempDirectory = DerbyUtilFile.createTempFolder();

        try {
            DerbyChecker derbyDb = new DerbyChecker(tempDirectory, "fake");
            this.exception.expect(SQLException.class);
            Connection connection = derbyDb.checkIfBootable(); //throws SQLException
            DerbyUtilCommon.close(connection);
        } finally {
            DerbyUtilFile.deleteDerbyFolder(tempDirectory);
        }
    }

    @Test
    public void testCheckIfBootable_Destroyed() throws Exception {
        File tempDirectory = DerbyUtilFile.createTempFolder();
        File derbyDirectory = new File(tempDirectory, "testDB");

        try {
            try {
                DerbyUtilFile.createDatabase(derbyDirectory);
            } finally {
                DerbyUtilFile.shutdown(derbyDirectory); //to be able to delete some folders
            }
            DerbyUtilFile.deleteFolderWithFiles(new File(derbyDirectory, "seg0"));
            DerbyChecker derbyDb = new DerbyChecker(tempDirectory, derbyDirectory.getName());
            this.exception.expect(SQLException.class);
            Connection connection = derbyDb.checkIfBootable();
            DerbyUtilCommon.close(connection); //in case exception was not thrown

        } finally {
            DerbyUtilFile.deleteDerbyFolder(derbyDirectory);
            DerbyUtilFile.deleteFolderWithFiles(tempDirectory);

        }
    }

    @Test
    public void testCheckAllTablesExist_None() throws Exception {
        String database = "testdb";

        try {

            Connection connection = DerbyUtilMemory.create(database);

            try {
                DerbyChecker derbyDb = new DerbyChecker("memory:testdb");
                this.exception.expect(XmppException.class);
                derbyDb.checkAllTablesExist(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }

        } finally {
            DerbyUtilMemory.shutdown(database);
        }

    }

    @Test
    public void testCheckAllTables_AllExceptOne() throws Exception {
        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);

            try {
                TABLE[] allExceptOne = Arrays.copyOf(TABLE.values(), TABLE.values().length - 1);
                DerbyUtilCommon.init(connection, allExceptOne, null, null);
                //omitting XMPP_STANZA on purpose

                DerbyChecker derbyDb = new DerbyChecker("memory:testdb");
                this.exception.expect(XmppException.class);
                derbyDb.checkAllTablesExist(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }

        } finally {
            DerbyUtilMemory.shutdown(database);
        }

    }

    @Test
    public void testCheckAllTables() throws Exception {
        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);
            try {
                DerbyUtilCommon.init(connection, TABLE.values(), null, null);

                DerbyChecker derbyDb = new DerbyChecker("memory:testdb");
                derbyDb.checkAllTablesExist(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }

        } finally {
            DerbyUtilMemory.shutdown(database);
        }

    }

    @Test
    public void testCheckAllProcedures_ExistNone() throws Exception {
        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);
            try {
                DerbyChecker derbyDb = new DerbyChecker("memory:testdb");
                this.exception.expect(XmppException.class);
                derbyDb.checkAllProceduresExist(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }

        } finally {
            DerbyUtilMemory.shutdown(database);
        }
    }

    @Test
    public void testCheckAllProceduresExist_AllExceptOne() throws Exception {
        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);

            try {
                //omitting the zero one for purpose
                PROCEDURE[] allExceptOne = Arrays.copyOfRange(PROCEDURE.values(), 1, PROCEDURE.values().length);
                DerbyUtilCommon.init(connection,
                        TABLE.values(),
                        allExceptOne,
                        null);

                DerbyChecker derbyDb = new DerbyChecker("memory:testdb");
                this.exception.expect(XmppException.class);
                derbyDb.checkAllProceduresExist(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }
        } finally {
            DerbyUtilMemory.shutdown(database);
        }
    }

    @Test
    public void testCheckAllProceduresExist() throws Exception {
        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);
            try {
                DerbyUtilCommon.init(connection, TABLE.values(), PROCEDURE.values(), null);

                DerbyChecker derbyDb = new DerbyChecker("memory:testdb");
                derbyDb.checkAllProceduresExist(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }

        } finally {

            DerbyUtilMemory.shutdown(database);

        }
    }

    @Test
    public void testCheckData() throws Exception {

        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);

            try {
                DerbyUtilCommon.init(connection, TABLE.values(), PROCEDURE.values(), DATA.values());

                DerbyChecker derbyDB = new DerbyChecker("memory:testdb");
                derbyDB.checkDataExists(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }
        } finally {
            DerbyUtilMemory.shutdown(database);
        }
    }

    @Test
    public void testCheckData_Empty() throws Exception {

        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);

            try {
                DerbyUtilCommon.init(connection, TABLE.values(), PROCEDURE.values(), null);

                DerbyChecker derbyDB = new DerbyChecker("memory:testdb");
                this.exception.expect(XmppException.class);
                derbyDB.checkDataExists(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }
        } finally {
            DerbyUtilMemory.shutdown(database);
        }
    }

    @Test
    public void testCheckData_Incomplete() throws Exception {

        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);

            try {
                DerbyUtilCommon.init(connection, TABLE.values(), PROCEDURE.values(), new DATA[]{DATA.DB_PROPERTIES});

                DerbyChecker derbyDB = new DerbyChecker("memory:testdb");
                this.exception.expect(XmppException.class);
                derbyDB.checkDataExists(connection);
            } finally {
                DerbyUtilCommon.close(connection);
            }
        } finally {

            DerbyUtilMemory.shutdown(database);
        }
    }

    @Test
    public void testInvalidSchema() throws Exception {
        String database = "testdb";

        try {
            Connection connection = DerbyUtilMemory.create(database);

            try {
                //init db, omit schema
                DerbyUtilCommon.init(connection, TABLE.values(), PROCEDURE.values(), new DATA[]{DATA.DB_PROPERTIES});
                //set previous schema
                DerbyUtilCommon.init(connection, SCHEMA.UPDATE_DB_SCHEMA.create(Version.V_5_2.toString()));
                //check that schema is not current
                assertThat(Version.CURRENT, is(not(Version.V_5_3)));

                DerbyChecker derbyDB = new DerbyChecker("memory:testdb");
                try {
                    derbyDB.checkSchemaVersion(connection);
                    fail();
                } catch (DerbyChecker.ObsoleteSchemaException e) {
                    assertThat(e.getFound(), is(Version.V_5_2));
                }
            } finally {
                DerbyUtilCommon.close(connection);
            }
        } finally {

            DerbyUtilMemory.shutdown(database);
        }
    }
}
