/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.derby;

import static com.kerio.im.tigase.derby.DerbyUtilCommon.close;
import com.kerio.im.tigase.derby.DerbyDatabase.TABLE;
import com.kerio.im.tigase.derby.DerbyDatabase.PROCEDURE;
import com.kerio.im.tigase.derby.DerbyDatabase.DATA;
import com.kerio.im.tigase.derby.DerbyDatabase.SCHEMA;
import com.kerio.im.tigase.derby.version.Version;
import com.kerio.im.xmpp.api.XmppException;
import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.MessageFormat;
import java.util.Collections;
import java.util.Date;
import java.util.EnumSet;
import org.apache.derby.jdbc.EmbeddedDataSource40;
import org.hamcrest.CoreMatchers;
import org.junit.After;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;
import tigase.db.DBInitException;
import tigase.db.DataTooBigException;
import static tigase.db.NonAuthUserRepository.OFFLINE_DATA_NODE;
import tigase.db.NonAuthUserRepositoryImpl;
import tigase.db.TigaseDBException;
import tigase.db.UserNotFoundException;
import tigase.db.jdbc.JDBCRepository;
import tigase.server.Message;
import tigase.server.amp.MsgRepository;
import tigase.util.TigaseStringprepException;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;
import static tigase.xmpp.XMPPImplIfc.CLIENT_XMLNS;
import tigase.xmpp.impl.OfflineMessages;

/**
 *
 * @author tskoda
 */
@RunWith(JUnit4.class)
public class DerbyHandlerTest {

    private static final String LEGACY_OFFLINE_COUNT = "select count(p.nid) from tig_pairs p, tig_nodes n where p.NID = n.nid and n.node = 'msgoffline'";
    private static final String AMP_OFFLINE_COUNT = "select count(o.MSG_ID) from APP.MSG_HISTORY o";

    private File folder;
    @Rule
    public ExpectedException exception = ExpectedException.none();
    private File completeDatab;

    @Before
    public void before() throws IOException {
        this.folder = DerbyUtilFile.createTempFolder();
        this.folder.deleteOnExit(); //folders and files are deleted in reverse order
    }

    @After
    public void after() throws IOException {
    }

    @Test
    public void testRemoveNonExisting() throws IOException {
        File nonExisting = new File(this.folder, "nonExisting");
        DerbyHandler handler = new DerbyHandler(this.folder, nonExisting.getName());
        assertFalse(nonExisting.exists());
        assertTrue(handler.remove());
        assertFalse(nonExisting.exists());
    }

    @Test
    public void testRemoveEmpty() throws IOException {
        File empty = DerbyUtilFile.createTempFolder(this.folder);
        try {
            DerbyHandler handler = new DerbyHandler(this.folder, empty.getName());
            assertTrue(empty.exists());
            assertTrue(handler.remove());
            assertFalse(empty.exists());
        } finally {
            empty.deleteOnExit();
        }
    }

    @Test
    public void testRemove() throws SQLException, IOException {
        File completeDatabase = new File(this.folder, "completeDatabase");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
            assertTrue(new File(completeDatabase, "seg0").isDirectory());
            assertTrue(new File(completeDatabase, "log").isDirectory());
            assertTrue(handler.remove());
            assertFalse(completeDatabase.exists());
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveNoSeq() throws IOException, SQLException {
        File completeDatabase = new File(this.folder, "noSeq");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            File seq = new File(completeDatabase, "seg0");
            DerbyUtilFile.deleteFolderWithFiles(seq);
            DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
            assertFalse(seq.exists());
            assertTrue(new File(completeDatabase, "log").isDirectory());
            assertTrue(handler.remove());
            assertFalse(completeDatabase.exists());
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveNoLog() throws SQLException, IOException {
        File completeDatabase = new File(this.folder, "noLog");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            File log = new File(completeDatabase, "log");
            DerbyUtilFile.deleteFolderWithFiles(log);
            DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
            assertFalse(log.exists());
            assertTrue(new File(completeDatabase, "seg0").isDirectory());
            assertTrue(handler.remove());
            assertFalse(completeDatabase.exists());
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveExtraExpectedFiles() throws SQLException, IOException {
        File completeDatabase = new File(this.folder, "completeDatabase");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            assertTrue(new File(completeDatabase, "dbex.lck").createNewFile());
            assertTrue(new File(completeDatabase, "BACKUP.HISTORY").createNewFile());

            DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
            assertTrue(new File(completeDatabase, "seg0").isDirectory());
            assertTrue(new File(completeDatabase, "log").isDirectory());

            assertTrue(handler.remove());
            assertFalse(completeDatabase.exists());
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveExtraFolder() throws IOException, SQLException {
        File completeDatabase = new File(this.folder, "extraFolder");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            File extraFolder = DerbyUtilFile.createTempFolder(completeDatabase);
            try {
                assertTrue(extraFolder.exists());
                DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
                assertFalse(handler.remove());
                assertTrue(extraFolder.exists());
                assertTrue(completeDatabase.exists());
                assertFalse(new File(completeDatabase, "seg0").exists());
                assertFalse(new File(completeDatabase, "log").exists());
            } finally {
                DerbyUtilFile.deleteDerbyFolder(extraFolder);
            }
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveExtraFolderInSeg() throws SQLException, IOException {
        File completeDatabase = new File(this.folder, "completeDatabase");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            File extraInSeg = DerbyUtilFile.createTempFolder(new File(completeDatabase, "seg0"));
            try {
                assertTrue(extraInSeg.isDirectory());
                DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
                assertTrue(new File(completeDatabase, "seg0").isDirectory());
                assertTrue(new File(completeDatabase, "log").isDirectory());
                assertFalse(handler.remove());
                assertTrue(completeDatabase.exists());
                assertFalse(new File(completeDatabase, "log").exists());
                assertTrue(new File(completeDatabase, "seg0").exists());
                assertTrue(extraInSeg.exists());
            } finally {
                DerbyUtilFile.deleteDerbyFolder(extraInSeg);
            }
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveExtraFile() throws SQLException, IOException {
        File completeDatabase = new File(this.folder, "completeDatabase");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            File extraInSeg = new File(completeDatabase, "extraFile");
            extraInSeg.createNewFile();
            try {
                assertTrue(extraInSeg.isFile());
                DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
                assertTrue(new File(completeDatabase, "seg0").isDirectory());
                assertTrue(new File(completeDatabase, "log").isDirectory());
                assertFalse(handler.remove());
                assertTrue(completeDatabase.exists());
                assertFalse(new File(completeDatabase, "seg").exists());
                assertFalse(new File(completeDatabase, "log").exists());
                assertTrue(extraInSeg.exists());
            } finally {
                DerbyUtilFile.deleteDerbyFolder(extraInSeg);
            }
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testRemoveExtraFileInSeq() throws IOException, SQLException {
        File completeDatabase = new File(this.folder, "completeDatabase");

        try {
            DerbyUtilFile.createDatabase(completeDatabase);
            File extraInSeg = new File(new File(completeDatabase, "seg0"), "extraFile");
            extraInSeg.createNewFile();
            try {
                assertTrue(extraInSeg.isFile());
                DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
                assertTrue(new File(completeDatabase, "seg0").isDirectory());
                assertTrue(new File(completeDatabase, "log").isDirectory());
                assertTrue(handler.remove());
                assertFalse(completeDatabase.exists());
            } finally {
                DerbyUtilFile.deleteDerbyFolder(extraInSeg);
            }
        } finally {
            DerbyUtilFile.deleteDerbyFolder(completeDatabase);
        }
    }

    @Test
    public void testUpgradeSchema() throws IOException, SQLException, TigaseStringprepException, UserNotFoundException, TigaseDBException, DBInitException, DataTooBigException {
        /*
         * implemented are only follofing upgrades:
         * 5_2 -> 5_4
         * 5_3 -> 5_4
         */
        for (Version from : Version.values()) {
            switch (from) {
                case V_5_2: {
                    System.out.println("Testing upgrade from 5_2 (with and without amp)");
                    for (int i = 0; i < 2; i++) {
                        boolean amp = (i == 0);
                        File db = new File(this.folder, "extraFolder_5_2" + (amp ? "_amp" : ""));

                        try {
                            prepareSchema_5_2(db);
                            prepareOfflineMessages(db, amp);

                            DerbyHandler handler = new DerbyHandler(this.folder, db.getName());
                            handler.upgradeSchema(from);

                            verifyAllSchemas(db, amp);
                        } finally {
                            DerbyUtilFile.deleteDerbyFolder(db);
                        }
                    }
                    break;
                }
                case V_5_3: {
                    System.out.println("Testing upgrade from 5_3 (with and without amp)");
                    for (int i = 0; i < 2; i++) {
                        boolean amp = (i == 0);
                        File db = new File(this.folder, "extraFolder_5_3" + (amp ? "_amp" : ""));
                        try {
                            prepareSchema_5_3(db);
                            prepareOfflineMessages(db, amp);

                            DerbyHandler handler = new DerbyHandler(this.folder, db.getName());
                            handler.upgradeSchema(from);

                            verifyAllSchemas(db, amp);
                        } finally {
                            DerbyUtilFile.deleteDerbyFolder(db);
                        }
                    }
                    break;
                }

                case V_5_4: {
                    System.out.println("Testing upgrade from 5_4 ... ");
                    break;
                }
                case CURRENT: {
                    System.out.println("Testing upgrade from CURRENT ... ");
                    break;
                }
                default:
                    File completeDatabase = new File(this.folder, "extraFolder_" + from.toString());
                    DerbyUtilFile.createDatabase(completeDatabase);
                    DerbyHandler handler = new DerbyHandler(this.folder, completeDatabase.getName());
                    //all other should fail
                    try {
                        handler.upgradeSchema(from);
                        fail();
                    } catch (XmppException e) {
                        assertTrue(e.getMessage().startsWith(MessageFormat.format("Upgrading {0} database schema from {1} to {2} failed", handler.getDbPath(), from, Version.CURRENT)));
                    } finally {
                        DerbyUtilFile.deleteDerbyFolder(completeDatabase);
                    }
            }
        }
    }

    public void prepareSchema_5_2(File database) throws SQLException, TigaseDBException, DBInitException, TigaseStringprepException {

        DerbyUtilFile.createDatabase(database);

        EmbeddedDataSource40 ds = new EmbeddedDataSource40();
        ds.setDatabaseName(database.getAbsolutePath());
        Connection connection = ds.getConnection();
        try {
            //fill db with 5_2 version - omit PERSONAL ROSTER TABLE
            System.out.println("- initializing db to version 5_2");
            DerbyUtilCommon.init(
                    connection,
                    new TABLE[]{TABLE.TIG_USERS, TABLE.TIG_NODES, TABLE.TIG_PAIRS, TABLE.TIG_SHORT_NEWS, TABLE.XMPP_STANZA},
                    PROCEDURE.values(),
                    DATA.values());

            //fill db with 5_2 version - set proper schema
            DerbyUtilCommon.init(
                    connection,
                    SCHEMA.UPDATE_DB_SCHEMA.create(Version.V_5_2.toString()));

            //personal roster table is not present
            PreparedStatement tableExistBefore = connection.prepareStatement(DerbyChecker.SELECT_TABLE);
            try {
                tableExistBefore.setString(1, TABLE.PERSONAL_ROSTER.name);
                ResultSet tableRS = tableExistBefore.executeQuery();
                try {
                    assertTrue(tableRS.next());
                    assertThat(tableRS.getInt(1), CoreMatchers.is(0));
                } finally {
                    close(tableRS);
                }

            } finally {
                close(tableExistBefore);
            }

        } finally {
            DerbyUtilCommon.close(connection);
        }

    }

    public void prepareSchema_5_3(File database) throws SQLException, DBInitException, TigaseStringprepException, UserNotFoundException, TigaseDBException {

        DerbyUtilFile.createDatabase(database);

        EmbeddedDataSource40 ds = new EmbeddedDataSource40();
        ds.setDatabaseName(database.getAbsolutePath());
        Connection connection = ds.getConnection();
        try {
            //fill db with 5_3 version - there is no change between 5_3 and current
            System.out.println("- initializing db to version 5_3");
            DerbyUtilCommon.init(
                    connection,
                    TABLE.values(),
                    PROCEDURE.values(),
                    DATA.values());

            //fill db with 5_3 version - set proper schema
            DerbyUtilCommon.init(
                    connection,
                    SCHEMA.UPDATE_DB_SCHEMA.create(Version.V_5_3.toString()));

        } finally {
            DerbyUtilCommon.close(connection);
        }

    }

    private void prepareOfflineMessages(File path, boolean amp) throws DBInitException, TigaseDBException, TigaseStringprepException, SQLException, DataTooBigException {
        JID user1 = JID.jidInstanceNS("user1", "domain", "resource1");
        JID user2 = JID.jidInstanceNS("user2", "domain", "resource1");
        Date hour = new Date(new Date().getTime() + (60 * 60 * 1000));

        //legacy offline messages, always present
        {
            JDBCRepository jdbc = new JDBCRepository();
            jdbc.initRepository("jdbc:derby:" + path, Collections.EMPTY_MAP);
            jdbc.addUser(user1.getBareJID());
            jdbc.addUser(user2.getBareJID());

            //store some offline messages with old offline provider
            OfflineMessages.MsgRepositoryImpl repository = new OfflineMessages().new MsgRepositoryImpl(
                    new NonAuthUserRepositoryImpl(jdbc, BareJID.bareJIDInstanceNS(null, "domain"), false),
                    null);
            repository.storeMessage(user1, user2, hour, message("offline 1", user1, user2));
            repository.storeMessage(user1, user2, hour, message("offline 2", user1, user2));
            repository.storeMessage(user2, user1, hour, message("offline 3", user2, user1));
        }

        //AMP offlline messages, optional
        if (amp) {
            MsgRepository repository = new MsgRepository();
            repository.initRepository("jdbc:derby:" + path, Collections.EMPTY_MAP);
            repository.storeMessage(user1, user2, hour, message("offline 4", user1, user2));
            repository.storeMessage(user1, user2, hour, message("offline 5", user1, user2));
            repository.storeMessage(user2, user1, hour, message("offline 6", user2, user1));
        }

        //verify both
        EmbeddedDataSource40 ds = new EmbeddedDataSource40();
        ds.setDatabaseName(path.getAbsolutePath());
        Connection connection = ds.getConnection();
        try {
            assertEquals(3, offlineCount(connection, LEGACY_OFFLINE_COUNT));

            if (amp) {
                assertEquals(3, offlineCount(connection, AMP_OFFLINE_COUNT));
            }
        } finally {
            DerbyUtilCommon.close(connection);
        }

    }

    private int offlineCount(Connection connection, String query) throws SQLException {
        PreparedStatement countStatement = connection.prepareStatement(query);
        try {
            ResultSet countResultSet = countStatement.executeQuery();
            try {
                countResultSet.next();
                return countResultSet.getInt(1);
            } finally {
                DerbyUtil.close(countResultSet);
            }
        } finally {
            DerbyUtil.close(countStatement);
        }
    }

    private Element message(String text, JID from, JID to) throws TigaseStringprepException {
        Element messageElement = new Element("message", new Element[]{new Element("body", text)}, null, null);
        messageElement.setXMLNS(CLIENT_XMLNS);
        messageElement.addAttribute("from", from.toString());
        messageElement.addAttribute("to", to.toString());
        messageElement.addAttribute("type", StanzaType.chat.toString());

        return messageElement;
    }

    public void verifyAllSchemas(final File db, boolean amp) throws SQLException {

        EmbeddedDataSource40 ds = new EmbeddedDataSource40();
        ds.setDatabaseName(db.getAbsolutePath());
        Connection connection = ds.getConnection();
        try {
            //changes of 5.3
            {
                System.out.println("- verifying " + Version.V_5_3.toString());

                //personal roster table is present
                PreparedStatement tableExistAfter = connection.prepareStatement(DerbyChecker.SELECT_TABLE);
                try {
                    tableExistAfter.setString(1, TABLE.PERSONAL_ROSTER.name);
                    ResultSet tableRS = tableExistAfter.executeQuery();
                    try {
                        assertTrue(tableRS.next());
                        assertThat(tableRS.getInt(1), CoreMatchers.is(1));
                        System.out.println("- personal roster table: ok");
                    } finally {
                        close(tableRS);
                    }

                } finally {
                    close(tableExistAfter);
                }
            }

            //changes of 5.4
            {
                System.out.println("- verifying " + Version.V_5_4.toString());
                Statement schemaStatement = connection.createStatement();
                try {
                    ResultSet schemaResult = schemaStatement.executeQuery(JDBCRepository.DERBY_GETSCHEMAVER_QUERY);
                    try {
                        if (schemaResult.next()) {
                            assertThat(schemaResult.getString(1), CoreMatchers.is(Version.V_5_4.toString()));
                        }
                    } finally {
                        close(schemaResult);
                    }
                } finally {
                    close(schemaStatement);
                }

                //no old offline messages
                assertEquals(0, offlineCount(connection, LEGACY_OFFLINE_COUNT));

                //no new offline messages
                if (amp) {
                    assertEquals(0, offlineCount(connection, AMP_OFFLINE_COUNT));
                }
            }
        } finally {
            DerbyUtilCommon.close(connection);
        }
    }

//    
//    public void checkExceptFolder(DerbyChecker checker) throws SQLException {
//        Connection connection = checker.checkIfBootable();
//        try {
//            checker.checkAllProceduresExist(connection);
//            checker.checkAllTablesExist(connection);
//            checker.checkDataExists(connection);
//        } finally {
//            close(connection);
//        }
//    }
//    @Test
//    public void testCreateDatabase() throws Exception {
//
//        try {
//            DerbyHandler derbyDB = new DerbyHandler("memory:testdb");
//            derbyDB.createDatabase();
//
//            DerbyChecker checker = new DerbyChecker("memory:testdb");
//            checkExceptFolder(checker);
//
//        } finally {
//            shutdown("jdbc:derby:memory:testdb");
//        }
//
//    }
//    @Test
//    public void testRepair_NonExisting() throws SQLException {
//        String connectionString = "jdbc:derby:memory:testdb";
//        try {
//            DerbyChecker checker = spy(new DerbyChecker("memory:testdb"));
//
//            //because using in memory database
//            doThrow(new DerbyChecker.NotExistException("Database does not exist", Code.RESOURCE_ERROR)).when(checker).checkIfExists();
//            DerbyHandler derbyDB = new DerbyHandler("memory:testdb");
//            derbyDB.repair();
//
//            Connection connection = checker.checkIfBootable();
//            try {
//                checker.checkAllProceduresExist(connection);
//                checker.checkAllTablesExist(connection);
//                checker.checkDataExists(connection);
//            } finally {
//                close(connection);
//            }
//        } finally {
//            shutdown(connectionString);
//        }
//
//    }
//    @Test
//    public void testRepair_Existing() throws SQLException {
//
//        String connectionString = "jdbc:derby:memory:testdb";
//        Connection connection = DriverManager.getConnection(connectionString + ";create=true");;
//        try {
//            DerbyChecker checker = spy(new DerbyChecker("memory:testdb"));
//            //because using in memory database
//            doThrow(new DerbyChecker.NotExistException("Database does not exist", Code.RESOURCE_ERROR)).when(checker).checkIfExists();
//
//            DerbyHandler derbyDB = new DerbyHandler("memory:testdb", checker);
//            initdb(connection, DerbyDatabase.TABLE.values(), DerbyDatabase.PROCEDURE.values(), DerbyDatabase.DATA.values());
//
//            derbyDB.repair();
//
//            Connection checkConnection = checker.checkIfBootable();
//            try {
//                checker.checkAllProceduresExist(checkConnection);
//                checker.checkAllTablesExist(checkConnection);
//                checker.checkDataExists(checkConnection);
//            } finally {
//                close(connection);
//            }
//        } finally {
//            close(connection);
//            shutdown(connectionString);
//        }
//
//    }
//
//    @Test
//    public void testRepair_NoTables() throws SQLException {
//        String connectionString = "jdbc:derby:memory:testdb";
//        Connection connection = DriverManager.getConnection(connectionString + ";create=true");;
//        try {
//            DerbyChecker checker = new DerbyChecker("memory:testdb");
//            DerbyHandler derbyDB = new DerbyHandler("memory:testdb", checker);
//            initdb(connection, null, null, null);
//
//            this.exception.expect(XmppException.class);
//            derbyDB.repair();
//
//        } finally {
//            close(connection);
//            shutdown(connectionString);
//        }
//    }
//
//    @Test
//    public void testRepair_NoProcedures() throws SQLException {
//
//        String connectionString = "jdbc:derby:memory:testdb";
//        Connection connection = DriverManager.getConnection(connectionString + ";create=true");
//        try {
//
//            DerbyChecker checker = new DerbyChecker("memory:testdb");
//            DerbyHandler derbyDB = new DerbyHandler("memory:testdb", checker);
//            initdb(connection, DerbyDatabase.TABLE.values(), null, null);
//
//            this.exception.expect(XmppException.class);
//            derbyDB.repair();
//
//        } finally {
//            close(connection);
//            shutdown(connectionString);
//        }
//    }
//
//    @Test
//    public void testRepair_NoData() throws SQLException {
//        String connectionString = "jdbc:derby:memory:testdb";
//
//        Connection connection = DriverManager.getConnection(connectionString + ";create=true");
//        try {
//
//            DerbyChecker checker = new DerbyChecker("memory:testdb");
//            DerbyHandler derbyDB = new DerbyHandler("memory:testdb", checker);
//            initdb(connection, DerbyDatabase.TABLE.values(), DerbyDatabase.PROCEDURE.values(), null);
//
//            this.exception.expect(XmppException.class);
//            derbyDB.repair();
//
//        } finally {
//            close(connection);
//            shutdown(connectionString);
//        }
//    }
}
