/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.user.UserMapper;
import com.kerio.im.tigase.user.UserUtil;
import java.util.HashMap;
import java.util.Map;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.mockito.Mockito.*;
import tigase.xmpp.BareJID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;

/**
 *
 * @author tskoda
 */
public class PersonalRosterListenerTest {

    private static final UserUtil as = new UserUtil();
    private static final ElementMapper map = new ElementMapper(new TestGroupService(), new UserMapper());

    private Map<BareJID, RosterContent.Buddy> defaultRosterContentMap;
    private RosterContent defaultRosterContent;

    private PersonalRoster.PersonalUserRoster personalRosterContent;
    private PersonalRoster personalRoster;
    private DefaultRoster defaultRoster;
    private GroupService groupService;
    private SessionUtil sessionUtil;

    private ElementMapper mapper;

    public PersonalRosterListenerTest() {
    }

    @Before
    public void setUp() {
        this.mapper = new ElementMapper(new TestGroupService(), new UserMapper());
        this.defaultRosterContentMap = new HashMap<BareJID, RosterContent.Buddy>();
        this.defaultRosterContent = new RosterContent(defaultRosterContentMap, mapper, 10, false);
        this.defaultRoster = mock(DefaultRoster.class);
        when(this.defaultRoster.roster(any(String.class))).thenReturn(defaultRosterContent);

        this.personalRosterContent = mock(PersonalRoster.PersonalUserRoster.class);
        this.personalRoster = mock(PersonalRoster.class);
        when(this.personalRoster.of(any(BareJID.class))).thenReturn(personalRosterContent);

        this.groupService = new TestGroupService();

        this.sessionUtil = mock(SessionUtil.class);
    }

    @After
    public void tearDown() {
    }

    private XMPPResourceConnection connection(final BareJID jid) throws NotAuthorizedException {
        XMPPResourceConnection result = mock(XMPPResourceConnection.class);
        when(this.sessionUtil.ownerOf(any(XMPPResourceConnection.class))).thenReturn(jid);
        return result;
    }

    @Test
    public void testRemoveBuddyNotDefault() throws Exception {

        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);

        //user 2 is removing user 1 from his default roster
        listener.removeBuddy(connection(as.bare(2)), as.jid(1, 0));

        verify(this.personalRoster, times(0)).of(any(BareJID.class));
        verify(this.personalRosterContent, times(0)).markRemoved(any(BareJID.class));
    }

    @Test
    public void testRemoveBuddyDefault() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        //user 2 is removing user 1 from his default roster
        listener.removeBuddy(connection(as.bare(2)), as.jid(1, 0));

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(1)).markRemoved(as.bare(1));
    }

    @Test
    public void testAddBuddyNoDefault() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);

        //user 2 is removing user 1 from his default roster
        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), "custom name", new String[]{"custom group"});

        verify(this.personalRoster, times(0)).of(any(BareJID.class));
        verify(this.personalRosterContent, times(0)).update(any(BareJID.class), any(String.class), any(String[].class));

    }

    @Test
    public void testAddBuddyDefaultNoChange() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        String name = as.user(1).getFullName();
        String[] groups = new String[]{TestGroupService.DEFAULT_GROUP};

        //user 2 is removing user 1 from his default roster
        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), name, groups);

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(0)).update(any(BareJID.class), any(String.class), any(String[].class));
        verify(this.personalRosterContent, times(1)).remove(any(BareJID.class));

    }

    @Test
    public void testAddBuddyNameChange() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        String name = "custom name";
        String[] groups = new String[]{TestGroupService.DEFAULT_GROUP};

        //user 2 is removing user 1 from his default roster
        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), name, groups);

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(1)).update(as.bare(1), name, null);

    }

    @Test
    public void testAddBuddyGroupsChange() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        String name = as.user(1).getFullName();
        String[] groups = new String[]{"custom 1", "custom 2"};

        //user 2 is removing user 1 from his default roster
        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), name, groups);

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(1)).update(as.bare(1), null, groups);

    }

    @Test
    public void testAddBuddyGroupsChangeEmpty() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        String name = as.user(1).getFullName();
        String[] groups = null;

        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), name, groups);

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(1)).update(as.bare(1), null, new String[]{null});
    }

    @Test
    public void testAddBuddyAllChange() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        String name = "custom name";
        String[] groups = new String[]{"custom 1", "custom 2"};

        //user 2 is removing user 1 from his default roster
        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), name, groups);

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(1)).update(as.bare(1), name, groups);

    }

    @Test
    public void testRemoved() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));
        this.defaultRosterContentMap.put(as.bare(2), map.value(as.user(2)));

        //not in default
        listener.removed(connection(as.bare(1)), as.jid(3, 0));
        verify(this.personalRoster, times(1)).of(as.bare(1));
        verify(this.personalRosterContent, times(1)).remove(as.bare(3));

        //in default
        listener.removed(connection(as.bare(1)), as.jid(2, 0));
        verify(this.personalRoster, times(2)).of(as.bare(1));
        verify(this.personalRosterContent, times(1)).remove(as.bare(2));
    }

    @Test
    public void testAdded() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));
        this.defaultRosterContentMap.put(as.bare(2), map.value(as.user(2)));

        //not in default
        listener.added(connection(as.bare(1)), as.jid(3, 0));
        verify(this.personalRoster, times(1)).of(as.bare(1));
        verify(this.personalRosterContent, times(1)).markAdded(as.bare(3));

        //in default
        listener.added(connection(as.bare(1)), as.jid(2, 0));
        verify(this.personalRoster, times(2)).of(as.bare(1));
        verify(this.personalRosterContent, times(1)).markAdded(as.bare(2));
    }

    @Test
    public void testSetBuddySubscription() throws Exception {
    }
    
    @Test
    public void testAddBuddySpecialChars() throws Exception {
        PersonalRosterListener listener = new PersonalRosterListener(personalRoster, defaultRoster, groupService, sessionUtil);
        this.defaultRosterContentMap.put(as.bare(1), map.value(as.user(1)));

        String name = "custom &amp; name";
        String[] groups = new String[]{"custom &amp; 1", null, "custom &amp; 2"};
        
        String nameUnescaped = "custom & name";
        String[] groupsUnescaped = new String[]{"custom & 1", null, "custom & 2"};

        //user 2 is adding user 1 to his default roster
        listener.addBuddy(connection(as.bare(2)), as.jid(1, 0), name, groups);

        verify(this.personalRoster, times(1)).of(as.bare(2));
        verify(this.personalRosterContent, times(1)).update(as.bare(1), nameUnescaped, groupsUnescaped);
        
    }

}
