/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.user.UserMapper;
import com.kerio.im.tigase.user.UserUtil;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;

/**
 *
 * @author tskoda
 */
public class RosterContentTest {

    private static final UserUtil as = new UserUtil();
    private static final PacketUtil packet = new PacketUtil();

    private static Map<BareJID, Element> toMap(List<Element> list) {
        Map<BareJID, Element> map = new HashMap<BareJID, Element>();
        for (Element el : list) {
            map.put(BareJID.bareJIDInstanceNS(el.getAttribute("jid")), el);
        }
        return map;
    }

    private Map<BareJID, RosterContent.Buddy> content;
    private ElementMapper mapper;
    private GroupService groupService;

    public RosterContentTest() {
    }

    @Before
    public void setUp() {
        this.content = new HashMap<BareJID, RosterContent.Buddy>();
        this.groupService = new TestGroupService();
        this.mapper = new ElementMapper(this.groupService, new UserMapper());
    }

    @After
    public void tearDown() {
    }

    @Test
    public void testFindNull() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        assertNull(roster.find(null));
    }

    @Test
    public void testFindEmpty() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        assertNull(roster.find(BareJID.bareJIDInstanceNS("name", "domain")));
    }

    @Test
    public void testFind() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        BareJID existing = BareJID.bareJIDInstanceNS("name", "domain");
        BareJID nonExisting = BareJID.bareJIDInstanceNS("name2", "domain");
        RosterContent.Buddy existingBuddy = mapper.value(existing, "Existing User", true, true);
        content.put(existing, existingBuddy);

        assertThat(roster.find(existing), is(existingBuddy));
        assertNull(roster.find(nonExisting));
    }

    @Test
    public void testAddNull() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        assertFalse(roster.add(null));
    }

    @Test
    public void testAddInvisibleOrDisabled() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        User disabled = as.disabled(1);
        User invisible = as.invisible(2);

        assertTrue(roster.add(disabled));
        assertTrue(roster.add(invisible));
    }

    @Test
    public void testAddAlreadyPresent() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        User twice = as.user(1);
        BareJID twiceJID = as.bare(twice);

        this.content.put(twiceJID, mapper.value(twice));

        assertFalse(roster.add(twice));
        assertFalse(roster.add(twice));
    }

    @Test
    public void testAdd() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        User user = as.user(1);
        RosterContent.Buddy userBuddy = mapper.value(user);

        assertTrue(roster.add(user));
        assertFalse(roster.add(user));

        assertThat(roster.find(as.bare(user)), is(userBuddy));
    }

    @Test
    public void testRemoveEmpty() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        User user = as.user(1);
        BareJID userJID = as.bare(user);

        assertFalse(roster.remove(userJID));
    }

    @Test
    public void testRemoveNull() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        assertFalse(roster.remove(null));
    }

    @Test
    public void testRemove() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        User user = as.user(1);
        BareJID userJID = as.bare(user);

        User nonExisting = as.user(2);
        BareJID nonExistingJID = as.bare(nonExisting);

        this.content.put(userJID, mapper.value(user));

        assertTrue(roster.remove(userJID));
        assertFalse(roster.remove(nonExistingJID));
    }

    @Test
    public void testRemoveTwice() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        User user = as.user(1);
        BareJID userJID = as.bare(user);

        this.content.put(userJID, mapper.value(user));

        assertTrue(roster.remove(userJID));
        assertFalse(roster.remove(userJID));
    }

    @Test
    public void testAllEmpty() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        assertThat(roster.all().size(), is(0));
    }

    @Test
    public void testAllAddRemove() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        roster.add(as.user(1));
        roster.add(as.user(2));
        assertThat(roster.all().size(), is(2));

        roster.remove(as.bare(1));
        assertThat(roster.all().size(), is(1));

        roster.remove(as.bare(2));
        assertThat(roster.all().size(), is(0));
    }

    @Test
    public void testAllImmutable() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        roster.add(as.user(1));
        roster.add(as.user(2));

        Collection<? extends BareJID> all = roster.all();
        assertThat(all.size(), is(2));
        assertThat(roster.all().size(), is(2));

        all.remove(as.bare(1));

        assertThat(all.size(), is(1));
        assertThat(roster.all().size(), is(2));
    }

    @Test
    public void testAsArrayNull() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        assertThat(roster.asArray(null, Collections.EMPTY_MAP).length, is(0));

    }

    @Test
    public void testAsArrayEmptyPersonal() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        roster.add(as.user(1));
        roster.add(as.user(2));
        roster.add(as.user(3));

        Set<JID> fullSet = new HashSet<JID>();
        fullSet.add(JID.jidInstance(as.bare(1)));
        fullSet.add(JID.jidInstance(as.bare(2)));
        fullSet.add(JID.jidInstance(as.bare(3)));

        JID[] fullArray = roster.asArray(as.bare(4), Collections.EMPTY_MAP);
        assertThat(fullArray.length, is(3));
        assertThat(new HashSet<JID>(Arrays.asList(fullArray)), is(fullSet));

        Set<JID> withoutOwnerSet = new HashSet<JID>();
        withoutOwnerSet.add(JID.jidInstance(as.bare(1)));
        withoutOwnerSet.add(JID.jidInstance(as.bare(2)));

        JID[] withoutOwnerArray = roster.asArray(as.bare(3), Collections.EMPTY_MAP);
        assertThat(withoutOwnerArray.length, is(2));
        assertThat(new HashSet<JID>(Arrays.asList(withoutOwnerArray)), is(withoutOwnerSet));
    }

    @Test
    public void testAsArrayNoIntersection() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        roster.add(as.user(1));
        roster.add(as.user(2));
        roster.add(as.user(3));

        Map<BareJID, PersonalRoster.Buddy> personal = new HashMap<BareJID, PersonalRoster.Buddy>();
        personal.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), "Full Name", new String[]{}, PersonalRoster.Buddy.FLAG.updated)); //not added, not removed, just updated
        personal.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.deleted)); //not added, removed
        personal.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), null, null, PersonalRoster.Buddy.FLAG.added)); //added, not removed

        Set<JID> resultSet = new HashSet<JID>();
        resultSet.add(JID.jidInstance(as.bare(2)));
        resultSet.add(JID.jidInstance(as.bare(3)));

        JID[] resultArray = roster.asArray(as.bare(1), personal);
        assertThat(new HashSet<JID>(Arrays.asList(resultArray)), is(resultSet));
    }

    @Test
    public void testAsArray() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        roster.add(as.user(1));
        roster.add(as.user(2));
        roster.add(as.user(3));
        roster.add(as.user(4));

        Map<BareJID, PersonalRoster.Buddy> personal = new HashMap<BareJID, PersonalRoster.Buddy>();
        //user 1 removed
        personal.put(as.bare(1), new PersonalRoster.Buddy(as.bare(1), null, null, PersonalRoster.Buddy.FLAG.deleted));
        //user 2 renamed
        personal.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), "Renamed", null, PersonalRoster.Buddy.FLAG.updated));
        //user 3 moved to another group
        personal.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), null, new String[]{"group1", "group2"}, PersonalRoster.Buddy.FLAG.updated));
        //user 5 added
        personal.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.added));

        Set<JID> resultSet = new HashSet<JID>();
        resultSet.add(JID.jidInstance(as.bare(2)));
        resultSet.add(JID.jidInstance(as.bare(3)));
        resultSet.add(JID.jidInstance(as.bare(4)));

        JID[] resultArray = roster.asArray(as.bare(6), personal);
        assertThat(new HashSet<JID>(Arrays.asList(resultArray)), is(resultSet));
    }

    @Test
    public void testAsListNull() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        assertThat(roster.asList(null, Collections.EMPTY_MAP).size(), is(0));
        //collection cant be null
        //assertNull(roster.asList(as.bare(1), null));
        //assertNull(roster.asList(null, null));

    }

    @Test
    public void testAsListEmptyPersonal() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        roster.add(as.user(1));
        roster.add(as.user(2));
        roster.add(as.user(3));

        List<Element> fullList = roster.asList(as.bare(4), Collections.EMPTY_MAP);
        assertThat(fullList.size(), is(3));

        Map<BareJID, Element> fullMap = toMap(fullList);
        assertThat(fullMap.get(as.bare(1)), is(packet.element(as.bare(1), "both", as.user(1).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(fullMap.get(as.bare(2)), is(packet.element(as.bare(2), "both", as.user(2).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(fullMap.get(as.bare(3)), is(packet.element(as.bare(3), "both", as.user(3).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));

        List<Element> withoutOwnerList = roster.asList(as.bare(3), Collections.EMPTY_MAP);
        assertThat(withoutOwnerList.size(), is(2));

        Map<BareJID, Element> withoutOwnerMap = toMap(withoutOwnerList);
        assertThat(withoutOwnerMap.get(as.bare(1)), is(packet.element(as.bare(1), "both", as.user(1).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(withoutOwnerMap.get(as.bare(2)), is(packet.element(as.bare(2), "both", as.user(2).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
    }

    @Test
    public void testAsListNoIntersection() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        roster.add(as.user(1));
        roster.add(as.user(2));
        roster.add(as.user(3));

        Map<BareJID, PersonalRoster.Buddy> personal = new HashMap<BareJID, PersonalRoster.Buddy>();
        personal.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), "Full Name", new String[]{}, PersonalRoster.Buddy.FLAG.updated)); //not added, not removed, just updated
        personal.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.deleted)); //not added, removed
        personal.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), null, null, PersonalRoster.Buddy.FLAG.added)); //added, not removed

        List<Element> resultList = roster.asList(as.bare(1), personal);
        assertThat(resultList.size(), is(2));

        Map<BareJID, Element> fullMap = toMap(resultList);
        assertThat(fullMap.get(as.bare(2)), is(packet.element(as.bare(2), "both", as.user(2).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(fullMap.get(as.bare(3)), is(packet.element(as.bare(3), "both", as.user(3).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
    }

    @Test
    public void testAsList() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);

        roster.add(as.user(1));
        roster.add(as.user(2));
        roster.add(as.user(3));
        roster.add(as.user(4));

        Map<BareJID, PersonalRoster.Buddy> personal = new HashMap<BareJID, PersonalRoster.Buddy>();
        //user 1 removed
        personal.put(as.bare(1), new PersonalRoster.Buddy(as.bare(1), null, null, PersonalRoster.Buddy.FLAG.deleted));
        //user 2 renamed
        personal.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), "Renamed", null, PersonalRoster.Buddy.FLAG.updated));
        //user 3 moved to another group
        personal.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), null, new String[]{"group1", "group2"}, PersonalRoster.Buddy.FLAG.updated));
        //user 5 added
        personal.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.added));

        List<Element> resultList = roster.asList(as.bare(6), personal);
        assertThat(resultList.size(), is(3));

        Map<BareJID, Element> resultMap = toMap(resultList);
        assertThat(resultMap.get(as.bare(2)), is(packet.element(as.bare(2), "both", "Renamed", Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(resultMap.get(as.bare(3)), is(packet.element(as.bare(3), "both", as.user(3).getFullName(), Arrays.asList("group1", "group2"))));
        assertThat(resultMap.get(as.bare(4)), is(packet.element(as.bare(4), "both", as.user(4).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));
    }

    @Test
    public void testSize() {
        RosterContent roster = new RosterContent(content, mapper, 10, false);
        assertThat(roster.size(), is(0));

        roster.add(as.user(1));
        assertThat(roster.size(), is(1));

        roster.remove(as.bare(1));
        assertThat(roster.size(), is(0));
    }

    @Test
    public void testSizeLimit() {
        RosterContent roster = new RosterContent(content, mapper, 2, false);

        assertThat(roster.size(), is(0));

        roster.add(as.user(1));
        assertThat(roster.size(), is(1));

        roster.add(as.user(2));
        assertThat(roster.size(), is(2));

        roster.add(as.user(3));
        assertThat(roster.size(), is(3)); //maximum 2 + 1 owner

        roster.add(as.user(4));
        assertThat(roster.size(), is(3));

        roster.remove(as.bare(4));
        assertThat(roster.size(), is(3));

        roster.remove(as.bare(3));
        assertThat(roster.size(), is(2));

        roster.remove(as.bare(2));
        assertThat(roster.size(), is(1));

    }

}
