/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.user;

import com.kerio.im.bean.User;
import com.kerio.im.tigase.Conversion;
import com.kerio.im.tigase.derby.DerbyDatabase;
import com.kerio.im.tigase.derby.DerbyUtilCommon;
import com.kerio.im.tigase.derby.DerbyUtilMemory;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.apache.derby.jdbc.EmbeddedDataSource40;
import org.hamcrest.BaseMatcher;
import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import tigase.xmpp.BareJID;
import tigase.xmpp.impl.roster.RosterAbstract;
import tigase.xmpp.impl.roster.RosterElement;
import tigase.xmpp.impl.roster.RosterFlat;

/**
 *
 * @author tskoda
 */
@Ignore
public abstract class XmppUsersStoreAbstract {

    protected static final String DOMAIN = "kerio.test";
    protected static final String USER_PREFIX = "user_";
    protected static final String ROSTER = ""
            + "select pr.PVAL as roster "
            + "  from APP.TIG_USERS as u "
            + "  left join APP.TIG_NODES as n "
            + "    on u.uid = n.UID and n.NODE = 'root' "
            + "  left join APP.TIG_PAIRS as pr "
            + "    on n.nid = pr.NID and pr.PKEY = 'roster' "
            + " where lower(u.user_id)=?";
    protected static final String ROSTER_UPDATE = ""
            + "update APP.TIG_PAIRS pr "
            + "   set pr.PVAL = ? "
            + " where pr.PKEY = 'roster' "
            + "   and pr.NID = ("
            + "select pn.NID"
            + "  from APP.TIG_NODES pn"
            + "  inner join APP.TIG_USERS u"
            + "     on pn.NODE = 'root' and pn.UID = u.UID and lower(u.user_id)=?)";
    protected static final String USER = ""
            + "select pf.PVAL as fullname, "
            + "        u.K_STATUS as status, "
            + "        u.K_GAL as visible "
            + "  from APP.TIG_USERS as u "
            + "  left join APP.TIG_NODES as n "
            + "    on u.uid = n.UID and n.NODE = 'root' "
            + "  left join APP.TIG_PAIRS as pf "
            + "    on n.nid = pf.NID and pf.PKEY = 'fullname' "
            + " where lower(u.user_id)=?";
    public static final String GROUP = "group";
    protected static final UserUtil as = new UserUtil();
    protected static final RosterElementUtil re = new RosterElementUtil();

    protected static class Matchers {

        protected static Matcher<User> visible() {
            return new BaseMatcher<User>() {
                @Override
                public boolean matches(Object item) {
                    User toTest = (User) item;
                    return toTest.isVisible();
                }

                @Override
                public void describeTo(Description description) {
                    description.appendText("is visible user");
                }
            };
        }

        protected static Matcher<User> enabled() {
            return new BaseMatcher<User>() {
                @Override
                public boolean matches(Object item) {
                    User toTest = (User) item;
                    return toTest.isEnabled();
                }

                @Override
                public void describeTo(Description description) {
                    description.appendText("is enabled user");
                }
            };
        }
    }

    protected static class RosterSerializer {

        public static RosterElement asRosterElement(User user, String[] groups, RosterAbstract.SubscriptionType subscription) {
            RosterElement result = new RosterElement(Conversion.toJID(user), user.getFullName(), groups, null);
            result.setSubscription(subscription);
            result.setManaged(true);
            return result;
        }

        public String get(Map<BareJID, RosterElement> roster) {
            StringBuilder result = new StringBuilder();
            for (RosterElement element : roster.values()) {
                result.append(asString(element));
            }
            return result.toString();
        }

        private String asString(RosterElement element) {
            return element.getRosterElement().toString();
        }

        private String asString(User user, String[] groups, RosterAbstract.SubscriptionType subscription) {
            return asString(asRosterElement(user, groups, subscription));
        }
    }

    private Connection connection;
    protected TestSessionProvider sessions;
    protected XmppUsersStore store;
    //test util classes
    protected Database database;

    protected XmppUsersStoreAbstract() {
        this.database = new Database();
    }

    @Before
    public void before() throws SQLException {
        repositoryReconnect();
        DerbyUtilCommon.init(this.connection, DerbyDatabase.TABLE.values(), DerbyDatabase.PROCEDURE.values(), DerbyDatabase.DATA.values());
    }

    @After
    public void after() throws SQLException {
        if (this.connection != null) {
            DerbyUtilCommon.close(this.connection);
        }
        DerbyUtilMemory.shutdown(DOMAIN);
    }

    protected void repositoryReconnect() throws SQLException {
        if (this.connection != null) {
            DerbyUtilCommon.close(this.connection);
        }

        this.sessions = new TestSessionProvider();
        this.connection = DerbyUtilMemory.create(DOMAIN);
        this.store = new XmppUsersStore(
                DOMAIN,
                DerbyUtilMemory.createDS(DOMAIN),
                this.sessions);
    }

    protected void repositoryAddAll(User[] users, Integer... except) {
        Set<Integer> exceptSet = except != null ? new HashSet(Arrays.asList(except)) : new HashSet();
        for (int i = 0; i < users.length; i++) {
            if (!exceptSet.contains(i)) {
                this.store.add(users[i]);
            }
        }
    }

    protected class Database {

        protected Roster roster = new Roster();

        protected User user(int i) throws SQLException {
            EmbeddedDataSource40 ds = DerbyUtilMemory.createDS(DOMAIN);
            Connection conn = ds.getConnection();
            PreparedStatement userStatement = conn.prepareStatement(USER);
            User result = null;
            try {
                userStatement.setString(1, XmppUsersStoreAbstract.this.as.bare(i).toString());
                ResultSet rosterResult = userStatement.executeQuery();
                try {
                    if (rosterResult.next()) {
                        result = new User(USER_PREFIX + i, DOMAIN, rosterResult.getString("fullname"), rosterResult.getBoolean("status"), rosterResult.getBoolean("visible"), false);
                    }
                } finally {
                    DerbyUtilCommon.close(rosterResult);
                }
            } finally {
                DerbyUtilCommon.close(userStatement);
            }

            return result;
        }

        protected Map<String, User> all() throws SQLException {
            EmbeddedDataSource40 ds = DerbyUtilMemory.createDS(DOMAIN);
            XmppUsersStore users = new XmppUsersStore(DOMAIN, ds, XmppUsersStoreAbstract.this.sessions);
            try {
                return users.all();
            } finally {
                users.bye();
            }
        }
    }

    protected class Roster {

        protected Map<BareJID, RosterElement> of(int user) throws SQLException {
            EmbeddedDataSource40 ds = DerbyUtilMemory.createDS(DOMAIN);
            Connection conn = ds.getConnection();
            Map<BareJID, RosterElement> result = new HashMap<BareJID, RosterElement>();
            PreparedStatement rosterStatement = conn.prepareStatement(ROSTER);
            try {
                rosterStatement.setString(1, USER_PREFIX + user + "@" + DOMAIN);
                ResultSet rosterResult = rosterStatement.executeQuery();
                try {
                    if (rosterResult.next()) {
                        RosterFlat.parseRosterUtil(rosterResult.getString(1), result, null);
                    }
                } finally {
                    DerbyUtilCommon.close(rosterResult);
                }
            } finally {
                DerbyUtilCommon.close(rosterStatement);
            }

            return result;
        }
    }
}
