/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.user;

import com.kerio.im.bean.User;
import com.kerio.im.xmpp.api.XmppUsers;
import java.sql.SQLException;
import java.util.EnumSet;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;
import org.junit.Test;
import static com.kerio.im.tigase.user.XmppUsersStoreAbstract.Matchers.*;

/**
 *
 * @author tskoda
 */
public class XmppUsersStoreUpdateTest extends XmppUsersStoreAbstract {

    @Test
    public void testUpdateName() throws SQLException {
        User[] users = {as.user(0), as.disabled(1), as.invisible(2), as.user(3)};
        repositoryAddAll(users);

        repositoryReconnect();

        User newNameUser0 = new User(users[0].getLogin(), users[0].getDomain(), "NEW FULLNAME NORMAL", users[0].isEnabled(), users[0].isVisible(), false);
        User newNameUser1 = new User(users[1].getLogin(), users[1].getDomain(), "NEW FULLNAME DISABLED", users[1].isEnabled(), users[1].isVisible(), false);
        User newNameUser2 = new User(users[2].getLogin(), users[2].getDomain(), "NEW FULLNAME DISABLED", users[2].isEnabled(), users[2].isVisible(), false);

        this.store.update(newNameUser0, EnumSet.of(XmppUsers.Changes.NAME));
        this.store.update(newNameUser1, EnumSet.of(XmppUsers.Changes.NAME));
        this.store.update(newNameUser2, EnumSet.of(XmppUsers.Changes.NAME));

        repositoryReconnect();

        assertThat(database.user(0).getFullName(), is(equalTo(newNameUser0.getFullName())));
        assertThat(database.user(1).getFullName(), is(equalTo(newNameUser1.getFullName())));
        assertThat(database.user(2).getFullName(), is(equalTo(newNameUser1.getFullName())));
        assertThat(database.user(3).getFullName(), is(equalTo(users[3].getFullName())));
    }

    @Test
    public void testUpdateNameNonExisting() throws SQLException {
        User user = as.user(0);
        this.store.add(user);

        assertThat(database.user(0).getFullName(), is(user.getFullName()));

        repositoryReconnect();
        assertFalse(this.store.update(as.user(1), EnumSet.of(XmppUsers.Changes.NAME)));
        assertThat(database.user(0).getFullName(), is(user.getFullName()));

    }

    @Test
    public void testUpdateStatusSingle() throws SQLException {
        User[] users = {as.user(0), as.disabled(1), as.user(2), as.user(3)};
        repositoryAddAll(users);

        assertThat(database.user(0), allOf(enabled(), visible()));
        assertThat(database.user(1), allOf(not(enabled()), visible()));
        assertThat(database.user(2), allOf(enabled(), visible()));
        assertThat(database.user(3), allOf(enabled(), visible()));

        //single change - disabled
        repositoryReconnect();
        assertThat(database.roster.of(0).size(), is(0));

        User user0disabled = as.disabled(users[0]);
        assertTrue(this.store.update(user0disabled, EnumSet.of(XmppUsers.Changes.STATUS)));

        assertThat(database.user(0), allOf(not(enabled()), visible()));

        repositoryReconnect();
        User user0enabled = as.enabled(users[0]);
        assertTrue(this.store.update(user0enabled, EnumSet.of(XmppUsers.Changes.STATUS)));

        assertThat(database.user(0), allOf(enabled(), visible()));
    }

    @Test
    public void testUpdateStatusMultiple() throws SQLException {
        User[] users = {as.user(0), as.disabled(1), as.user(2), as.user(3)};
        repositoryAddAll(users);

        assertThat(database.user(0), allOf(enabled(), visible()));
        assertThat(database.user(1), allOf(not(enabled()), visible()));
        assertThat(database.user(2), allOf(enabled(), visible()));
        assertThat(database.user(3), allOf(enabled(), visible()));

        //multiple change - front and back
        repositoryReconnect();

        User user0disabled = as.disabled(users[0]);
        User user1enabled = as.enabled(users[1]);
        User user3disabled = as.disabled(users[3]);

        assertTrue(this.store.update(user0disabled, EnumSet.of(XmppUsers.Changes.STATUS)));
        assertTrue(this.store.update(user1enabled, EnumSet.of(XmppUsers.Changes.STATUS)));
        assertTrue(this.store.update(user3disabled, EnumSet.of(XmppUsers.Changes.STATUS)));

        /**
         * disabled, enabled, enabled, disabled
         *
         * user0: 1, 2 (none) user1: 2 (both) user2: 1 (both) user3: 1, 2 (none)
         */
        assertThat(database.user(0), allOf(not(enabled()), visible()));
        assertThat(database.user(1), allOf(enabled(), visible()));
        assertThat(database.user(2), allOf(enabled(), visible()));
        assertThat(database.user(3), allOf(not(enabled()), visible()));
    }

    @Test
    public void testUpdateVisibilitySingle() throws SQLException {
        User[] users = {as.user(0), as.invisible(1), as.user(2), as.user(3)};
        repositoryAddAll(users);

        assertThat(database.user(0), allOf(enabled(), visible()));
        assertThat(database.user(1), allOf(enabled(), not(visible())));
        assertThat(database.user(2), allOf(enabled(), visible()));
        assertThat(database.user(3), allOf(enabled(), visible()));

        //single change - user 0 invisible
        repositoryReconnect();

        User user0invisible = as.invisible(users[0]);
        assertTrue(this.store.update(user0invisible, EnumSet.of(XmppUsers.Changes.VISIBILITY)));

        assertThat(database.user(0), allOf(enabled(), not(visible())));

        //single change - enable
        repositoryReconnect();
        User user0visible = as.visible(users[0]);
        assertTrue(this.store.update(user0visible, EnumSet.of(XmppUsers.Changes.VISIBILITY)));

        assertThat(database.user(0), allOf(enabled(), visible()));
    }

    @Test
    public void testUpdateVisibilityMultiple() throws SQLException {
        User[] users = {as.user(0), as.invisible(1), as.user(2), as.user(3)};
        repositoryAddAll(users);

        assertThat(database.user(0), allOf(enabled(), visible()));
        assertThat(database.user(1), allOf(enabled(), not(visible())));
        assertThat(database.user(2), allOf(enabled(), visible()));
        assertThat(database.user(3), allOf(enabled(), visible()));

        //multiple change - front and back
        repositoryReconnect();

        User user0invisible = as.invisible(users[0]);
        User user1visible = as.visible(users[1]);
        User user3invisible = as.invisible(users[3]);

        assertTrue(this.store.update(user0invisible, EnumSet.of(XmppUsers.Changes.VISIBILITY)));
        assertTrue(this.store.update(user1visible, EnumSet.of(XmppUsers.Changes.VISIBILITY)));
        assertTrue(this.store.update(user3invisible, EnumSet.of(XmppUsers.Changes.VISIBILITY)));

        /**
         * disabled, enabled, enabled, disabled
         *
         * user0: 1, 2 (none) user1: 2 (both) user2: 1 (both) user3: 1, 2 (none)
         */
        assertThat(database.user(0), allOf(enabled(), not(visible())));
        assertThat(database.user(1), allOf(enabled(), visible()));
        assertThat(database.user(2), allOf(enabled(), visible()));
        assertThat(database.user(3), allOf(enabled(), not(visible())));
    }

    @Test
    public void testUpdateNotExisting() throws SQLException {
        User[] users = {as.user(0), as.disabled(1), as.user(2), as.user(3)};
        repositoryAddAll(users);

        assertThat(database.roster.of(0).size(), is(0));
        assertThat(database.roster.of(1).size(), is(0));
        assertThat(database.roster.of(2).size(), is(0));
        assertThat(database.roster.of(3).size(), is(0));

        repositoryReconnect();

        assertFalse(this.store.update(as.user(4), EnumSet.allOf(XmppUsers.Changes.class)));

        assertThat(database.user(4), is(nullValue()));
    }
}
