/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.service.task;

import com.kerio.im.common.Task;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.lang.model.type.NullType;

/**
 *
 * @author tskoda
 */
public class PrintSystemConfiguration implements Task<NullType> {

//    private interface Environment {
//
//    }
    private class UnixUtils {

        void output(Process process, List<String> output) throws IOException {
            output(process.getInputStream(), output);

            List<String> errors = new LinkedList<String>();
            output(process.getErrorStream(), errors);

            if (errors.size() > 0) {
                throw new IOException(errors.toString());
            }
        }

        void output(InputStream stream, List<String> output) throws IOException {
            BufferedReader is = new BufferedReader(new InputStreamReader(stream));
            try {
                String line;
                while ((line = is.readLine()) != null) {
                    output.add(line);
                }
            } finally {
                try {
                    is.close();
                } catch (IOException e) {
                }
            }
        }
        
        String getSysCtl(String value, String delimiter) {
            String[] exec = {"/bin/sh", "-c", "sysctl -a | grep '"+ value + "'"};
            String result = "-";

            try {
                Runtime runtime = Runtime.getRuntime();
                Process process = runtime.exec(exec);
                List<String> output = new LinkedList<String>();

                output(process, output);
                for (String o : output) {
                    String[] parsed = o.split(delimiter, 2);
                    if ((parsed.length == 2) && (parsed[1] != null)) {
                        result = parsed[1].trim();
                        break;
                    }
                }

            } catch (Exception e) {
                LOG.log(Level.INFO, "Getting '{0}' from sysctl output failed, {1}", new String[]{value, e.getMessage()});
            }

            return result;
        }

        String getUlimit() {
            //String exec = "ulimit -n";
            String[] exec = {"/bin/sh", "-c", "ulimit -n"};
            String result = "-";

            try {
                Runtime runtime = Runtime.getRuntime();
                Process process = runtime.exec(exec);
                List<String> output = new LinkedList<String>();

                output(process, output);
                for (String o : output) {
                    if ((o != null) && (!o.isEmpty())) {
                        result = o.trim();
                        break;
                    }
                }

            } catch (Exception e) {
                LOG.log(Level.INFO, "Getting ulimit output failed, {0}", e.getMessage());
            }

            return result;
        }
    }

    private class UnixEnvironment {
        private final UnixUtils util = new UnixUtils();

        public String getCurrentSoftLimit() {
            return util.getUlimit();
        }

        public String getKernelMaxFiles() {
            return util.getSysCtl("fs.file-max", "=");
        }

    }

    private class MacEnvironment {

        protected final UnixUtils util = new UnixUtils();

        public String getCurrentSoftLimit() {
            return util.getUlimit();
        }

        private String getKernelMaxFiles() {
              return util.getSysCtl("kern.maxfiles:", ":");
        }

        private String getKernelMaxProcFiles() {
            return util.getSysCtl("kern.maxfilesperproc:", ":");
        }
    }

    private enum OS {

        MAC,
        LINUX,
        WIN,
        UNKNOWN;

        private OS() {
        }

        static OS current(String osName) {
            osName = osName == null ? "unknown" : osName.toLowerCase();

            if (osName.startsWith("windows")) {
                return WIN;
            } else if (osName.startsWith("linux")) {
                return LINUX;
            } else if (osName.startsWith("mac") || osName.startsWith("darwin")) {
                return MAC;
            } else {
                return UNKNOWN;
            }
        }

    }

    private static final Logger LOG = Logger.getLogger(PrintSystemConfiguration.class.getName());

    public PrintSystemConfiguration() {
    }

    @Override
    public void run(NullType parameter) {
        if (LOG.isLoggable(Level.INFO)) {
            try {

                OS current = OS.current(System.getProperties().getProperty("os.name"));
                Properties systemProperties = System.getProperties();
                LOG.log(Level.INFO, "Printing environment report  ... ");
                LOG.log(Level.INFO, "- operating system {0} (name / version / architecture): {1} / {2} / {3}", new Object[]{current, systemProperties.getProperty("os.name", "-"), systemProperties.getProperty("os.version", "-"), systemProperties.getProperty("os.arch", "-")});
                LOG.log(Level.INFO, "- memory (current / free / max): {0} MB / {1} MB / {2} MB", new Object[]{Runtime.getRuntime().totalMemory() / 1000000, Runtime.getRuntime().freeMemory() / 1000000, Runtime.getRuntime().maxMemory() / 1000000});

                switch (current) {
                    case MAC:
                        MacEnvironment env = new MacEnvironment();
                        LOG.log(Level.INFO, "- open files limits (soft / kernel per process / kernel total): {0} / {1} / {2}", new Object[]{env.getCurrentSoftLimit(), env.getKernelMaxProcFiles(), env.getKernelMaxFiles()});
                        break;
                    case LINUX:
                        UnixEnvironment unixEnv = new UnixEnvironment();
                        LOG.log(Level.INFO, "- open files limits (soft / kernel total): {0} / {1} ", new Object[]{unixEnv.getCurrentSoftLimit(), unixEnv.getKernelMaxFiles()});
                        break;
                    default:
                        break;
                }

                List<String> propertiesNames = Collections.list((Enumeration<String>) systemProperties.propertyNames());
                Collections.sort(propertiesNames);
                for (String property : propertiesNames) {
                    if ((property != null)
                            && (property.startsWith("java."))
                            && (!property.startsWith("java.vendor") && !property.startsWith("java.specification") && !property.startsWith("java.vm.specification")
                            && !property.startsWith("java.awt"))) {// && (!property.startsWith("java.library.path"))) {
                        LOG.log(Level.INFO, "- {0}: {1}", new Object[]{property, systemProperties.getProperty(property)});
                    }
                }
            } catch (Exception e) {
                LOG.log(Level.INFO, "Printing environment report failed", e);
            }
        }

    }

}
