/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.service.task;

import com.kerio.im.common.Lookup;
import com.kerio.im.common.Task;
import com.kerio.im.common.error.Code;
import com.kerio.im.core.CoreRuntime;
import com.kerio.im.core.CoreRuntimeFactory;
import com.kerio.im.core.ErrorHandler;
import com.kerio.im.core.api.event.StartupFinishedEvent;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.api.service.ThreadService;
import com.kerio.im.core.api.service.ThreadService.ShutdownListener;
import com.kerio.im.core.auth.ControlAuthEvent;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.external.ExtService;
import com.kerio.im.external.ExtServiceFactory;
import com.kerio.im.external.ImService;
import com.kerio.im.external.ImServiceFactory;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.XmppService;
import com.kerio.im.xmpp.api.XmppServiceFactory;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class StartServer implements Task<Map<String, String>> {

    private static final Logger LOG = Logger.getLogger(StartServer.class.getName());
    private StartupTimeout startupTimeout;

    @Override
    public void run(Map<String, String> cmdProperties) {

        CoreRuntimeFactory coreRuntimeFactory = new CoreRuntimeFactory();
        LOG.log(Level.FINE, "Initializing core component");
        CoreRuntime coreRuntime = coreRuntimeFactory.get(cmdProperties);
        LOG.log(Level.INFO, "Initializing core component successful");

        this.startupTimeout = new StartupTimeout(cmdProperties, coreRuntime.errors(), coreRuntime.threads());

        coreRuntime.bus().subscribe(CommandEvent.class, new LazyExtServiceLoader(coreRuntime));

        LOG.log(Level.FINE, "Initializing xmpp component");
        XmppServiceFactory xmppServiceFactory = Lookup.service(XmppServiceFactory.class);
        XmppService xmppService = xmppServiceFactory.get(cmdProperties, coreRuntime);
        LOG.log(Level.INFO, "Initializing xmpp component successful");

        coreRuntime.setXmppService(xmppService);
        LOG.log(Level.INFO, "Instant Messaging service is waiting for control user login");
        coreRuntime.bus().subscribe(ControlAuthEvent.class, new ControlUserAuthListener());
    }

    protected class ControlUserAuthListener implements ServiceBus.EventListener<ControlAuthEvent> {

        @Override
        public void event(ControlAuthEvent event) {
            StartServer.this.startupTimeout.cancel();
            LOG.log(Level.INFO, "Instant Messaging service is waiting for initial configuration message");
        }
    }

    protected class LazyExtServiceLoader implements ServiceBus.EventListener<CommandEvent> {

        private static final String COMMAND = "config";
        private final CoreRuntime coreRuntime;

        public LazyExtServiceLoader(CoreRuntime coreRuntime) {
            this.coreRuntime = coreRuntime;
        }

        @Override
        public void event(CommandEvent event) {
            if (event.interested(COMMAND)) {
                try {
                    LOG.log(Level.FINE, "Initializing external component");
                    ExtServiceFactory factory = Lookup.service(ExtServiceFactory.class);
                    ExtService service = factory.getService(null, event.getContent());
                    LOG.log(Level.INFO, "Initializing external component successful");
                    this.coreRuntime.setExtService(service);
                    this.coreRuntime.bus().publish(new StartupFinishedEvent(event.getContent()));
                    LOG.log(Level.INFO, "Instant Messaging server is successfuly configured and running");
                } catch (Exception e) {
                    LOG.log(Level.INFO, "Initializing external component failed", e);
                    LOG.log(Level.WARNING, "Initializing external component failed, reason: {0}", e.getMessage());
                    event.fail("Initializing external component failed");
                    this.coreRuntime.errors().error(Code.EXT_SERVICE_ERROR);
                }
            }
        }
    }

    protected class StartupTimeout {

        private static final String TIMEOUT_PROPERTY = "service.im.timeout";
        public static final boolean DO_NOT_INTERUPT_IF_RUNNING = false;
        private final ShutdownListener submit;

        public StartupTimeout(Map<String, String> cmdProperties, ErrorHandler errorHandler, ThreadService threadService) {

            if (cmdProperties.containsKey(TIMEOUT_PROPERTY)) {
                try {
                    long timeout = Long.valueOf(cmdProperties.get(TIMEOUT_PROPERTY));
                    if (timeout > 0) {
                        submit = threadService.submit(new StartupTimeoutTask(errorHandler), timeout, TimeUnit.SECONDS);
                        LOG.log(Level.INFO, "Startup timeout set to {0} seconds and counting ... ", timeout);
                    } else {
                        submit = null;
                    }
                } catch (NumberFormatException e) {
                    throw new XmppException("Startup property {0} has incorrect value '{1}'", Code.TIMEOUT, TIMEOUT_PROPERTY, cmdProperties.get(TIMEOUT_PROPERTY));
                }
            } else {
                submit = null;
            }
        }

        public void cancel() {
            if (this.submit != null) {
                if (submit.getFuture().cancel(DO_NOT_INTERUPT_IF_RUNNING)) {
                    LOG.log(Level.FINE, "Startup timeout dismissed");
                }
            }
        }
    }

    protected class StartupTimeoutTask implements Runnable {

        private final ErrorHandler errorHandler;

        public StartupTimeoutTask(ErrorHandler errorHandler) {
            this.errorHandler = errorHandler;
        }

        @Override
        public void run() {
            LOG.log(Level.WARNING, "Startup timeout reached, closing server");
            this.errorHandler.error(Code.TIMEOUT);
        }
    }
}
