/*
 * Tigase Jabber/XMPP Multi-User Chat Component
 * Copyright (C) 2008 "Bartosz M. Małkowski" <bartosz.malkowski@tigase.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://www.gnu.org/licenses/.
 *
 * $Rev$
 * Last modified by $Author$
 * $Date$
 */
package tigase.muc;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import tigase.collections.TwoHashBidiMap;
import tigase.muc.repository.RepositoryException;
import tigase.util.TigaseStringprepException;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;

/**
 * @author bmalkow
 * 
 */
public class Room {

	private static class OccupantEntry {

		public BareJID jid;

		private final Set<JID> jids = new HashSet<JID>();

		private Role role = Role.none;

	}

	public static interface RoomListener {
		void onChangeSubject(Room room, String nick, String newSubject, Date changeDate);

		void onSetAffiliation(Room room, BareJID jid, Affiliation newAffiliation);
	}

	/**
	 * <bareJID, Affiliation>
	 */
	private final Map<BareJID, Affiliation> affiliations = new ConcurrentHashMap<BareJID, Affiliation>();

	private final RoomConfig config;

	private final Date creationDate;

	private final BareJID creatorJid;

	private final List<RoomListener> listeners = new CopyOnWriteArrayList<RoomListener>();

	/**
	 * < nickname,real JID>
	 */
	private final TwoHashBidiMap<String, OccupantEntry> occupants = new TwoHashBidiMap<String, OccupantEntry>();

	private final PresenceStore presences = new PresenceStore();

	private boolean roomLocked;

	private String subject;

	private Date subjectChangeDate;

	private String subjectChangerNick;

	/**
	 * @param rc
	 * @param creationDate
	 * @param creatorJid2
	 */
	public Room(RoomConfig rc, Date creationDate, BareJID creatorJid) {
		this.config = rc;
		this.creationDate = creationDate;
		this.creatorJid = creatorJid;
	}

	/**
	 * @param jid
	 * @param owner
	 * @throws RepositoryException
	 */
	public void addAffiliationByJid(BareJID jid, Affiliation affiliation) throws RepositoryException {
		if (affiliation == Affiliation.none) {
			this.affiliations.remove(jid);
		} else {
			this.affiliations.put(jid, affiliation);
		}
		fireOnSetAffiliation(jid, affiliation);
	}

	public void addListener(RoomListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * @param senderJid
	 * @param nickName
	 */
	public void addOccupantByJid(JID senderJid, String nickName, Role role) {
		OccupantEntry entry = this.occupants.get(nickName);
		if (entry == null) {
			entry = new OccupantEntry();
			entry.jid = senderJid.getBareJID();
			this.occupants.put(nickName, entry);
		}

		entry.role = role;
		entry.jids.add(senderJid);
	}

	/**
	 * @param senderJid
	 * @param nickName
	 */
	public void changeNickName(JID senderJid, String nickName) {
		OccupantEntry occ = getBySenderJid(senderJid);
		String oldNickname = this.occupants.getKey(occ);

		this.occupants.remove(oldNickname);
		this.occupants.put(nickName, occ);
	}

	private void fireOnSetAffiliation(BareJID jid, Affiliation affiliation) {
		for (RoomListener listener : this.listeners) {
			listener.onSetAffiliation(this, jid, affiliation);
		}
	}

	private void fireOnSetSubject(String nick, String subject, Date changeDate) {
		for (RoomListener listener : this.listeners) {
			listener.onChangeSubject(this, nick, subject, changeDate);
		}
	}

	/**
	 * @param value
	 *            user JID
	 * @return
	 */
	public Affiliation getAffiliation(BareJID jid) {
		Affiliation result = null;
		if (jid != null) {
			result = this.affiliations.get(jid);
		}
		return result == null ? Affiliation.none : result;
	}

	/**
	 * @param occupantNickname
	 * @return
	 */
	public Affiliation getAffiliation(String nickname) {
		OccupantEntry entry = this.occupants.get(nickname);
		return getAffiliation(entry == null ? null : entry.jid);
	}

	/**
	 * @return
	 */
	public Collection<BareJID> getAffiliations() {
		return this.affiliations.keySet();
	}

	private OccupantEntry getBySenderJid(JID sender) {
		for (Entry<String, OccupantEntry> e : occupants.entrySet()) {
			if (e.getValue().jids.contains(sender)) {
				return e.getValue();
			}
		}
		return null;
	}

	public RoomConfig getConfig() {
		return config;
	}

	/**
	 * @return
	 */
	public Date getCreationDate() {
		return this.creationDate;
	}

	public BareJID getCreatorJid() {
		return creatorJid;
	}

	public Element getLastPresenceCopyByJid(BareJID occupantJid) {
		Element e = this.presences.getBestPresence(occupantJid);
		if (e != null) {
			return e.clone();
		} else {
			return null;
		}
	}

	/**
	 * @return
	 */
	public int getOccupantsCount() {
		return this.occupants.size();
	}

	/**
	 * @param occupantNickname
	 * @return
	 */
	public BareJID getOccupantsJidByNickname(String nickname) {
		OccupantEntry entry = this.occupants.get(nickname);
		if (entry == null)
			return null;

		if (!entry.jids.isEmpty()) {
			return entry.jids.iterator().next().getBareJID();
		}

		return null;
	}

	/**
	 * @param recipientNickame
	 * @return
	 */
	public Collection<JID> getOccupantsJidsByNickname(final String nickname) {
		OccupantEntry entry = this.occupants.get(nickname);
		if (entry == null)
			return new ArrayList<JID>();

		return Collections.unmodifiableCollection(entry.jids);
	}

	/**
	 * @param jid
	 * @return
	 */
	public String getOccupantsNickname(JID jid) {
		OccupantEntry e = getBySenderJid(jid);
		if (e == null)
			return null;

		String nickname = occupants.getKey(e);

		return nickname;
	}

	/**
	 * @return
	 */
	public Collection<String> getOccupantsNicknames() {
		return Collections.unmodifiableCollection(this.occupants.keySet());
	}

	/**
	 * @param occupantBareJid
	 * @return
	 */
	public Collection<String> getOccupantsNicknames(BareJID bareJid) {
		Set<String> result = new HashSet<String>();

		for (Entry<String, OccupantEntry> e : this.occupants.entrySet()) {
			if (e.getValue().jid.equals(bareJid)) {
				result.add(e.getKey());
			}
		}

		return Collections.unmodifiableCollection(result);
	}

	/**
	 * @param nickName
	 * @return
	 */
	public Role getRole(String nickname) {
		OccupantEntry entry = this.occupants.get(nickname);
		if (entry == null)
			return Role.none;
		return entry.role == null ? Role.none : entry.role;
	}

	public BareJID getRoomJID() {
		return this.config.getRoomJID();
	}

	/**
	 * @return
	 */
	public String getSubject() {
		return subject;
	}

	public Date getSubjectChangeDate() {
		return subjectChangeDate;
	}

	/**
	 * @return
	 */
	public String getSubjectChangerNick() {
		return subjectChangerNick;
	}

	/**
	 * @param senderJID
	 * @return
	 */
	public boolean isOccupantInRoom(final JID jid) {
		return getBySenderJid(jid) != null;
	}

	public boolean isRoomLocked() {
		return roomLocked;
	}

	public void removeListener(RoomListener listener) {
		this.listeners.remove(listener);
	}

	/**
	 * 
	 * @param jid
	 * @return <code>true</code> if no more JIDs assigned to nickname. In other
	 *         words: nickname is removed
	 */
	public boolean removeOccupant(JID jid) {
		OccupantEntry e = getBySenderJid(jid);
		if (e != null) {
			e.jids.remove(jid);
			if (e.jids.isEmpty()) {
				this.occupants.removeValue(e);
				return true;
			}
		}
		return false;
	}

	/**
	 * @param occupantNick
	 */
	public void removeOccupant(String occupantNick) {
		this.occupants.remove(occupantNick);
	}

	/**
	 * @param affiliations2
	 */
	public void setAffiliations(Map<BareJID, Affiliation> affiliations) {
		this.affiliations.clear();
		this.affiliations.putAll(affiliations);
	}

	/**
	 * @param occupantNick
	 * @param newRole
	 */
	public void setNewRole(String nickname, Role newRole) {
		OccupantEntry entry = this.occupants.get(nickname);
		if (entry != null)
			entry.role = newRole;

	}

	/**
	 * @param msg
	 * @param senderRoomJid
	 * @throws RepositoryException
	 */
	public void setNewSubject(String msg, String senderNickname) throws RepositoryException {
		this.subjectChangerNick = senderNickname;
		this.subject = msg;
		this.subjectChangeDate = new Date();
		fireOnSetSubject(senderNickname, msg, this.subjectChangeDate);
	}

	public void setRoomLocked(boolean roomLocked) {
		this.roomLocked = roomLocked;
	}

	public void setSubjectChangeDate(Date subjectChangeDate) {
		this.subjectChangeDate = subjectChangeDate;
	}

	/**
	 * @param nickName
	 * @param element
	 * @throws TigaseStringprepException
	 */
	public void updatePresenceByJid(JID jid, Element cp) throws TigaseStringprepException {
		if (cp == null)
			this.presences.remove(jid);
		else
			this.presences.update(cp);
	}
        
        public int getNumberOf(Affiliation affiliation) {
            int result = 0;
            for (Affiliation existing: this.affiliations.values()) {
                if (existing == affiliation) {
                    result++;
                }
            }
            
            return result;
        }

}
